/**
	$Id$

	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.list;

import java.util.Collection;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Type;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.extensions.Action;
import net.sf.basedb.util.formatter.CollectionFormatter;
import net.sf.basedb.util.formatter.Formatter;

/**
	Action that inserts a new column into a table listing. The action is used multiple
	times while generating the table. The first use is creating a column definition. 
	This uses the id, title, value type, properties, sortable and filterable flags, 
	formatter, etc. Then, for each row in the table, the {@link #getValue(DbControl, Object)}
	is used to get the actual value that should be displayed in the table. If 
	a formatter has been defined, it is used to format the value, otherwise
	the {@link Object#toString()} method is used.
	
	@author Nicklas
	@since 3.2
	@base.modified $Date$
*/
public interface ListColumnAction<I, V>
	extends Action
{

	/**
		ID attribute of the column. This is requried and must be
		a unique id among all columns in the list.
	*/
	public String getId();

	/**
		Optional HTML class attribute that is used to style the column
		through CSS.
	*/
	public String getClazz();

	/**
		Optional HTML style attribute that is used to style the column
		through CSS.
	*/
	public String getStyle();
	
	
	/**
		The column title. Required.
	*/
	public String getTitle();
	
	/**
		An optional subtitle.
		@since 3.18
	*/
	public default String getSubtitle()
	{
		return null;
	}
	
	/**
		Optional extra classes to add to table cell (td)
		tags for this column.
		@since 3.18
	*/
	public default String getCellClass()
	{
		return null;
	}
	
	/**
		Optional extra style definition to add to table cell (td)
		tags for this column.
		@since 3.18
	*/
	public default String getCellStyle()
	{
		return null;
	}
	
	/**
		If set, text overflow check is disabled for this column.
		If not set, the global settings for the logged in user is used.
		@since 3.18
	*/
	public default boolean getDisableOverflowCheck()
	{
		return false;
	}
	
	/**
		An optional tooltip that is displayed when the mouse is over the 
		column header.
	*/
	public String getTooltip();
	
	/**
		The property name for referencing data in the column.
		This is optional but is needed to enable filtering, sorting
		and/or exporting.
		@return The property or null
	*/
	public String getProperty();
	
	/**
		The data type of the values that are referenced by the property.
		Required if a property has been set.
		@return The data type of the values
	*/
	public Type getValueType();
	
	/**
		A flag to indicate if sorting by this column is enabled for
		not. The sort uses the default property unless it has been
		overridden by {@link #getSortProperty()}.
		@return TRUE to enable sorting, FALSE to disable
	*/
	public boolean isSortable();
	
	/**
		Override the default property when this column is used for sorting.
		@return NULL if the default property should be used for sorting
	*/
	public String getSortProperty();

	/**
		A flag to indicate if filtering on this column is enabled or
		not. The filter uses the default property unless it has been
		overridden by {@link #getFilterProperty()}.
		@return TRUE to enable filtering, FALSE to disable
	*/
	public boolean isFilterable();
	
	/**
		Override the default property when this column is used for filtering.
		@return NULL if the default property should be used for filtering
	*/
	public String getFilterProperty();

	/**
		A flag to indicate if it is possible to export this column with the
		generic table exporter or not. The export uses the default property 
		unless it has been overridden by {@link #getExportProperty()}.
		@return TRUE to enable exporting, FALSE to disable
	*/
	public boolean isExportable();
	
	/**
		Override the default property when this column is used for exporting.
		@return NULL if the default property should be used for exporting
	*/
	public String getExportProperty();
	
	/**
		Get a formatter that converts the exported value to a string that is written
		to the file. The {@link Formatter#format(Object)} method is called once for 
		every item that is listed in the table. 
		@return A formatter, or null to simply use <code>Object.toString()</code>
	*/
	@SuppressWarnings("rawtypes")
	public Formatter getExportFormatter();
	
	/**
		Get the value that should be used when exporting to a file. This method is 
		called once for every item that is listed in the table. The returned
		value is then passed to the formatter (if any) that is retrieved from
		{@link #getExportFormatter()}.
		@param dc An open DbControl than can be used to query the database
		@param item The current item
		@return The value that should be exported
	*/
	public Object getExportValue(DbControl dc, I item);
	
	/**
		Get an enumeration of possible values that can be found in the
		column. This is used to generate a filter with a drop-down
		containing checkboxes for each option. If no enumeration is
		specified, a default filter input field is used.
	*/
	public Enumeration<String, String> getEnumeration();
	
	/**
		Get a formatter that converts the current value to a string that is displayed
		in the column. The string may include HTML that is required for proper display.
		The {@link Formatter#format(Object)} method is called once for 
		every item that is listed in the table. 
		@return A formatter, or null to simply use <code>Object.toString()</code>
	*/
	public Formatter<? super V> getFormatter();
	
	/**
		Get a formatter that is intended to format a collection of values
		that have been retrieved from this action. The default implementation
		will return a formatter that creates a comma-separated string of
		the individual values where each value is formatted with the formatter
		from {@link #getFormatter()}.
		@since 3.19.4
	*/
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public default Formatter<Collection<? super V>> getCollectionFormatter()
	{
		return new CollectionFormatter(getFormatter());
	}
	
	/**
		Get the value that should be displayed in the column. This method is 
		called once for every item that is listed in the table. The returned
		value is then passed to the formatter (if any) that is retrieved from
		{@link #getFormatter()}.
		@param dc An open DbControl than can be used to query the database
		@param item The current item
		@return The value that should be displayed
	*/
	public V getValue(DbControl dc, I item);
	
	
}
