package net.sf.basedb.clients.web.extensions.list;

import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.clients.web.extensions.AbstractJspActionFactory;
import net.sf.basedb.clients.web.extensions.ExtensionsControl;
import net.sf.basedb.clients.web.extensions.JspContext;
import net.sf.basedb.clients.web.extensions.list.RelatedItemColumn.Specification;
import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.Metadata.PropertyPath;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.SyncFilter.SourceItemTransform;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.ExtensionsFilter;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.InvokationContext;
import net.sf.basedb.util.extensions.SingleExtensionFilter;

/**
	Action factory for adding columns to a table that pull data from a parent
	or child item from an arbitrary location up or down the chain of linked items.
	The factory will scan the "columns" setting from the current context for 
	entries matching the pattern: /ITEMTYPE/Subtype/property-or-annotation
	
	It will use the item list API to find a
	parent item that matches the ITEMTYPE and Subtype and the retrieve
	the property using a {@link PropertyPath} or an annotation.

	@author nicklas
	@since 3.16, 3.17
*/
public class RelatedItemColumnActionFactory
	extends AbstractJspActionFactory<ListColumnAction<Annotatable, Object>>
{
	
	public RelatedItemColumnActionFactory()
	{}

	/**
		We get the "columns" setting from the current ItemContext and check if
		there seems to be an entry for a related item.
	*/
	@Override
	public boolean prepareContext(InvokationContext<? super ListColumnAction<Annotatable, Object>> context) 
	{
		JspContext jspContext = (JspContext)context.getClientContext();
		SessionControl sc = jspContext.getSessionControl();
		GuiContext guiContext = jspContext.getGuiContext();
		Item item = guiContext.getItem();
		String subContext = Values.getString(guiContext.getSubContext(), "");
		
		String allColumns = null;
		SourceItemTransform requiredDirection = null;
		if (RelatedItemExtensionColumn.SUBCONTEXT.equals(subContext)) 
		{
			// The current item should be a Specification instance that may specify
			// the column. Otherwise we load the columns from the regular list context
			// which makes it possible to select linked item columns from a parent/child
			// that are currently available
			Specification spec = jspContext.getCurrentItem();
			if (spec != null) 
			{
				allColumns = spec.actionId;
				if (spec.direction == SourceItemTransform.CHILD_TO_PARENT)
				{
					// We may only select CHILD columns from a PARENT
					requiredDirection = SourceItemTransform.PARENT_TO_CHILD;
					subContext = "";
				}
			}
		}
		if (allColumns == null)
		{
			ItemContext cc = sc.getCurrentContext(item, subContext, null);
			if (cc == null) return false;
			allColumns = cc.getSetting("columns");
		}
		if (allColumns == null || !allColumns.contains("/")) return false;

		boolean lazy = !Boolean.FALSE.equals(jspContext.getAttribute("lazy-loading"));
		if (lazy)
		{
			jspContext.addScript(jspContext.getRoot() + "/include/scripts/lazy-items.js");
		}
		
		for (String col : allColumns.split(","))
		{
			if (col.startsWith("/") && col.contains("!"))
			{
				Specification spec = Specification.parse(item, col);
				if (requiredDirection != null && requiredDirection != spec.direction) continue;
				if (spec.property.startsWith("/")) 
				{
					// We support multi-hop in max two steps
					spec = Specification.parse(spec.targetType, spec.property);
				}
				
				// Create a child context for invoking the specified extensions as if we are on the relevant list page
				GuiContext childGuiContext = GuiContext.list(spec.targetType, RelatedItemExtensionColumn.SUBCONTEXT);
				JspContext childJspContext = ExtensionsControl.createContext(jspContext.getDbControl(), jspContext.getPageContext(), childGuiContext, spec);
				childJspContext.linkAttributes(jspContext);
				ExtensionsFilter filter = spec.extensionId == null ? null : new SingleExtensionFilter(spec.extensionId);
				ExtensionsInvoker<ListColumnAction<Annotatable, ?>> childInvoker = ListColumnUtil.useExtensions(childJspContext, filter);

				if (childInvoker.getNumExtensions() > 0)
				{
					// Copy scripts and stylesheets to the main JSP context
					jspContext.addScripts(childJspContext.getScripts());
					jspContext.addStylesheets(childJspContext.getStylesheets());
					
					// Save the invoker in the main JSP context so that we can use the same 
					// instance in getActions()->RelatedItemColumn.create()
					jspContext.setAttribute("RelatedItemInvoker."+spec.expression, childInvoker);
				}
			}
		}
		
		return super.prepareContext(context);
	}

	
	@SuppressWarnings({"unchecked", "rawtypes"})
	@Override
	public ListColumnAction<Annotatable, Object>[] getActions(InvokationContext<? super ListColumnAction<Annotatable, Object>> context) 
	{
		JspContext jspContext = (JspContext)context.getClientContext();
		SessionControl sc = jspContext.getSessionControl();
		GuiContext guiContext = jspContext.getGuiContext();
		Item item = guiContext.getItem();
		String subContext = Values.getString(guiContext.getSubContext(), "");

		String allColumns = null;
		SourceItemTransform requiredDirection = null;
		if (RelatedItemExtensionColumn.SUBCONTEXT.equals(subContext)) 
		{
			// The current item should be a Specification instance that may specify
			// the column. Otherwise we load the columns from the regular list context
			Specification spec = jspContext.getCurrentItem();
			if (spec != null) 
			{
				allColumns = spec.actionId;
				if (spec.direction == SourceItemTransform.CHILD_TO_PARENT)
				{
					// We may only select CHILD columns from a PARENT
					requiredDirection = SourceItemTransform.PARENT_TO_CHILD;
					subContext = "";
				}
			}
		}
		ItemContext cc = sc.getCurrentContext(item, subContext, null);
		if (allColumns == null)
		{
			if (cc == null) return null;
			allColumns = cc.getSetting("columns");
		}
		if (allColumns == null || !allColumns.contains("/")) return null;
		
		boolean lazy = !Boolean.FALSE.equals(jspContext.getAttribute("lazy-loading"));
		DbControl dc = jspContext.getDbControl();
		RelatedItemHelper helper = new RelatedItemHelper(sc, jspContext, lazy);
		List<ListColumnAction<Annotatable, Object>> actions = new ArrayList<>(); 
		for (String col : allColumns.split(","))
		{
			if (col.startsWith("/"))
			{
				Specification spec = Specification.parse(item, col);
				if (requiredDirection != null && requiredDirection != spec.direction) continue;
				ListColumnAction<Annotatable, Object> action = RelatedItemColumn.create(dc, actions.size(), spec, helper);
				if (action != null) actions.add(action);
			}
		}
		
		if (actions.size() == 0) return null;
		// If we are lazy-loading child/parent items we need to store context attributes
		// so that the lazy-loading feature can pick them up again
		if (lazy) jspContext.storeAttributes(cc, "lazy-related-items-attributes");
		return actions.toArray(new ListColumnAction[actions.size()]);
	}
	
}
