package net.sf.basedb.clients.web.extensions.list;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.clients.web.Base;
import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.formatter.LinkedItemFormatter;
import net.sf.basedb.clients.web.formatter.StringFormatter;
import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Metadata.PropertyPath;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.formatter.CollectionFormatter;
import net.sf.basedb.util.formatter.NameableFormatter;
import net.sf.basedb.util.formatter.ToStringFormatter;
import net.sf.basedb.util.listable.SourceItemTransformer;

/**
	Related item column that get values from parent or child items by specifying
	a "property path". See {@link Metadata}.

	@author nicklas
	@since 3.16, 3.17
*/
public class RelatedItemPropertyColumn 
	extends RelatedItemColumn
{

	
	private static Map<String, String> pathTitles = new HashMap<>();
	static
	{
		// See also add_relateditem_columns.js
		pathTitles.put("externalId", "External ID");
		pathTitles.put("bioWell.bioPlate", "Bioplate");
		pathTitles.put("entryDate", "Registered");
		pathTitles.put("remainingQuantity", "Remaining quantity");
		pathTitles.put("originalQuantity", "Original quantity");
		pathTitles.put("creationEvent.entryDate", "Registration date");
		pathTitles.put("creationEvent.eventDate", "Creation date");
		pathTitles.put("creationEvent.protocol", "Protocol");
		pathTitles.put("creationEvent.hardware", "Hardware");
		pathTitles.put("protocol", "Protocol");
		pathTitles.put("hardware", "Hardware");
		pathTitles.put("software", "Software");
		pathTitles.put("tag", "Tag");
		pathTitles.put("job.itemSubtype", "Job type");
		pathTitles.put("$itemLists", "Item lists");
	}
	
	/**
		Convert some paths to a more user-friendly title instead.
	*/
	private static String pathToTitle(String path)
	{
		String title = pathTitles.get(path);
		return title == null ? path.substring(0, 1).toUpperCase() + path.substring(1) : title;
	}
	
	private final PropertyPath<? super Annotatable, ?> property;
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	RelatedItemPropertyColumn(DbControl dc, int index, Specification spec, RelatedItemHelper helper, PropertyPath<? super Annotatable, ?> property)
	{
		super(dc, index, spec, helper);
		this.property = property;
		if (property == null)
		{
			setTitle(spec.generateTitle(null));
			setTooltip(spec.generateTooltip(null));
		}
		else
		{
			setTitle(spec.generateTitle(pathToTitle(spec.property)));
			setTooltip(spec.generateTooltip(spec.property));
		}
		// If the property ends with an item...
		if (property == null || property.getHibernateType().isEntityType())
		{
			setExportFormatter(new NameableFormatter());
			if (!helper.lazy) setFormatter(new CollectionFormatter(new LinkedItemFormatter()));
			// Specical case for Job subtypes which we set as an enumeration of existing subtypes
			if ("job.itemSubtype".equals(spec.property))
			{
				ItemQuery<ItemSubtype> jobSubtypesQuery = Base.getSubtypesQuery(Item.JOB);
				setEnumeration(Enumeration.fromItems(jobSubtypesQuery.list(dc), null));
				setValueType(Type.INT);
			}
			else
			{
				// Filtering is made on the "name"
				if (getFilterProperty()==null) 
				{
					// Add '.name' to make filtering work
					String nameProperty = getProperty() + ".name";
					if (nameProperty.contains("/..")) nameProperty = nameProperty.replace("/..", "/");
					setFilterProperty(nameProperty);
				}
				setValueType(Type.STRING);
			}
		}
		else
		{
			Type colType = Type.fromHibernateType(property.getHibernateType());
			if (colType != null)
			{
				setValueType(colType);
				setExportFormatter(FormatterFactory.getTypeFormatter(dc.getSessionControl(), colType, false));
			}
			else
			{
				// Last fallback -- if we get here the Type.fromHibernateType() should be update with special cases
				setValueType(Type.STRING);
				setExportFormatter(new ToStringFormatter<>());
			}
			if (!helper.lazy) 
			{
				setFormatter(new CollectionFormatter(colType == null ? new StringFormatter() : FormatterFactory.getTypeFormatter(dc.getSessionControl(), colType, true)));
			}
		}

	}
	
	@Override
	public Object getValue(RelatedItemHelper helper, Annotatable item, SourceItemTransformer preTransform) 
	{
		Set<Annotatable> items = getRelatedItems(item, preTransform);
		if (items.size() == 0) return null;

		List<Object> values = new ArrayList<>(items.size());
		for (Annotatable p : items)
		{
			Object v = property == null ? p : property.getValue(helper.dc, p);
			if (v != null) values.add(v);
		}
		
		return values.size() == 0 ? null : values;
	}

}
