/*
	$Id: BasicItem.java 8171 2023-11-21 09:31:51Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.NoAutoUnlinkAnyToAny;
import net.sf.basedb.core.hibernate.TypeWrapper;

import java.lang.ref.WeakReference;
import java.util.Set;
import java.util.TreeSet;

/**
	This is the root superclass of all item classes. All items must inherit from this
	class. This class provides access to the id, type and version of the item as
	well as permission checking methods.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2023-11-21 10:31:51 +0100 (Tue, 21 Nov 2023) $
*/
public abstract class BasicItem
	implements Identifiable
{
	/**
		The data class from the net.sf.basedb.core.data layer that holds the
		data for this item.
	*/
	private final BasicData data;
	
	/**
		A reference to the current DbControl object. We use a WeakReference since
		we it is always the client applications responsibility to keep control
		over the DbControl object.
	*/
	private WeakReference<DbControl> dc;
	
	/**
		The logged in user's permission to this item. Calculated by
		the {@link #initPermissions(int, int)} method.
	*/
	private int permissions;

	/**
		Create a new instance. Do not call this directly. Use the
		{@link DbControl#newItem(Class, Object[])} or 
		{@link DbControl#getItem(Class, BasicData, Object[])} 
		methods instead.
	*/
	BasicItem(BasicData data)
	{
		this.data = data;
		this.permissions = 127;  // All item permissions
	}

	/*
		From the AccessControlled interface
		-------------------------------------------
	*/
	/**
		Checks if the logged in user has the specified permission on
		this item. The default implementation only checks the role
		keys for permissions. A subclass may not override this method
		since it would make it possible to bypass security checks.
		<p>
		Subclasses that needs to check other keys, such as the
		{@link SharedItem} should override the {@link
		#initPermissions(int,int)} method instead.

		@param permission A value from the {@link Permission} class
		@return <code>TRUE</code> if the user has the permission,
			<code>FALSE</code> otherwise
	*/
	@Override
	public final boolean hasPermission(Permission permission)
	{
		return Permission.hasPermission(permissions, permission);
	}
	
	/**
		Checks if the logged in user has the specified permission on
		this item. If not, a {@link PermissionDeniedException} is thrown.
		@param permission A value from the {@link Permission} class
		@throws PermissionDeniedException If the logged in user doesn't
			have the requested permission
	*/
	@Override
	public final void checkPermission(Permission permission)
		throws PermissionDeniedException
	{
		if (!hasPermission(permission))
		{
			throw new PermissionDeniedException(permission, this.toString());
		}
	}
	@Override
	public final Set<Permission> getPermissions()
	{
		return Permission.fromInt(permissions);
	}
	// -------------------------------------------
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public final int getId()
	{
		return data.getId();
	}
	@Override
	public final int getVersion()
	{
		return data.getVersion();
	}
	// The getType() method must be implemented by each subclass
	// -------------------------------------------

	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		String id = getId() == 0 ? "new" : "id="+getId();
		if (this instanceof Nameable)
		{
			Nameable n = (Nameable)this;
			id += "; name="+n.getName();
		}
		return getType().toString()+"["+id+"]";
	}
	/**
		Check if this item is equal to another item. They are considered 
		to be equal if their BasicData objects are equal.
		@see BasicData#equals(Object)
	*/
	@Override
	public final boolean equals(Object o)
	{
		if (this == o) return true;
		if (o == null || this.getClass() != o.getClass()) return false;
		BasicItem item = (BasicItem)o;
		return data.equals(item.data);
	}
	
	/**
		The hash code is calculated at object construction and remains
		the same during the objects lifetime. For new items the hash code
		is randomly generated. Items loaded from the database calculates
		the hash code from the from the id.
		@see BasicData#hashCode()
	*/
	@Override
	public final int hashCode()
	{
		return data.hashCode();
	}
	// -------------------------------------------
	
	/**
		Checks if the item has been saved to the database or not.
		@return TRUE if the item is saved to the database, FALSE otherwise
	*/
	public final boolean isInDatabase()
	{
		return getId() != 0;
	}

	/**
		Check if this item is attached to a {@link DbControl} object or not.
		An item is detached if there is no <code>DbControl</code>, if it is closed, 
		or if is not managed by the <code>DbControl</code>.

		@return TRUE if the item is detached, FALSE otherwise
		@see DbControl#detachItem(BasicItem)
		@see DbControl#reattachItem(BasicItem, boolean)
		@see DbControl#saveItem(BasicItem)
	*/
	public final boolean isDetached()
	{
		DbControl tmp = dc == null ? null : dc.get();
		return tmp == null || tmp.isClosed() || !tmp.isAttached(this);
	}
	
	/**
		Check if this item is used by some other item. With used we mean that
		another item is linking to this item in way that prevents this item from
		beeing deleted. Ie. if we tried to delete an item that is used, we would
		get a foreign key violation error from the database.
		If the subclass overrides this method it should propagate the call
		to the superclass, ie. <code>super.isUsed()</code>, unless it has decided
		that the item is used by some other means.
		@return TRUE if this item is used, FALSE otherwise
	 	@throws BaseException If not able to tell if item is used or not.
		@see #getUsingItems()
	*/
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
			"COUNT_USED_ANYTOANY_TO", Long.class);
		/*
			SELECT count(*)
			FROM AnyToAnyData ana
			WHERE ana.toId = :toId AND ana.toType = :toType AND ana.usingTo = true
		*/
		query.setParameter("toId", this.getId(), TypeWrapper.H_INTEGER);
		query.setParameter("toType", this.getType().getValue(), TypeWrapper.H_INTEGER);
		return HibernateUtil.loadData(query) > 0;
	}
	
	/**
		Get all items that are using this item. This method must be implemented
		in a way that is consistent with the {@link #isUsed()} method. Ie. 
		a subclass must override both or none of these methods. The method must
		return a set with proxies for all items (even if the logged in user doesn't
		have read permission for all of them). With the proxy it is possible to
		get the ID and type of the item. To get full access to the item the logged
		in user must of course have read permission to it.
		@return A set containing proxies for the items, or an empty set if
			no items are using this item
		@throws BaseException If there is an error loading the items
		@since 2.2
		@see #addUsingItems(Set, Item, org.hibernate.query.Query)
	*/
	public Set<ItemProxy> getUsingItems()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Object[]> query = HibernateUtil.getPredefinedQuery(session, 
			"FIND_USING_ANYTOANY", Object[].class);
		/*
			SELECT ana.fromId, ana.fromType
			FROM AnyToAnyData ana
			WHERE ana.toId = :toId AND ana.toType = :toType AND ana.usingTo = true
		*/
		query.setParameter("toId", this.getId(), TypeWrapper.H_INTEGER);
		query.setParameter("toType", this.getType().getValue(), TypeWrapper.H_INTEGER);
		Set<ItemProxy> using = new TreeSet<ItemProxy>();
		for (Object[] u : HibernateUtil.loadList(query, getSessionControl()))
		{
			using.add(new ItemProxy((Integer)u[0], Item.fromValue((Integer)u[1])));
		}
		return using;
	}

	/**
		Break references to items that may lead to circular references.
		This method is intended to be implemented by subclasses and should
		only break references leading to items that are found in the 
		given set of items. This method is called from the {@link Trashcan}
		which has already confirmed that the given items are about to
		be deleted and may have circular references between them.
		NOTE! This method should not check {@link AnyToAny} links. This has
		already been done in the trashcan code.
		@since 3.8
	*/
	int breakCircularReferences(Set<ItemProxy> items)
	{
		return 0;
	}
	
	private boolean hasBeenCalledFromGetAnnotationSet = false;
	private boolean hasBeenCalledFromAnnotationBatcher = false;
	
	/**
		Annotatable items that are used with the {@link AnnotationBatcher}
		must not call {@link Annotatable#getAnnotationSet()} in the same
		transaction. This method is a helper method for keeping track of
		which method that has been called.
		@param annotationSet TRUE when called from the getAnnotationSet method
		@param batcher TRUE when called from the batcher
		@since 3.8
	*/
	void checkBatchAnnotatableUsage(boolean annotationSet, boolean batcher)
	{
		if (annotationSet)
		{
			if (hasBeenCalledFromAnnotationBatcher)
			{
				throw new IllegalStateException("Can't call getAnnotationSet() after the item has been used with an AnnotationBatcher: " + this);
			}
			hasBeenCalledFromGetAnnotationSet = true;
		}
		if (batcher)
		{
			if (hasBeenCalledFromGetAnnotationSet)
			{
				throw new IllegalStateException("Can't use with AnnotationBatcher after getAnnotationSet() has been called: " + this);
			}
			if (hasBeenCalledFromAnnotationBatcher)
			{
				throw new IllegalStateException("Can't use an item more than once with an AnnotationBatcher: " + this);
			}
			hasBeenCalledFromAnnotationBatcher = true;
		}
	}
	
	/**
		Add items to the set of items using this one.
		@param using The set to add to
		@param itemType The type of items to add
		@param query A Hibernate query that returns the ID of items that are using
			this one
		@since 2.2
	*/
	protected void addUsingItems(Set<ItemProxy> using, Item itemType, org.hibernate.query.Query<Integer> query)
	{
		for (Integer itemId : HibernateUtil.loadList(query, getSessionControl()))
		{
			if (itemId != null) using.add(new ItemProxy(itemId, itemType));
		}
	}
	
	/**
		Add items to the set of items using this one.
		@param using The set to add to
		@param query A Hibernate query that returns BasicData objects
		@since 2.2
	*/
	protected void addUsingItems(Set<ItemProxy> using, org.hibernate.query.Query<? extends BasicData> query)
	{
		for (BasicData item : HibernateUtil.loadList(query, getSessionControl()))
		{
			using.add(new ItemProxy(item.getId(), Item.fromDataObject(item)));
		}
	}

	/**
		Try to load the the item-class representation of all the
		data-class items ignoring permission problems. The loaded
		items are added to the 'parents' set. Typically used
		by {@link Annotatable#getAnnotatableParents()} implementations to
		load all annotatable parent items.
		@since 3.7
	*/
	protected void addAnnotatableParents(DbControl dc, Set<Annotatable> parents, AnnotatableData... items)
	{
		if (items == null || items.length == 0) return;
		for (AnnotatableData data : items)
		{
			if (data != null)
			{
				BasicData d = (BasicData)data;
				Item itemType = Item.fromDataObject(d);
				Class<? extends BasicItem> itemClass = itemType.getItemClass();
				try
				{
					Annotatable parent = (Annotatable)dc.getItem(itemType.getItemClass(), d);
					parents.add(parent);
				}
				catch (PermissionDeniedException ex)
				{}
			}
		}
	}
	
	
	/**
		Get the {@link BasicData} object that holds all data for this item.
	*/
	abstract BasicData getData();
	
	final BasicData getBasicData()
	{
		return data;
	}
	
	/**
		Get the {@link DbControl} object that currently manages this item.
		@throws ConnectionClosedException If the item is connected to a closed 
			<code>DbControl</code> object, or not connected at all
	*/
	public final DbControl getDbControl()
		throws ConnectionClosedException
	{
		DbControl tmp = dc == null ? null : dc.get();
		if (tmp == null || tmp.isClosed()) throw new ConnectionClosedException();
		return tmp;
	}

	/**
		Set the {@link DbControl} object that should manage this item.
		A null value detaches the item
	*/
	final void setDbControl(DbControl dbControl)
	{
		dc = dbControl == null ? null : new WeakReference<DbControl>(dbControl);
	}
	
	/**
		Get the {@link SessionControl} object that manages this item.
		@throws ConnectionClosedException If the item is not connected to a
			<code>DbControl</code> object
	*/
	public final SessionControl getSessionControl()
		throws ConnectionClosedException
	{
		DbControl tmp = dc == null ? null : dc.get();
		if (tmp == null) throw new ConnectionClosedException();
		return tmp.getSessionControl();
	}
	
	/**
		Initialise the logged in user's permissions for this item.
		For items loaded from the database, the default implementation 
		checks the role keys. For new items, write permission is added.
		<p>
		Subclasses that needs to check other keys or properties, such 
		as the {@link OwnedItem} and {@link SharedItem} should override 
		this method. The subclass should calculate additional permissions 
		to be granted or denied, and combine those with whatever was passed
		as parameters. Use the binary OR operator ( | ) to combine
		the permissions. Finally the subclass must call 
		<code>super.initPermissions(granted, denied)</code>.
		
		@param granted Permissions that have been granted by the subclass
		@param denied Permissions that have been denied by the subclass
		@throws BaseException If the permissions couldn't be initialised
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		Item itemType = getType();

		int rolePermissions = getSessionControl().getRolePermissions(itemType);
		if (Permission.hasPermission(rolePermissions, Permission.DENIED))
		{
			granted = 0;
		}
		else
		{
			PluginPermission pp = getPluginPermissions();
			granted |= rolePermissions | pp.granted;
			denied |= pp.denied & ~pp.granted;
		}
		
		if (!isInDatabase())
		{
			granted |= Permission.grant(Permission.WRITE);
		}
		this.permissions = granted & ~denied; // granted AND NOT denied
	}
	
	PluginPermission getPluginPermissions()
	{
		return getSessionControl().getPluginPermissions(getType());
	}
	

	/**
		This method is called on each {@link Validatable} item 
		before is saved to the database. If the subclass
		overrides this method it should also propagate the call
		to the superclass, ie. <code>super.validate()</code>.

		@throws BaseException If there is an error
		@see Validatable
		@see <a href="../../../../../html/developer/api/core_api.datavalidation.html">Developer documentation: Data validation</a>
		@see <a href="../../../../../html/developer/core_ref.rules.itemclass.html">Developer documentation: Coding rules and guidelines for item classes</a>
	*/
	void validate()
		throws InvalidDataException, BaseException
	{}

	/**
		This method is called on each {@link Transactional} item and on
		all items if the action is {@link Transactional.Action#CREATE}
		or {@link Transactional.Action#DELETE} before a commit is issued to the database. 
		If the subclass overrides this method it must also propagate the call
		to the superclass, ie. <code>super.onBeforeCommit(action)</code>.

		@throws BaseException If there is an error
		@see Transactional
		@see <a href="../../../../../html/developer/api/core_api.transactions.html">Developer documentation: Transactions</a>
		@see <a href="../../../../../html/developer/core_ref.rules.itemclass.html">Developer documentation: Coding rules and guidelines for item classes</a>
	*/
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		if (action == Transactional.Action.DELETE)
		{
			// Delete all any-to-any links from this item...
			org.hibernate.Session session = getDbControl().getHibernateSession();
			if (!(this.data instanceof NoAutoUnlinkAnyToAny))
			{
				AnyToAny.unlinkAllFrom(getDbControl(), this);
				
				// ...and all links to this item that doesn't count as using it
				org.hibernate.query.Query<?> query = HibernateUtil.getPredefinedQuery(session, 
					"DELETE_UNUSED_ANYTOANY_TO");
				/*
					DELETE FROM AnyToAnyData ata 
					WHERE ata.toId = :toId AND ata.toType = :toType AND ata.usingTo = false
				*/
				query.setParameter("toId", this.getId(), TypeWrapper.H_INTEGER);
				query.setParameter("toType", this.getType().getValue(), TypeWrapper.H_INTEGER);
				HibernateUtil.executeUpdate(query);
			}
			
			// Make sure ItemList.size is up to date
			if (this instanceof Listable)
			{
				org.hibernate.query.Query<?> query = HibernateUtil.getPredefinedSQLQuery(session, 
					"UPDATE_ITEMLIST_SIZE");
					/*
						UPDATE [ItemLists] 
						SET [size] = [size] + :delta
						WHERE [member_type] = :memberType 
						AND [id] = ANY(
							SELECT [list_id]
							FROM [ItemListMembers] 
							WHERE [item_id] = :itemId
						)
					*/
				query.setParameter("delta", -1, TypeWrapper.H_INTEGER);
				query.setParameter("itemId", getId(), TypeWrapper.H_INTEGER);
				query.setParameter("memberType", getType().getValue(), TypeWrapper.H_INTEGER);
				int numUpdated = HibernateUtil.executeUpdate(query);
				if (numUpdated > 0)
				{
					// Delete rows from ItemListMembers table
					query = HibernateUtil.getPredefinedSQLQuery(session, 
						"DELETE_ITEMLIST_MEMBER");
						/*
							DELETE FROM [ItemListMembers]
							WHERE [item_id] = :itemId
							AND [list_id] = ANY(
								SELECT [id]
								FROM [ItemLists]
								WHERE [member_type] = :memberType
							)
						*/
					query.setParameter("itemId", getId(), TypeWrapper.H_INTEGER);
					query.setParameter("memberType", getType().getValue(), TypeWrapper.H_INTEGER);
					int numMembers = query.executeUpdate();
				}
			}
		}
	}

	/**
		This method is called before inserting new items into the database
		to allow items to propagate default values from the active project.
		This method is only called when a project is active. 
		<p>
		Implementation note! Subclasses should only set default values
		for properties that hasn't been set explicitly by client code,
		also accounting for setAbc(null) calls.
		
		@param activeProject The currently active project
		@since 2.9
	*/
	void setProjectDefaults(Project activeProject)
		throws BaseException
	{}
	
	/**
		This method is called on all items directly after Hibernate has inserted 
		it into the database. This method can be used in place of the 
		{@link #onBeforeCommit(Transactional.Action)} in case the id is needed. The id 
		has not been generated when the <code>onBeforeCommit</code> is called.
		@throws BaseException If there is an error
		@see Transactional
		@see <a href="../../../../../html/developer/api/core_api.transactions.html">Developer documentation: Transactions</a>
		@see <a href="../../../../../html/developer/core_ref.rules.itemclass.html">Developer documentation: Coding rules and guidelines for item classes</a>
	*/
	void onAfterInsert()
		throws BaseException
	{}
	
	/**
		This method is called on each {@link Transactional} object 
		after a successful commit to the database. If the subclass
		overrides this method it should also propagate the call
		to the superclass, ie. <code>super.onAfterCommit(action)</code>. It is forbidden
		to access the <code>DbControl</code> object from this method and
		it must not throw any exceptions. All exceptions should be logged
		using the {@link Application#getLogger()} object.

		@see Transactional
		@see <a href="../../../../../html/developer/api/core_api.transactions.html">Developer documentation: Transactions</a>
		@see <a href="../../../../../html/developer/core_ref.rules.itemclass.html">Developer documentation: Coding rules and guidelines for item classes</a>
	*/
	void onAfterCommit(Transactional.Action action)
	{}

	/**
		This method is called on each {@link Transactional} object 
		after an unsuccessful commit to the database. If the subclass
		overrides this method it should also propagate the call
		to the superclass, ie. <code>super.onRollback(action)</code>. It is forbidden
		to access the <code>DbControl</code> object from this method and
		it must not throw any exceptions. All exceptions should be logged
		using the {@link Application#getLogger()} object.

		@see Transactional
		@see <a href="../../../../../html/developer/api/core_api.transactions.html">Developer documentation: Transactions</a>
		@see <a href="../../../../../html/developer/core_ref.rules.itemclass.html">Developer documentation: Coding rules and guidelines for item classes</a>
	*/
	void onRollback(Transactional.Action action)
	{}

}
