/*
	$Id: DynamicRawDataQuery.java 7433 2017-12-08 13:18:12Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.query.AutoJoiner;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.ReporterListExpression;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.util.jep.JepFunction;
import net.sf.basedb.util.jep.RawFunction;

/**
	Represents a query for raw data in a "dynamic" way. The main difference 
	from a client applications point of view, is that it is possible to 
	specify which columns that should be selected. In a techinical point of
	view the difference is that we are using SQL instead of HQl.
	<p>
	This type of query is needed to bridge the gap between the dynamic
	and regular parts of the database. For example, a query of this type
	can be used by a batcher to insert the original intensities in a raw
	bioassayset.
	
	@author Nicklas
	@version 2.0
	@see RawBioAssay#getDynamicRawData()
	@see net.sf.basedb.util.IntensityCalculatorUtil
	@base.modified $Date: 2017-12-08 14:18:12 +0100 (Fri, 08 Dec 2017) $
*/
public class DynamicRawDataQuery
	extends AbstractSqlQuery
{

	/**
		Create a new dynamic query for raw data.		
		@param raw The raw bioassay to query on
		
		@since 3.9
	*/
	public static DynamicRawDataQuery getQuery(RawBioAssay raw)
	{
		return raw.getDynamicRawData();
	}
	
	/**
		Create a new dynamic query for raw data.		
		@param raw The raw bioassays to query on which must all be
			of the same raw data type
		@since 3.9
	*/
	public static DynamicRawDataQuery getQuery(Collection<RawBioAssay> raw)
	{
		RawDataType rdt = null;
		Set<Integer> ids = new HashSet<Integer>();
		for (RawBioAssay r : raw)
		{
			ids.add(r.getId());
			if (rdt == null) 
			{
				rdt = r.getRawDataType();
			}
			else if (!rdt.equals(r.getRawDataType()))
			{
				throw new InvalidDataException("Raw bioassay '" + r.getName() + " doesn't use the expected data type: " + rdt.getName());
			}
		}
		DynamicRawDataQuery query = new DynamicRawDataQuery(rdt);
		query.restrictPermanent(Restrictions.in(Dynamic.rawData("rawBioAssay"), Expressions.parameter("rawBioAssays")));
		query.setPermanentParameter("rawBioAssays", ids, Type.INT);
		return query;
	}
	
	private final RawDataType rawDataType;
	private final RealTable rootTable;
	private final Set<Object> joinedItems;
	private RawBioAssay rawBioAssay;

	/**
		Create a new dynamic query for raw data.
		NOTE! This constructor doesn't add any
		restriction to the query. This must be 
		done by the caller.
		@param rawBioAssay The raw bioassay to query on
	*/
	DynamicRawDataQuery(RawBioAssay rawBioAssay)
	{
		this(rawBioAssay.getRawDataType());
		this.rawBioAssay = rawBioAssay;
	}
	
	/**
		Create a new dynamic query for raw data.
		@param rawDataType The raw data type
		@since 3.1
	*/
	DynamicRawDataQuery(RawDataType rawDataType)
	{
		super(rawDataType.getRealTable().getQualifiedTableName());
		if (!rawDataType.isStoredInDb())
		{
			throw new BaseException("Raw data for raw data type '" + rawDataType + "' is not stored in the database.");
		}
		this.rawDataType = rawDataType;
		this.rootTable = rawDataType.getRealTable();
		joinedItems = new HashSet<Object>();
		setAutoJoinType(JoinType.LEFT);
	}
	
	/*
		From the Query interface
		-------------------------------------------
	*/
	/**
		The alias of the item that is returned by this query.
		@see Item#getAlias()
	*/
	@Override
	public String getRootAlias()
	{
		return HibernateUtil.quote(rootTable.getAlias());
	}
	@Override
	public boolean isReadonly()
	{
		return false;
	}
	// -------------------------------------------

	/*
		From the AbstractSqlQuery class
		-------------------------------------------
	*/
	/**
		No default selection.
	*/
	@Override
	Select[] getDefaultSelects()
	{
		return null;
	}
	
	@Override
	public JepFunction[] getJepFunctions(DbControl dc, boolean restrictions) 
	{
		List<JepFunction> functions = new ArrayList<JepFunction>();
		functions.add(new RawFunction(true));
		return functions.toArray(new JepFunction[functions.size()]);
	}
	// -------------------------------------------


	public RawDataType getRawDataType()
	{
		return rawDataType;
	}
	
	RealTable getRootTable()
	{
		return rootTable;
	}
	
	public RawBioAssay getRawBioAssay()
	{
		return rawBioAssay;
	}
	
	/**
		Specify the join type of automatic joins. The default join type is
		{@link JoinType#LEFT}. This setting doesn't affect joins that has been made
		directly to the query.
		@param joinType The type of the automatic joins
		@since 3.9
	*/
	public void setAutoJoinType(JoinType joinType)
	{
		addAutoJoiner(new ReporterJoiner(joinType));
		addAutoJoiner(new RawBioAssayJoiner(joinType));
	}

	/**
		Join information about reporters in the query.
		@param joinType The type of join. Can not be null.
	 	@throws InvalidDataException If joinType parameter is null.
		@see Dynamic#selectReporter(String)
		@see Dynamic#reporter(String)
		@since 3.1
	*/
	public void joinReporters(JoinType joinType)
	{
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (!joinedItems.contains(RealTable.REPORTERS))
		{
			join(new RealJoin(joinType, getRootTable(), RealTable.REPORTERS, "reporter_id"));
			joinedItems.add(RealTable.REPORTERS);
		}
	}

	
	/**
	  	Makes a join with a <code>ReporterList</code>
	  	@param reporterList To do the join on.
	  	@param joinType The join type to use. 
		@see net.sf.basedb.core.query.JoinType
		@since 3.9
	*/
	public void joinReporterList(ReporterList reporterList, JoinType joinType)
	{
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (reporterList == null) throw new InvalidUseOfNullException("reporterList");
		if (!joinedItems.contains(reporterList))
		{
			join(new ReporterListJoin(reporterList, joinType, getRootTable(), "reporter_id"));
			joinedItems.add(reporterList);
		}
	}

	/**
		Join information about raw bioassays in the query.
		@param joinType The type of join. Can not be null.
	 	@throws InvalidDataException If joinType parameter is null.
		@since 3.9
	*/
	public void joinRawBioAssay(JoinType joinType)
	{
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (!joinedItems.contains(RealTable.RAWBIOASSAYS))
		{
			join(new RealJoin(joinType, getRootTable(), RealTable.RAWBIOASSAYS, "rawbioassay_id"));
			joinedItems.add(RealTable.RAWBIOASSAYS);
		}
	}
	
	/**
		Auto joiner for reporters.
	*/
	static class ReporterJoiner
		implements AutoJoiner<DynamicRawDataQuery, ReporterPropertyExpression>
	{
		private final JoinType joinType;
	
		private ReporterJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<DynamicRawDataQuery> getQueryClass()
		{
			return DynamicRawDataQuery.class;
		}
		@Override
		public Class<ReporterPropertyExpression> getElementClass()
		{
			return ReporterPropertyExpression.class;
		}
		@Override
		public void joinIfNeeded(DynamicRawDataQuery query, ReporterPropertyExpression element)
		{
			query.joinReporters(joinType);
		}
		@Override
		public void reset(DynamicRawDataQuery query)
		{}
		// -------------------------------------------
	}
	
	/**
		Auto joiner for raw bioassays.
	*/
	static class RawBioAssayJoiner
		implements AutoJoiner<DynamicRawDataQuery, RawBioAssayPropertyExpression>
	{
		private final JoinType joinType;
	
		private RawBioAssayJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<DynamicRawDataQuery> getQueryClass()
		{
			return DynamicRawDataQuery.class;
		}
		@Override
		public Class<RawBioAssayPropertyExpression> getElementClass()
		{
			return RawBioAssayPropertyExpression.class;
		}
		@Override
		public void joinIfNeeded(DynamicRawDataQuery query, RawBioAssayPropertyExpression element)
		{
			query.joinRawBioAssay(joinType);
		}
		@Override
		public void reset(DynamicRawDataQuery query)
		{}
		// -------------------------------------------
	}
	
	/**
		Auto joiner for reporter lists.
	*/
	static class ReporterListJoiner
		implements AutoJoiner<DynamicRawDataQuery, ReporterListExpression>
	{
		private final JoinType joinType;
	
		private ReporterListJoiner(JoinType joinType)
		{
			this.joinType = joinType;
		}
		
		/*
			From the AutoJoiner interface
			-------------------------------------------
		*/
		@Override
		public Class<DynamicRawDataQuery> getQueryClass()
		{
			return DynamicRawDataQuery.class;
		}
		@Override
		public Class<ReporterListExpression> getElementClass()
		{
			return ReporterListExpression.class;
		}
		@Override
		public void joinIfNeeded(DynamicRawDataQuery query, ReporterListExpression element)
		{
			JoinType jt = element.getAutoJoinType();
			query.joinReporterList(element.getReporterList(), jt == null ? joinType : jt);
		}
		@Override
		public void reset(DynamicRawDataQuery query)
		{}
		// -------------------------------------------
	}

}
