/*
	$Id: DynamicSpotQuery.java 6465 2014-06-04 12:41:12Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.util.jep.ChannelFunction;
import net.sf.basedb.util.jep.ExtraValueFunction;
import net.sf.basedb.util.jep.InReporterListFunction;
import net.sf.basedb.util.jep.JepFunction;
import net.sf.basedb.util.jep.NotInReporterListFunction;
import net.sf.basedb.util.jep.PositionFunction;
import net.sf.basedb.util.jep.RawChannelFunction;
import net.sf.basedb.util.jep.RawFunction;
import net.sf.basedb.util.jep.ReporterFunction;
import net.sf.basedb.util.jep.ScoreFunction;

/**
	Represents a query for spot information in the dynamic database. A single
	query object may query against data in one bioassayset at a time.
	Use factory methods in the {@link Dynamic} class to join other tables and 
	create restrictions based on the data in those tables. You may of course
	also use the regular expressions and restrictions from the {@link
	net.sf.basedb.core.query.Expressions}, {@link net.sf.basedb.core.query.Restrictions} 
	and other classes.
	<p>
	This type of query is executed by the {@link #iterate(DbControl)} method
	and returns the result as a {@link DynamicResultIterator}.
	
	@author Nicklas
	@version 2.0
	@see BioAssaySet#getSpotData()
	@base.modified $Date: 2014-06-04 14:41:12 +0200 (Wed, 04 Jun 2014) $
*/
public class DynamicSpotQuery
	extends DynamicQuery
{

	private final BioAssaySet bioAssaySet;
	private boolean disableRawJoinOnMergedData;
	private Select[] defaultSelects;
	
	private Set<Object> joinedItems;
	
	/**
		Create a new dynamic query.
		@param bioAssaySet The bioassay set to query against
	*/
	DynamicSpotQuery(BioAssaySet bioAssaySet)
	{
		super(bioAssaySet.getDataCube(), VirtualTable.SPOT, bioAssaySet.getIntensityTransform());
		this.bioAssaySet = bioAssaySet;
		joinedItems = new HashSet<Object>();
		setAutoJoinType(JoinType.LEFT);
	}
	
	/*
		From the AbstractSqlQuery class
		-------------------------------------------
	*/
	/**
		Default selection: column, postion, ch1, ch2, ...
	*/
	@Override
	Select[] getDefaultSelects()
	{
		if (defaultSelects == null)
		{
			int channels = getDataCube().getRawDataType().getChannels();
			defaultSelects = new Select[2+channels];
			defaultSelects[0] = Dynamic.select(VirtualColumn.COLUMN);
			defaultSelects[1] = Dynamic.select(VirtualColumn.POSITION);
			for (int i = 1; i <= channels; ++i)
			{
				defaultSelects[i+1] = Dynamic.select(VirtualColumn.channelIntensity(i));
			}
		}
		return defaultSelects;
	}
	// -------------------------------------------
	/*
		From the AbstractSqlQuery class
		-------------------------------------------
	*/

	/**
	 	Reset join information.
	*/
	@Override
	public void reset()
	{
		super.reset();
		joinedItems.clear();
	}
	// -------------------------------------------

	/*
		From the DynamicQuery class
		-------------------------------------------
	*/
	/**
		The following JEP functions can be used:
		<ul>
		<li>ch(n): {@link ChannelFunction}
		<li>rawCh(n): {@link RawChannelFunction}
		<li>pos(): {@link PositionFunction}
		<li>rep(property): {@link ReporterFunction}
		<li>score(reporter-list-id): {@link ScoreFunction}
		<li>xtra(extra-value-id): {@link ExtraValueFunction} (if a bioassay set
			has been specified)
		<li>raw(property): {@link RawFunction}
		<li>inList(reporter-list-id): {@link InReporterListFunction} (only if restrictions=true)
		<li>notInList(reporter-list-id): {@link NotInReporterListFunction} (only if restrictions=true)
		</ul>
		@since 2.15
	*/
	@Override
	public JepFunction[] getJepFunctions(DbControl dc, boolean restrictions)
	{
		List<JepFunction> functions = new ArrayList<JepFunction>();
		functions.add(new ChannelFunction());
		functions.add(new RawChannelFunction());
		functions.add(new PositionFunction());
		functions.add(new ReporterFunction());
		functions.add(new ScoreFunction(dc));
		functions.add(new ExtraValueFunction(dc, bioAssaySet));
		functions.add(new RawFunction(!hasDisabledRawJoin()));
		if (restrictions)
		{
			functions.add(new InReporterListFunction(dc));
			functions.add(new NotInReporterListFunction(dc));
		}
		return functions.toArray(new JepFunction[functions.size()]);
	}
	// -------------------------------------------

	/**
		Disable automatic and manual joining of raw data if the bioassay set
		contains merged data. Eg. if the {@link BioAssaySet#getMaxRawMappingsForSpot()}
		returns a value != 1. Note! Joins that have already been applied before this 
		method is called are not disabled. 
		@since 2.15
	*/
	public void setDisableRawJoinOnMergedData(boolean disable)
	{
		this.disableRawJoinOnMergedData = disable;
	}
	
	/**
		@return TRUE if joining raw data has been disabled for merged data
	*/
	public boolean getDisableRawJoinOnMergedData()
	{
		return disableRawJoinOnMergedData;
	}
	
	/**
		Check if joining to raw data has been disabled. It is disabled only
		if {@link BioAssaySet#getMaxRawMappingsForSpot()} != 1 and 
		{@link #getDisableRawJoinOnMergedData()} == true
	 	@return TRUE if raw data join has been disabled, FALSE if not
	*/
	public boolean hasDisabledRawJoin()
	{
		return disableRawJoinOnMergedData && bioAssaySet.getMaxRawMappingsForSpot() != 1;
	}
	
	/**
		Join information about reporters in the query. Now it becomes possible
		to select reporter data or restrict the results based on reporter
		information.
		<p>
		NOTE! Since BASE 3.1 this method automatically join to the cloned reporter
		table if available. Use {@link #joinReporters(JoinType, boolean)}, or
		{@link #setUseClonedReporters(boolean)} with a FALSE value to force using 
		the master reporters table.
		
		@param joinType The type of join. Can not be null.
	 	@throws InvalidDataException If joinType parameter is null.
		@see Dynamic#selectReporter(String)
		@see Dynamic#reporter(String)
	*/
	public void joinReporters(JoinType joinType)
		throws InvalidDataException
	{
		joinReporters(joinType, useClonedReporters());
	}
	
	/**
		Join information about reporters in the query. If the useCloned
		parameter is TRUE cloned reporter annotations are used if available,
		otherwise the master reporter annotations are used.

		@param joinType The type of join. Can not be null.
	 	@throws InvalidDataException If joinType parameter is null.
		@see Dynamic#selectReporter(String)
		@see Dynamic#reporter(String)
		@since 3.1
	*/
	public void joinReporters(JoinType joinType, boolean useCloned)
	{
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (!joinedItems.contains(VirtualTable.POSITION))
		{
			join(new DynamicJoin(getVirtualDb(), joinType, null, VirtualTable.POSITION, 
				VirtualColumn.CUBE, VirtualColumn.POSITION));
			joinedItems.add(VirtualTable.POSITION);
		}
		if (useCloned && hasClonedReporters())
		{
			if (!joinedItems.contains(VirtualTable.CLONED_REPORTERS))
			{
				VirtualColumn[] posColumns = new VirtualColumn[] { VirtualColumn.REPORTER_ID };
				VirtualColumn[] cloneColumns = new VirtualColumn[] { VirtualColumn.REPORTER_ID_PK };
				join(new DynamicJoin(getVirtualDb(), joinType, 
						VirtualTable.POSITION, posColumns, 
						VirtualTable.CLONED_REPORTERS, cloneColumns));
				joinedItems.add(VirtualTable.CLONED_REPORTERS);
			}
		}
		else
		{
			if (!joinedItems.contains(RealTable.REPORTERS))
			{
				join(new RealJoin(joinType, VirtualTable.POSITION, RealTable.REPORTERS, 
					VirtualColumn.REPORTER_ID));
				joinedItems.add(RealTable.REPORTERS);
			}
		}
	}
	
	/**
	  	Makes a join with a <code>ReporterList</code>
	  	@param reporterList To do the join on.
	  	@param joinType The join type to use. 
	  		@see net.sf.basedb.core.query.JoinType
	*/
	public void joinReporterList(ReporterList reporterList, JoinType joinType)
	{
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (reporterList == null) throw new InvalidUseOfNullException("reporterList");
		if (!joinedItems.contains(VirtualTable.POSITION))
		{
			join(new DynamicJoin(getVirtualDb(), joinType, null, VirtualTable.POSITION, 
				VirtualColumn.CUBE, VirtualColumn.POSITION));
			joinedItems.add(VirtualTable.POSITION);
		}
		if (!joinedItems.contains(reporterList))
		{
			join(new ReporterListJoin(reporterList, joinType, VirtualTable.POSITION, VirtualColumn.REPORTER_ID));
			joinedItems.add(reporterList);
		}
	}
	
	
	/**
		Join information about raw data in the query. Now it becomes possible
		to select raw data or restrict the results based on values in the raw
		data.
		@param joinType The type of join. Can not be null.
	 	@throws InvalidDataException If joinType parameter is null. 
		@see Dynamic#selectRawData(String)
		@see Dynamic#rawData(String)
	*/
	public void joinRawData(JoinType joinType)
		throws InvalidDataException
	{
		if (hasDisabledRawJoin())
		{
			throw new PermissionDeniedException("Not allowed to join raw data with merged spot data");
		}
		RawDataType rdt = getDataCube().getRawDataType();
		if (!rdt.isStoredInDb())
		{
			throw new BaseException("Raw data for raw data type: " + rdt + " is not stored in the database.");
		}
		
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (!joinedItems.contains(VirtualTable.RAWPARENTS))
		{
			join(new DynamicJoin(getVirtualDb(), joinType, null, VirtualTable.RAWPARENTS, 
				VirtualColumn.CUBE, VirtualColumn.COLUMN, VirtualColumn.POSITION));
			joinedItems.add(VirtualTable.RAWPARENTS);
		}
		if (!joinedItems.contains(rdt))
		{
			join(new RealJoin(joinType, VirtualTable.RAWPARENTS, rdt.getRealTable(),
				VirtualColumn.RAWDATA_ID));
			joinedItems.add(rdt);
		}
	}
	
	/**
		Join extra values in the query. Now it becomes possible to select
		those extra values or use them for restrictions.
		@param extraValue The extra values to join. Can not be null.
		@param joinType The type of join. Can not be null.
	 	@throws InvalidDataException If any of the required parameters are null. 
		@see Dynamic#select(ExtraValue, String)
		@see Dynamic#extraValue(ExtraValue)
	*/
	public void joinExtraValue(ExtraValue extraValue, JoinType joinType)
		throws InvalidDataException
	{
		if (extraValue == null) throw new InvalidUseOfNullException("extraValue");
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (!getDataCube().equals(extraValue.getDataCubeExtraValue().getDataCube()))
		{
			throw new InvalidDataException("Extra value "+extraValue+
				" is not part of the same data cube as this query: " + getDataCube());
		}
		if (!joinedItems.contains(extraValue))
		{
			join(new ExtraValueJoin(extraValue, joinType));
			joinedItems.add(extraValue);
		}
	}
	
	/**
		Specify the join type of automatic joins. The default join type is
		{@link JoinType#LEFT}. This setting doesn't affect joins that has been made
		directly to the query.
		@param joinType The type of the automatic joins
	*/
	public void setAutoJoinType(JoinType joinType)
	{
		addAutoJoiner(DynamicSpotQueryJoiners.getExtraValueJoiner(joinType));
		addAutoJoiner(DynamicSpotQueryJoiners.getReporterJoiner(joinType));
		if (!hasDisabledRawJoin())
		{
			addAutoJoiner(DynamicSpotQueryJoiners.getRawDataJoiner(joinType));
		}
		addAutoJoiner(DynamicSpotQueryJoiners.getReporterListJoiner(joinType));
	}
}
