/*
	$Id: Extract.java 7587 2019-02-06 09:41:56Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.data.ExtractData;
import net.sf.basedb.core.hibernate.TypeWrapper;

import java.util.Set;

/**
	This class is used to represent extract items. An extract is
	created from a {@link Sample} or by combining other extracts
	(pooling). It is also possible to create
	standalone extracts, with no reference to a sample or to other
	extracts.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2019-02-06 10:41:56 +0100 (Wed, 06 Feb 2019) $
*/
public class Extract
	extends MeasuredBioMaterial
{

	/**
		The type of item represented by this class.
		@see Item#EXTRACT
		@see #getType()
	*/
	public static final Item TYPE = Item.EXTRACT;
	
	/**
		The id for the <code>ExtractType</code> item representing a
		labeled extract.
	 	@since 3.0
	*/
	public static final String LABELED = "net.sf.basedb.core.ExtractType.LABELED";

	/**
		The id for the <code>ExtractType</code> item representing a library.
	 	@since 3.0
	*/
	public static final String LIBRARY = "net.sf.basedb.core.ExtractType.LIBRARY";

	/**
		Create a new <code>Extract</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>Extract</code> item
		@throws BaseException If there is an error
	*/
	public static Extract getNew(DbControl dc)
		throws BaseException
	{
		return getNew(dc, null);
	}

	/**
		Create a new <code>Extract</code> item as part
		of a master bioplate event.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param master The master bioplate event participant or null
		@return The new <code>Extract</code> item
		@throws BaseException If there is an error
		@since 2.17
	*/
	public static Extract getNew(DbControl dc, BioPlateEventParticipant master)
		throws BaseException
	{
		Extract e = dc.newItem(Extract.class);
		e.setName("New extract");
		e.setBioPlateEventParticipant(master);
		return e;
	}

	
	/**
		Get an <code>Extract</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Extract</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static Extract getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Extract e = dc.loadItem(Extract.class, id);
		if (e == null) throw new ItemNotFoundException("Extract[id="+id+"]");
		return e;
	}

	/**
		Get a query that returns extracts.
		@return An {@link ItemQuery} object.
	*/
	public static ItemQuery<Extract> getQuery()
	{
		return new ItemQuery<Extract>(Extract.class);
	}
	
	
	Extract(ExtractData extractData)
	{
		super(extractData);
	}

	@Override
	ExtractData getData()
	{
		return (ExtractData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Subtypable interface
		-----------------------------
	*/
	@Override
	@SupportsPushAnnotations
	@SubtypableRelatedItems({Item.PROTOCOL, Item.KIT, Item.HARDWARE, Item.TAG, Item.SAMPLE, Item.EXTRACT})
	public ItemSubtype getItemSubtype()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getItemSubtype());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that:
		<ul>
		<li>no {@link PhysicalBioAssay}:s has been created from this item
		<li>no {@link Extract}:s has been created from this item
		<li>no {@link DerivedBioAssay} is referencing this item
		<li>no {@link RawBioAssay} is referencing this item
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		boolean used = false;
		
		if (!used)
		{
			org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
				"GET_DERIVEDBIOASSAYS_FOR_EXTRACT", Long.class, "count(*)");
			/*
				SELECT {1}
				FROM DerivedBioAssayData dba
				WHERE dba.extract = :extract
			*/
			query.setParameter("extract", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		
		if (!used)
		{
			org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
				"GET_RAWBIOASSAYS_FOR_EXTRACT", Long.class, "count(*)");
			/*
				SELECT {1}
				FROM RawBioAssayData rba
				WHERE rba.parentExtract = :extract
			*/
			query.setParameter("extract", getId(), TypeWrapper.H_INTEGER);
			used = HibernateUtil.loadData(query) > 0;
		}
		
		return used || super.isUsed();
	}
	
	/**
		Get all:
		<ul>
		<li>Child extracts created from the extract
		<li>{@link PhysicalBioAssay}:s created from this extract
		<li>{@link DerivedBioAssay}:s referencing the extract
		<li>{@link RawBioAssay}:s referencing the extract
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Physical bioassays
		org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SOURCEEVENTS_FOR_BIOMATERIAL", Integer.class, "bme.physicalBioAssay.id");
			/*
				SELECT {1}
				FROM BioMaterialEventData bme
				JOIN bme.sources src
				WHERE src.bioMaterial = :bioMaterial
			*/
		query.setParameter("bioMaterial", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.PHYSICALBIOASSAY, query);
		
		// Derived bioassays
		query = HibernateUtil.getPredefinedQuery(session, 
				"GET_DERIVEDBIOASSAYS_FOR_EXTRACT", Integer.class, "dba.id");
				/*
					SELECT {1}
					FROM DerivedBioAssayData dba
					WHERE dba.extract = :extract
				*/
		query.setParameter("extract", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.DERIVEDBIOASSAY, query);
		
		// Raw bioassays
		query = HibernateUtil.getPredefinedQuery(session, 
				"GET_RAWBIOASSAYS_FOR_EXTRACT", Integer.class, "rba.id");
				/*
					SELECT {1}
					FROM RawBioAssayData rba
					WHERE rba.parentExtract = :extract
				*/
		query.setParameter("extract", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.RAWBIOASSAY, query);
		
		return using;
	}
	// -------------------------------------------

	/**
		Get the tag that has been applied to this extract in order
		to be able to distinguish it from other extracts on a 
		physical bioassay.
		@return A tag or null if the extract is not tagged
		@since 3.0
	*/
	public Tag getTag()
	{
		return getDbControl().getItem(Tag.class, getData().getTag());
	}
	
	/**
		Set the tag that was applied to this extract in order
		to be able to distinguish it from other extracts on a 
		physical bioassay.
		@param tag The tag, or null if the extract is not tagged
		@since 3.0
	*/
	public void setTag(Tag tag)
	{
		checkPermission(Permission.WRITE);
		if (tag != null) tag.checkPermission(Permission.USE);
		getData().setTag(tag == null ? null : tag.getData());
	}
	
	/**
		Creates a query that returns all child (pooled) extracts that
		has this extracts as one of it's sources.
		@return A query
		@since 2.10
	*/
	public ItemQuery<Extract> getChildExtracts()
	{
		ItemQuery<Extract> query = Extract.getQuery();
		query.joinPermanent(Hql.innerJoin("creationEvent", Item.BIOMATERIALEVENT.getAlias()));
		query.joinPermanent(Hql.innerJoin(Item.BIOMATERIALEVENT.getAlias(), "sources", "src"));
		query.restrictPermanent(Restrictions.eq(Hql.index("src", null), Hql.entity(this)));
		return query;
	}
	
	/**
		Counts all extracts that have been created from this extract. The count
		includes extracts that the logged in user doesn't have access to.
		@return The number of extracts
	 	@throws BaseException If there is some kind of error.
	 	@since 2.16
	*/
	public long countExtracts()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SOURCEEVENTS_FOR_BIOMATERIAL", Long.class, "count(*)");
		/*
			SELECT {1}
			FROM BioMaterialEventData bme
			JOIN bme.sources src
			WHERE src.bioMaterial = :bioMaterial
		*/
		query.setParameter("bioMaterial", getId(), TypeWrapper.H_INTEGER);
		return HibernateUtil.loadData(query);
	}
	
	/**
		Count the number of child extracts. Since an extract
		can only have other extracts as children, this method always return 
		0 if extractsOnly = false.
		@since 2.16
	*/
	@Override
	public long countChildren(boolean extractsOnly)
	{
		return extractsOnly ? countExtracts() : 0;
	}

	/**
		Get a query that returns all bioassays created from this 
		extract.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<PhysicalBioAssay> getPhysicalBioAssays()
	{
		ItemQuery<PhysicalBioAssay> query = PhysicalBioAssay.getQuery();
		query.joinPermanent(Hql.innerJoin("creationEvent.sources", Item.EXTRACT.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.index(Item.EXTRACT.getAlias(), null),
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Get a query that returns all derived bioassays created from this 
		extract.
		@return An {@link ItemQuery} object
		@since 3.14.1
	*/
	public ItemQuery<DerivedBioAssay> getDerivedBioAssays()
	{
		ItemQuery<DerivedBioAssay> query = DerivedBioAssay.getQuery();
		query.joinPermanent(Hql.innerJoin("extract", Item.EXTRACT.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.EXTRACT.getAlias()),
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Get a query that returns all raw bioassays created from this 
		extract.
		@return An {@link ItemQuery} object
		@since 3.14.1
	*/
	public ItemQuery<RawBioAssay> getRawBioAssays()
	{
		ItemQuery<RawBioAssay> query = RawBioAssay.getQuery();
		query.joinPermanent(Hql.innerJoin("parentExtract", Item.EXTRACT.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.EXTRACT.getAlias()),
				Hql.entity(this)
			)
		);
		return query;
	}
	
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		Get the tag and parent extracts.
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		Set<Annotatable> annotatable = super.getAnnotatableParents();
		ExtractData data = getData();
		addAnnotatableParents(getDbControl(), annotatable,
			data.getTag()
		);
		return annotatable;
	}
}

