package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.snapshot.SnapshotManager;


/**
	Implements an API for simple and fast loading of annotation values optimized for
	situations when a single annotation is needed for many (=thousands) of items.
	The regular annotation API is typically very slow and requires multiple SQL queries
	for each annotation. The {@link SnapshotManager} API is typically faster, but only if
	the annotations have already been cached.
	
	This API have limited filtering capabilities and can be used for reading only, but it
	can be combined with the {@link AnnotationBatcher} API for updating annotations.
	
	The result is returned as an iterator with each entry corresponding
	to a single annotation value. The values are not sorted in any particular
	order. If the annotation type allows multiple values each value is returned
	as separate iterator entries in no particular order.
		
	If the annotation type supports project-specific values, only default values
	and values for the current project are returned.

	@since 3.19.10
	@param <T> The type of values for the annotation
	@see AnnotationType#getFastLoader(DbControl)
*/
public class FastAnnotationLoader<T>
{
	
	private final DbControl dc;
	private final AnnotationType type;
	
	private int projectId;
	private boolean includeDefault;
	
	FastAnnotationLoader(DbControl dc, AnnotationType type) 
	{
		this.dc = dc;
		this.type = type;
		this.projectId = dc.getSessionControl().getActiveProjectId();
		this.includeDefault = true;
	}
	
	/**
		If the annotation type supports project-specific annotations this method can be
		used to control which values that are loaded. The default setting is to load 
		default values and values for the currently active project.
		@param p The project that values should belong to, or null to not load any project-specific values
	*/
	public void setProject(Project p)
	{
		this.projectId =  p == null ? -1 : p.getId();
	}
	
	/**
		If the annotation type supports project-specific annotations this method can be
		used to control which values that are loaded. The default setting is to load 
		default values and values for the currently active project.
		@param includeDefault TRUE to include default values, FALSE to not include them
	*/
	public void setIncludeDefault(boolean includeDefault)
	{
		this.includeDefault = includeDefault;
	}
	
	/**
		Load the annotation values for all items returned by the query. This method is
		typically faster than the {@link #load(Collection)} method since the
		{@link ItemQuery#idList(DbControl)} is used to only load item ids and not
		full items.
	*/
	public Iterator<FastAnnotation<T>> load(ItemQuery<? extends Annotatable> query)
	{
		return new FastIterator<>(dc, type, query.getItemType(), projectId, includeDefault, query.idList(dc));
	}
	
	/**
		Load the annotation values for all items in the given collection. 
		Note! All items in the collection must be of the same type.
	*/
	public Iterator<FastAnnotation<T>> load(Collection<? extends Annotatable> items)
	{
		if (items.isEmpty()) return Collections.emptyIterator();
		List<Integer> itemIds = new ArrayList<Integer>(items.size());
		Item itemType = null;
		for (Annotatable item : items)
		{
			itemIds.add(item.getId());
			if (itemType == null) itemType = item.getType();
		}
		return new FastIterator<>(dc, type, itemType, projectId, includeDefault, itemIds);
	}
	
	/**
		Load the annotation values for all items returned by the given iterator.
		Note! All items returned by the iterator must be of the same type.
	*/
	public Iterator<FastAnnotation<T>> load(Iterator<? extends Annotatable> items)
	{
		List<Integer> itemIds = new ArrayList<Integer>();
		Item itemType = null;
		while (items.hasNext())
		{
			Annotatable item = items.next();
			itemIds.add(item.getId());
			if (itemType == null) itemType = item.getType();
		}
		return new FastIterator<>(dc, type, itemType, projectId, includeDefault, itemIds);
	}

	/**
		Iterator implementation that splits the query into multiple calls depending
		on the max number of parameters that can be used in a query.
	*/
	static class FastIterator<T>
		implements Iterator<FastAnnotation<T>>
	{
		private final DbControl dc;
		private final String sql;
		private final List<Integer> itemIds;
		private final int maxParameters;
		
		private Iterator<Object[]> it;
		private int startIndex;
		private int endIndex;
		private FastAnnotation<T> next;
		
		FastIterator(DbControl dc, AnnotationType annotationType, Item itemType, int projectId, boolean includeDefault, List<Integer> itemIds)
		{
			this.dc = dc;
			this.itemIds = itemIds;
			this.sql = PredefinedQuery.getQueryString("FAST_ANNOTATION_LOADER", 
				annotationType.getValueType().getTableName(),
				Integer.toString(annotationType.getId()),
				Integer.toString(annotationType.getProjectAnnotations() ? projectId : 0),
				Integer.toString(annotationType.getProjectAnnotations() ? (includeDefault ? 0 : -1) : 0),
				Integer.toString(itemType.getValue()));
			this.maxParameters = HibernateUtil.getDbEngine().getMaxParametersInQuery();
			this.it = Collections.emptyIterator(); // This will make sure that hasNext() calls nextIterator() at least once
		}
		
		/**
			Create a new query that load annotations for items between startIndex and endIndex.
			@return TRUE if an iterator was created, FALSE if not
		*/
		private boolean nextIterator()
		{
			startIndex = endIndex;
			endIndex = Math.min(startIndex + maxParameters, itemIds.size());
			if (startIndex < itemIds.size())
			{
				org.hibernate.query.Query<Object[]>	loadValues = HibernateUtil.createSqlQuery(dc.getHibernateSession(), sql, Object[].class);
				loadValues.setParameterList("listOfIds", itemIds.subList(startIndex, endIndex), TypeWrapper.H_INTEGER);
				it = HibernateUtil.loadIterator(loadValues, true, dc.getSessionControl());
			}
			else
			{
				it = null;
			}
			return it != null;
		}
		
		@Override
		public boolean hasNext() 
		{
			if (next != null) return true;
			if (it == null) return false;
			
			while (!it.hasNext())
			{
				if (!nextIterator()) return false;
			}
			
			next = new FastAnnotation<T>(it.next());
			return true;
		}

		@Override
		public FastAnnotation<T> next() 
		{
			FastAnnotation<T> v = next;
			next = null;
			return v;
		}
	}
	
	/**
		Holds information about a single annotation value.
	*/
	public static class FastAnnotation<T>
	{
		/** 
			The value.
		*/
		public final T value;
		
		/**
			The ID of the item the annotation belongs to.
		*/
		public final int itemId;
		
		/**
			The ID of the annotation. If the annotation type support multiple values, they
			all have the same value here.
		*/
		public final int annotationId;
		
		/**
			The ID of the project if the value is a project-specific annotation, 
			or 0 if this is a default annotation.
		*/
		public final int projectId;
		
		/**
			The ID of the default annotation this value is overriding. This value is 0 for
			default annotations and for project-specific annotations where no default value
			exists.
		*/
		public final int overrideId;

		/**
			The ID of the unit that has been selected for this annotation. Note that
			the 'value' is always in the unit specified by the annotation type.
		*/
		public final Integer unitId;
		
		@SuppressWarnings("unchecked")
		FastAnnotation(Object[] data)
		{
			this.value = (T)data[0];
			this.itemId = (Integer)data[1];
			this.annotationId = (Integer)data[2];
			this.projectId = (Integer)data[3];
			this.overrideId = (Integer)data[4];
			this.unitId = (Integer)data[5];
		}
	}
}