/**
	$Id: JobAgent.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.data.JobAgentData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.jobagent.JobAgentConnection;
import net.sf.basedb.util.jobagent.JobAgentInfo;

/**
	This class represents a job agent, which is a program running
	in another virtual machine on the same or a different server. 
	The job agent is regularly checking the database for jobs that
	are waiting to be executed. It is not neccessary for a job agent to
	be registered in the database but if it is the core can provide services
	that are otherwise inaccesible. 
	<ul>
	<li>Keep track of which plugins are installed on the agent's server
	<li>Provide information to other clients about what the agent is
		doing, ie. memory/cpu usage, which plugins that are executing, etc.
	<li>Priority boosting for plugins that should bypass the queing system.
	</ul>

	@author nicklas
	@version 2.0
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (Mon, 22 May 2017) $
*/
public class JobAgent
	extends CommonItem
{
	
	/**
		Log events.
	*/
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(JobAgent.class);
	
	/**
		The type of item represented by this class.
		@see Item#JOBAGENT
		@see #getType()
	*/
	public static final Item TYPE = Item.JOBAGENT;

	/**
		The default port (47822) where job agents are listening for
		connections. If 47822 is converted to a hexadecimal number it
		becomes BACE!
	*/
	public static final int DEFAULT_PORT = 0xBACE;
	
	/**
		Create a new <code>JobAgent</code> item.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param externalId The external ID, which must be unique, for the job agent
		@return The new <code>JobAgent</code> item
		@throws BaseException If there is an error
	*/
	public static JobAgent getNew(DbControl dc, String externalId) 
		throws BaseException
	{
		JobAgent jag = dc.newItem(JobAgent.class);
		jag.setName("New job agent");
		jag.setExternalId(externalId);
		return jag;
	}
	
	/**
		Get a <code>JobAgent</code> object when you know the ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>JobAgent</code> item
		@throws ItemNotFoundException If an item with the specified ID is
			not found
		@throws PermissionDeniedException If the logged in user doesn't
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static JobAgent getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		JobAgent jag = dc.loadItem(JobAgent.class, id);
		if (jag == null) throw new ItemNotFoundException("JobAgent[id=" + id + "]");
		return jag;
	}
	
	/**
		Get a <code>JobAgent</code> item when you know the external ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param externalId The external ID of the item to load
		@return The <code>JobAgent</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the item
		@throws BaseException If there is another error
	*/
	public static JobAgent getByExternalId(DbControl dc, String externalId)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (externalId != null) externalId = externalId.trim();
		org.hibernate.query.Query<JobAgentData> query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
			"GET_JOBAGENT_FOR_EXTERNALID", JobAgentData.class);
		/*
			SELECT jag
			FROM JobAgentData jag
			WHERE jag.externalId = :externalId
		*/
		query.setParameter("externalId", externalId, TypeWrapper.H_STRING);
		JobAgent jag = dc.getItem(JobAgent.class, HibernateUtil.loadData(query));
		if (jag == null) throw new ItemNotFoundException("JobAgent[externalId="+externalId+"]");
		return jag;
	}
	
	/**
		Get an {@link ItemQuery} that returns job agents.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<JobAgent> getQuery()
	{
		return new ItemQuery<JobAgent>(JobAgent.class);
	}
	
	JobAgent(JobAgentData data)
	{
		super(data);
	}

	@Override
	JobAgentData getData()
	{
		return (JobAgentData)getBasicData();
	}
	/*
	 From the Identifiable interface
	 -------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/**
		The maximum length of the external ID that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = JobAgentData.MAX_EXTERNAL_ID_LENGTH;
	/**
		Get the external id for the job agent. The external id is used to
		identify each job agent and must be unique.
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	/**
		Set the external id for the job agent.
	 	@param externalId External id to be set. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_EXTERNAL_ID_LENGTH}
		@throws BaseException If there is another error
	*/
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNullableString(externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}
	
	/**
		The maximum length of the server name that can be stored in the database.
		@see #setServer(String)
	*/
	public static final int MAX_SERVER_LENGTH = JobAgentData.MAX_SERVER_LENGTH;
	/**
		Get the name or IP-adress of the server were the job agent is running.
		@return The name or IP, or null if not known
	*/
	public String getServer()
	{
		return getData().getServer();
	}
	/**
		Set the name/IP of the server where the job agent is running.
		@param server The name/IP or null if not known
		@throws InvalidDataException If the new value is longer than {@link 
			#MAX_SERVER_LENGTH}
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
	*/
	public void setServer(String server)
		throws InvalidDataException, PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setServer(StringUtil.setNullableString(server, "server", MAX_SERVER_LENGTH));
	}
	
	/**
		Get the port the job agent is listening on.
		@return The port number or null if the job agent isn't listening
	*/
	public Integer getPort()
	{
		return getData().getPort();
	}
	/**
		Set the port the job agent is listening on.
		@param port The new port number or null if the job agent isn't listening
	*/
	public void setPort(Integer port)
	{
		checkPermission(Permission.WRITE);
		getData().setPort(port);
	}
	
	/**
		Get info about the running job agent. For performance reasons the
		information is cached and only updated if older than 
		specified by the <code>agent.maxage</code> configuration setting.
		@return A <code>JobAgentInfo</code> object
	*/
	public JobAgentInfo getInfo()
	{
		return getAgentInfo(this);
	}
	
	/**
		Get a connection object for connecting to the job agent. It is only
		possible to connect to job agents which has specified a server and port.
		@param timeout The timeout for the connection, null will use the default
			timeout as configued by the <code>agent.connection.timeout</code> setting
			in the <code>base.config</code> file
		@return The connection object or null if none could be created
		@throws PermissionDeniedException If the logged in user doesn't have
			WRITE permission on the job agent
	*/
	public JobAgentConnection getConnection(Integer timeout)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		String server = getServer();
		Integer port = getPort();
		if (timeout == null)
		{
			timeout = Config.getInt("agent.connection.timeout", 1000);
		}
		JobAgentConnection conn = null;
		if (port != null && server != null)
		{
			conn = new JobAgentConnection(server, port, timeout);
			agentInfo.remove(getId());
		}
		return conn;
	}
	
	/**
		Get the current CPU usage in percent of the server where the job agent is running. 
		The CPU usage is not available unless a server name/IP and port has been specified. 
		For performance reasons the CPU usage is cached and only updated if older than 
		specified by the <code>agent.maxage</code> configuration setting.
		@return The CPU usage of the job agent's server in percent, or null if not known
	*/
	public Integer getCpuUsage()
	{
		return getAgentInfo(this).getCpuUsage();
	}
	/**
		Get the current memory usage in bytes of the server where the job agent 
		is running. The memory usage is not available unless a server name/IP and port 
		has been specified. For performance reasons the memory usage is cached and only 
		updated if older than specified by the <code>agent.maxage</code> configuration setting.
		@return The memory usage on the job agent's server, or null if not known
	*/
	public Long getUsedMemory()
	{
		return getAgentInfo(this).getUsedMemory();
	}
	
	/**
		Get the total memory in bytes of the server where the job agent 
		is running. The memory is not available unless a server name/IP and port 
		has been specified. For performance reasons the memory is cached and only 
		updated if older than specified by the <code>agent.maxage</code> configuration 
		setting.
		@return The memory on the job agent's server, or null if not known
	*/
	public Long getTotalMemory()
	{
		return getAgentInfo(this).getTotalMemory();
	}
	
	/**
		Get the settings for the specified plugin. If no settings exists a new 
		<code>JobAgentSettings</code> item is created if the <code>create</code>
		parameter is true. You must call {@link DbControl#saveItem(BasicItem)}
		to save the new settings to the database.
		
		@param plugin The plugin to the settings for, or null if none exists 
			and the <code>create</code> parameter is false
	 	@param create TRUE if new settings should be created if no one was found, FALSE otherwise. 
		@return A <code>JobAgentSettings</code> object for the specified plugin
	*/
	public JobAgentSettings getSettings(PluginDefinition plugin, boolean create)
	{
		if (plugin == null) throw new InvalidUseOfNullException("plugin");
		DbControl dc = getDbControl();
		JobAgentSettings settings = dc.getItem(JobAgentSettings.class, 
			getData().getPlugins().get(plugin.getData()));
		
		if (settings == null && create)
		{
			settings = JobAgentSettings.getNew(dc, this, plugin);
		}
		return settings;
	}
	
	/**
		Get the settings for the plugin that is used for the specified job. 
		@param job The job to get the settings for
		@return A <code>JobAgentSettings</code> object or null if not
			settings exist for the job
	*/
	public JobAgentSettings getSettings(Job job)
	{
		DbControl dc = getDbControl();
		JobAgentSettings settings = dc.getItem(JobAgentSettings.class, 
			getData().getPlugins().get(job.getData().getPluginDefinition()));
		return settings;
	}
	
	/**
		Get a query that returns all plugin-specific settings for this job agent.
		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<JobAgentSettings> getSettings()
	{
		return JobAgentSettings.getQuery(this);
	}

	/**
		Get a query that returns all jobs that are waiting to be executed sorted
		by their priority taking the priority boost into account.

		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<Job> getWaitingJobs()
	{
		ItemQuery<Job> query = Job.getQuery();
		
		/*
			The final query will look something like:
			SELECT job 
			FROM JobData job
			INNER JOIN job.pluginDefinition pld
			INNER JOIN pld.jobAgentSettings jas
			WHERE jas.jobAgent = this
			AND (job.owner IN (:sharedToUsers) OR job.activeProjectId IN (:sharedToProjects)
			AND job.status = Job.Status.WAITING
			AND job.jobAgentId IS NULL OR job.jobAgentId = this.id
			ORDER BY job.priority - jas.priorityBoost ASC, job.scheduled ASC
		*/
		
		// Join the plugin-specific settings
		query.joinPermanent(Hql.innerJoin("pluginDefinition", Item.PLUGINDEFINITION.getAlias()));
		query.joinPermanent(Hql.innerJoin(Item.PLUGINDEFINITION.getAlias(), 
			"jobAgentSettings", Item.JOBAGENTSETTINGS.getAlias()));
		
		// Only load jobs with plugins for which this job agent has settings
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property(Item.JOBAGENTSETTINGS.getAlias(), "jobAgent"),
				Hql.entity(this)
			)
		);
		
		// Create restriction for the owner of the job
		query.include(Include.MINE, Include.OTHERS);
		
		// Load all users that this job agent is shared to
		Restriction userRestriction = null;
		Set<Integer> userIds = new HashSet<Integer>();
		userIds.add(getData().getOwner().getId());
		boolean everyone = false;
		ItemKey itemKey = getItemKey();
		if (itemKey != null)
		{
			Set<Integer> groupIds = itemKey.getGroupIds(Permission.USE);
			if (groupIds != null && groupIds.contains(SystemItems.getId(Group.EVERYONE)))
			{
				everyone = true;
			}
			else
			{
				userIds.addAll(itemKey.getUserIds(Permission.USE));
				groupIds = Group.getGroupsRecursive(getDbControl(), groupIds);
				userIds.addAll(User.getAllMembers(getDbControl(), groupIds));
			}
		}
		if (userIds.size() > 0)
		{
			userRestriction = Restrictions.in(
					Hql.property("owner"),
					Expressions.parameter("users")
				);
		}
		
		// Load all projects this job agent is shared to
		Restriction projectRestriction = null;
		Set<Integer> projectIds = new HashSet<Integer>();
		if (!everyone)
		{
			ProjectKey projectKey = getProjectKey();
			if (projectKey != null)
			{
				// Filter on projects that have USE permission for this agent
				projectIds.addAll(projectKey.getProjectIds(Permission.USE));
			}
		}
		if (projectIds.size() > 0)
		{
			projectRestriction = Restrictions.in(
					Hql.property("activeProjectId"),
					Expressions.parameter("projects")
				);
		}
		
		// Add user and project restrictions to the query
		if (!everyone)
		{
			if (userIds.size() > 0) query.setPermanentParameter("users", userIds, Type.INT);
			if (projectIds.size() > 0) query.setPermanentParameter("projects", projectIds, Type.INT);
			if (userRestriction != null && projectRestriction != null)
			{
				query.restrictPermanent(Restrictions.or(userRestriction, projectRestriction));
			}
			else if (userRestriction != null)
			{
				query.restrictPermanent(userRestriction);
			}
			else if (projectRestriction != null)
			{
				query.restrictPermanent(projectRestriction);
			}
			else
			{
				// No one is allowed to use this agent (should never happen since it must have an owner)
				query.restrictPermanent(Restrictions.eq(Expressions.integer(0), Expressions.integer(1)));
			}
		}

		// Filter out jobs that are not waiting
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("status"), 
				Expressions.integer(Job.Status.WAITING.getValue())
			)
		);
		
		// Filter out jobs that have been marked for other job agents
		// jobAgentId = this.id OR jobAgentId IS NULL
		query.restrictPermanent(
			Restrictions.or(
				Restrictions.eq(
					Hql.property("jobAgentId"),
					Expressions.integer(getId())
				),
				Restrictions.eq(Hql.property("jobAgentId"), null)
			)
		);
		
		// Order by priority corrected by the priority boost
		query.order(
			Orders.asc(
				Expressions.subtract(
					Hql.property("priority"),
					Hql.property(Item.JOBAGENTSETTINGS.getAlias(), "priorityBoost")
				)
			)
		);
		
		// Order by the waiting time
		query.order(Orders.asc(Hql.property("scheduled")));
		
		return query;
	}
	
	
	private static final Map<Integer, JobAgentInfo> agentInfo = new HashMap<Integer, JobAgentInfo>();
	
	/**
		Get a <code>JobAgentInfo</code> for a job agent. If the job agent hasn't specified
		a server and port or if there is an error while connecting to the job agent
		an empty info object is returned.
		
		@param agent The job agent to get the info for
		@return A <code>JobAgentInfo</code> object
	*/
	private static JobAgentInfo getAgentInfo(JobAgent agent)
	{
		JobAgentInfo info = agentInfo.get(agent.getId());
		if (info == null || info.getAge() > Config.getInt("agent.maxage", 60) * 1000)
		{
			String server = agent.getServer();
			Integer port = agent.getPort();
			if (port != null && server != null)
			{
				try
				{
					JobAgentConnection conn = new JobAgentConnection(server, port, 
						Config.getInt("agent.connection.timeout", 1000));
					info = conn.getInfo(true);
				}
				catch (Throwable t)
				{
					log.warn("Could not get info for job agent on server " + server + ":" + port, t);
					info = new JobAgentInfo();
				}
			}
			else
			{
				info = new JobAgentInfo();
			}
			agentInfo.put(agent.getId(), info);
		}
		return info;
	}
	
}

