/*
	$Id: MappingBatcher.java 6881 2015-04-21 07:59:42Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.hibernate.ExecuteUpdateWork;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Selects;

import java.util.List;
import java.util.LinkedList;
import java.sql.PreparedStatement;
import java.sql.SQLException;

/**
	Batcher class for inserting parent mapping information in the
	dynamic database. This batcher is used to map data cube spots to 
	their parent {@link RawData}.
	<p>
	Code example: <a href="../../../../../development/overview/dynamic/rootbioassayset.html">Dynamic API - Creating a root bioassayset</a>

	@author Nicklas
	@version 2.0
	@see BioAssaySet#getMappingBatcher()
	@base.modified $Date: 2015-04-21 09:59:42 +0200 (Tue, 21 Apr 2015) $
*/
public class MappingBatcher
	extends AbstractBatcher
{

	/**
		The datacube this batcher inserts data for.
	*/
	private final DataCube dataCube;
	
	/**
		If the data cube has already received it's data and
		no raw data can be linked.
	*/
	private final boolean noRawMapping;

	/**
		The number of queued inserts. Used to check when a flush
		is needed.
	*/
	private int queuedInsertCount = 0;
	
	/**
		The total number of inserts done by this batcher.
	*/
	private int totalInsertCount = 0;
	
	/**
		The number of bytes this batcher has added to the experiment.
	*/
	private long bytes = 0;

	/**
		The number of bytes used by a single row.
	*/
	private final long bytesPerRow;

	/**
		The statement that inserts the spot/raw data link.
<pre class="code">
INSERT INTO Dynamic#RawParents (cube, column, position, rawdata_id)
VALUES (cube, ?, ?, ?)
</pre>
	*/
	private PreparedStatement rawMappingSql = null;

	/**
	 	The SQL string for the {@link #rawMappingSql} statement.
	*/
	private String rawMappingSqlStatement;

	/**
		Create a new position batcher for a data cube. It is assumed that
		the calling code has checked that the cube is allowed
		to receive more data (ie. it hasn't been committed to the database).
	*/
	MappingBatcher(DbControl dc, DataCube dataCube)
	{
		super();
		this.dataCube = dataCube;
		this.noRawMapping = dataCube.isInDatabase();
		this.bytesPerRow = VirtualTable.RAWPARENTS.getBytesPerRow(dataCube.getVirtualDb());
		setDbControl(dc);
	}

	/*
		From the Batcher interface
		-------------------------------------------
	*/
	@Override
	public void flush()
		throws BaseException
	{
		try
		{
			if (rawMappingSql != null) rawMappingSql.executeBatch();
		}
		catch (SQLException ex)
		{
			logSql.error(rawMappingSqlStatement, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
		queuedInsertCount = 0;
	}
	// -------------------------------------------
	/*
		From the AbstractBatcher class
		-------------------------------------------
	*/
	/**
		Update bytes on data cube and close open SQL statements.
	*/
	@Override
	void onBeforeClose()
		throws BaseException
	{
		dataCube.addBytes(bytes);
		try
		{
			if (rawMappingSql != null) 
			{
				rawMappingSql.close();
				rawMappingSql = null;
			}
		}
		catch (SQLException ex)
		{
			logSql.error(ex.getMessage(), ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
		}
		if (totalInsertCount > 0 && !getDbControl().isRolledBack())
		{
			String catalog = Application.getDynamicCatalog();
			String schema = Application.getDynamicSchema();
			String table = VirtualTable.RAWPARENTS.getTableName(dataCube.getVirtualDb());
			analyzeTable(catalog, schema, table);
		}
	}
	// -------------------------------------------

	/**
		Link a spot to a raw data. One spot may be linked to several
		raw data items.
		@param column The column coordinate of the spot
		@param position The position coordinate of the spot
		@param rawData The raw data object
		@throws InvalidDataException If the raw data is null
		@throws BaseException If there is an error
	*/
	public void insert(short column, int position, RawData rawData)
		throws BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		updateLastAccess();
		if (noRawMapping)
		{
			throw new PermissionDeniedException("Data has already been committed to the database.");
		}
		if (rawData == null)
		{
			throw new InvalidUseOfNullException("rawData");
		}
		try
		{
			if (rawMappingSql == null) buildRawMappingSql();
			rawMappingSql.setShort(1, column);
			rawMappingSql.setInt(2, position);
			rawMappingSql.setInt(3, rawData.getId());
			rawMappingSql.addBatch();
			bytes += bytesPerRow;
			totalInsertCount++;
		}
		catch (SQLException ex)
		{
			logSql.error(rawMappingSqlStatement, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
		if (++queuedInsertCount == getBatchSize())
		{
			flush();
		}
	}
	

	/**
		Insert mapping to raw spots using a query. The query MUST select data in
		the following order: column, position, rawDataId
		<p>
		Internally this method converts the query into a 
		<code>INSERT INTO ... SELECT ...</code> query, making it very fast,
		since the data for each position doesn't have to be retreived, processed and
		then inserted one row at a time. It is possible to call
		this method multiple times with the same or different query. 
		
		@param query The query that selects the data to be inserted
		@return int The number of rows inserted
		@throws BaseException If there is an error
	*/
	public int insert(final AbstractSqlQuery query)
		throws BaseException
	{
		if (isClosed()) throw new ConnectionClosedException();
		updateLastAccess();
		if (query == null) throw new InvalidUseOfNullException("query");
		
		query.resetTemporary();
		// Add static values cubeNo to the selection
		query.temporarySelect(
			Selects.expression(
				Expressions.integer(dataCube.getCubeNo()), 
				VirtualColumn.CUBE.getName()
			)
		);
		
		String selectSql = query.getMainQuery(getDbControl(), true);
		
		// Holds the names of the parameters in the query in the order they appear
		final List<String> parameterOrder = new LinkedList<String>();
		
		// Parse out named parameters
		selectSql = query.parseParameters(selectSql, parameterOrder);
		String insertSql = buildInsertSelectSql(selectSql);

		int rowsInserted = 0;
		try
		{
			rowsInserted = HibernateUtil.doReturningWork(getDbControl().getHibernateSession(), 
				new ExecuteUpdateWork(getSessionControl(), insertSql)
				{
					@Override
					public void setParameters(PreparedStatement ps)
						throws SQLException
					{
						query.setParameters(ps, parameterOrder, 1);
					}
				}
			);
			bytes += rowsInserted * bytesPerRow;
			totalInsertCount += rowsInserted;
		}
		catch (SQLException ex)
		{
			logSql.error(insertSql, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
		return rowsInserted;
	}

	/**
		Get the total number of mappings to raw data create by this mapper.
		@since 2.2
	*/
	public int getTotalInsertCount()
	{
		return totalInsertCount;
	}
	
	/**
		Builds the insert SQL statement. Also creates the Dynamic#RawParents table
		if it doesn't exists.
		@see BatchUtil#buildInsertSql(VirtualDb, VirtualTable, VirtualColumn[], Object[])
	*/
	private void buildRawMappingSql()
		throws SQLException, BaseException
	{
		VirtualDb vdb = dataCube.getVirtualDb();
		if (!vdb.hasTable(VirtualTable.RAWPARENTS))
		{
			vdb.createTables(VirtualTable.RAWPARENTS);
		}
		short cube = dataCube.getCubeNo();
		rawMappingSqlStatement = BatchUtil.buildInsertSql(vdb,
			VirtualTable.RAWPARENTS,
			VirtualTable.RAWPARENTS.getColumns(vdb), 
			new Object[] {cube}
		);
		rawMappingSql = HibernateUtil.getConnection(getDbControl().getHibernateSession()).prepareStatement(rawMappingSqlStatement);
	}
	
	/**
		Build the INSERT INTO ... SELECT ... statment. Also creates the 
		Dynamic#RawParents table if it doesn't exists.
		@param selectSql The SELECT part of the query
		@see BatchUtil#buildInsertSelectSql(VirtualDb, VirtualTable, VirtualColumn[], String)
	*/
	private String buildInsertSelectSql(String selectSql)
		throws BaseException
	{
		VirtualDb vdb = dataCube.getVirtualDb();
		if (!vdb.hasTable(VirtualTable.RAWPARENTS))
		{
			vdb.createTables(VirtualTable.RAWPARENTS);
		}
		String sql = BatchUtil.buildInsertSelectSql(vdb, VirtualTable.RAWPARENTS, VirtualTable.RAWPARENTS.getColumns(vdb), selectSql);
		return sql;
	}

}
