/**
	$Id: SessionWrapper.java 7730 2019-06-17 06:40:32Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.io.Serializable;
import java.sql.Connection;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityGraph;
import javax.persistence.EntityManagerFactory;
import javax.persistence.FlushModeType;
import javax.persistence.LockModeType;
import javax.persistence.StoredProcedureQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaDelete;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.CriteriaUpdate;
import javax.persistence.metamodel.Metamodel;

import org.hibernate.CacheMode;
import org.hibernate.Criteria;
import org.hibernate.Filter;
import org.hibernate.FlushMode;
import org.hibernate.HibernateException;
import org.hibernate.IdentifierLoadAccess;
import org.hibernate.LobHelper;
import org.hibernate.LockMode;
import org.hibernate.LockOptions;
import org.hibernate.MultiIdentifierLoadAccess;
import org.hibernate.NaturalIdLoadAccess;
import org.hibernate.ReplicationMode;
import org.hibernate.Session;
import org.hibernate.SessionEventListener;
import org.hibernate.SessionFactory;
import org.hibernate.SharedSessionBuilder;
import org.hibernate.SimpleNaturalIdLoadAccess;
import org.hibernate.Transaction;
import org.hibernate.TypeHelper;
import org.hibernate.UnknownProfileException;
import org.hibernate.graph.RootGraph;
import org.hibernate.jdbc.ReturningWork;
import org.hibernate.jdbc.Work;
import org.hibernate.procedure.ProcedureCall;
import org.hibernate.query.Query;
import org.hibernate.query.NativeQuery;
import org.hibernate.stat.SessionStatistics;

import net.sf.basedb.core.HibernateUtil;

/**
	Wrapper around Hibernate sessions that includes a query cache.
	This wrapper forwards all calls to the underlying real Hibernate 
	session.
	
	@author Nicklas
	@since 2.16
	@base.modified $Date: 2019-06-17 08:40:32 +0200 (Mon, 17 Jun 2019) $
*/
@SuppressWarnings("unchecked")
public class SessionWrapper
	extends AbstractQueryCache
	implements Session
{

	private static final long serialVersionUID = 8075255339774194513L;

	private final Session session;
	
	/**
		Creates a new session wrapper
		@param session The real Hibernate session
	*/
	public SessionWrapper(Session session)
	{
		this.session = session;
	}
	
	/**
		Get the underlying parent session from Hibernate.
	*/
	public Session getParentSession()
	{
		return session;
	}
	
	@Override
	public Transaction beginTransaction()
		throws HibernateException
	{
		return session.beginTransaction();
	}

	@Override
	public LockRequest buildLockRequest(LockOptions options)
	{
		return session.buildLockRequest(options);
	}

	@Override
	public void cancelQuery() 
		throws HibernateException
	{
		session.cancelQuery();
	}

	@Override
	public void clear()
	{
		session.clear();
	}

	@Override
	public void close() 
		throws HibernateException
	{
		session.close();
	}


	@Override
	public boolean contains(Object object)
	{
		return session.contains(object);
	}

	@SuppressWarnings("rawtypes")
	@Override
	@Deprecated
	public Criteria createCriteria(Class persistentClass, String alias)
	{
		return session.createCriteria(persistentClass, alias);
	}

	@SuppressWarnings("rawtypes")
	@Override
	@Deprecated
	public Criteria createCriteria(Class persistentClass)
	{
		return session.createCriteria(persistentClass);
	}

	@Override
	@Deprecated
	public Criteria createCriteria(String entityName, String alias)
	{
		return session.createCriteria(entityName, alias);
	}

	@Override
	@Deprecated
	public Criteria createCriteria(String entityName)
	{
		return session.createCriteria(entityName);
	}

	@Override
	@Deprecated
	public org.hibernate.Query<?> createFilter(Object collection, String queryString)
		throws HibernateException
	{
		return session.createFilter(collection, queryString);
	}

	/**
		If the same query has been executed before, return the cached
		instance, otherwise create a new query from the parent session
		and cache it.
	*/
	@Override
	public Query<?> createQuery(String queryString)
		throws HibernateException
	{
		Query<?> q = getCachedQuery(queryString);
		if (q == null)
		{
			q = session.createQuery(queryString);
			q.setFetchSize(HibernateUtil.getJdbcFetchSize());
			cacheQuery(queryString, q);
		}
		return q;
	}

	/**
		If the same query has been executed before, return the cached
		instance, otherwise create a new query from the parent session
		and cache it.
	*/
	@SuppressWarnings("deprecation")
	@Override
	public NativeQuery<?> createSQLQuery(String queryString) 
		throws HibernateException
	{
		Query<?> q = getCachedQuery(queryString);
		if (q == null || !(q instanceof NativeQuery))
		{
			q = session.createSQLQuery(queryString);
			q.setFetchSize(HibernateUtil.getJdbcFetchSize());
			cacheQuery(queryString, q);
		}
		return (NativeQuery<?>)q;
	}

	@Override
	public void delete(Object object) 
		throws HibernateException
	{
		session.delete(object);
	}

	@Override
	public void delete(String entityName, Object object)
		throws HibernateException
	{
		session.delete(entityName, object);
	}

	@Override
	public void disableFetchProfile(String name) 
		throws UnknownProfileException
	{
		session.disableFetchProfile(name);
	}

	@Override
	public void disableFilter(String filterName)
	{
		session.disableFilter(filterName);
	}

	@Override
	public Connection disconnect() 
		throws HibernateException
	{
		return session.disconnect();
	}

	@Override
	public void doWork(Work work) 
		throws HibernateException
	{
		session.doWork(work);
	}

	@Override
	public void enableFetchProfile(String name) 
		throws UnknownProfileException
	{
		session.enableFetchProfile(name);
	}

	@Override
	public Filter enableFilter(String filterName)
	{
		return session.enableFilter(filterName);
	}

	@Override
	public void evict(Object object) 
		throws HibernateException
	{
		session.evict(object);
	}

	@Override
	public void flush() 
		throws HibernateException
	{
		session.flush();
	}

	@Override
	public <T> T get(Class<T> clazz, Serializable id, LockMode lockMode)
		throws HibernateException
	{
		return session.get(clazz, id, lockMode);
	}

	@Override
	public <T> T get(Class<T> clazz, Serializable id, LockOptions lockOptions)
		throws HibernateException
	{
		return session.get(clazz, id, lockOptions);
	}

	@Override
	public <T> T get(Class<T> clazz, Serializable id) 
		throws HibernateException
	{
		return session.get(clazz, id);
	}

	@Override
	public Object get(String entityName, Serializable id, LockMode lockMode)
		throws HibernateException
	{
		return session.get(entityName, id, lockMode);
	}

	@Override
	public Object get(String entityName, Serializable id, LockOptions lockOptions)
		throws HibernateException
	{
		return session.get(entityName, id, lockOptions);
	}

	@Override
	public Object get(String entityName, Serializable id)
		throws HibernateException
	{
		return session.get(entityName, id);
	}

	@Override
	public CacheMode getCacheMode()
	{
		return session.getCacheMode();
	}

	@Override
	public LockMode getCurrentLockMode(Object object) 
		throws HibernateException
	{
		return session.getCurrentLockMode(object);
	}

	@Override
	public Filter getEnabledFilter(String filterName)
	{
		return session.getEnabledFilter(filterName);
	}

	@Override
	public String getEntityName(Object object) 
		throws HibernateException
	{
		return session.getEntityName(object);
	}

	@Override
	public FlushModeType getFlushMode()
	{
		return session.getFlushMode();
	}

	@Override
	public Serializable getIdentifier(Object object) 
		throws HibernateException
	{
		return session.getIdentifier(object);
	}
	
	@Override
	public LobHelper getLobHelper()
	{
		return session.getLobHelper();
	}

	/**
		If the same query has been executed before, return the cached
		instance, otherwise create a new query from the parent session
		and cache it.
	*/
	@Override
	public Query<?> getNamedQuery(String queryName) 
		throws HibernateException
	{
		Query<?> q = getCachedQuery(queryName);
		if (q == null)
		{
			q = session.getNamedQuery(queryName);
			cacheQuery(queryName, q);
		}
		return q;
	}

	@Override
	public SessionFactory getSessionFactory()
	{
		return session.getSessionFactory();
	}

	@Override
	public SessionStatistics getStatistics()
	{
		return session.getStatistics();
	}

	@Override
	public Transaction getTransaction()
	{
		return session.getTransaction();
	}

	@Override
	public TypeHelper getTypeHelper()
	{
		return session.getTypeHelper();
	}
	
	@Override
	public boolean isConnected()
	{
		return session.isConnected();
	}

	@Override
	public boolean isDefaultReadOnly()
	{
		return session.isDefaultReadOnly();
	}

	@Override
	public boolean isDirty() 
		throws HibernateException
	{
		return session.isDirty();
	}

	@Override
	public boolean isFetchProfileEnabled(String name)
		throws UnknownProfileException
	{
		return session.isFetchProfileEnabled(name);
	}

	@Override
	public boolean isOpen()
	{
		return session.isOpen();
	}

	@Override
	public boolean isReadOnly(Object entityOrProxy)
	{
		return session.isReadOnly(entityOrProxy);
	}

	@Override
	public <T> T load(Class<T> theClass, Serializable id, LockMode lockMode)
		throws HibernateException
	{
		return session.load(theClass, id, lockMode);
	}

	@Override
	public <T> T load(Class<T> theClass, Serializable id, LockOptions lockOptions)
		throws HibernateException
	{
		return session.load(theClass, id, lockOptions);
	}

	@Override
	public <T> T load(Class<T> theClass, Serializable id)
		throws HibernateException
	{
		return session.load(theClass, id);
	}

	@Override
	public void load(Object object, Serializable id) 
		throws HibernateException
	{
		session.load(object, id);
	}

	@Override
	public Object load(String entityName, Serializable id, LockMode lockMode)
		throws HibernateException
	{
		return session.load(entityName, id, lockMode);
	}

	@Override
	public Object load(String entityName, Serializable id, LockOptions lockOptions) 
		throws HibernateException
	{
		return session.load(entityName, id, lockOptions);
	}

	@Override
	public Object load(String entityName, Serializable id)
		throws HibernateException
	{
		return session.load(entityName, id);
	}

	@Override
	public void lock(Object object, LockMode lockMode) 
		throws HibernateException
	{
		session.lock(object, lockMode);
	}

	@Override
	public void lock(String entityName, Object object, LockMode lockMode)
		throws HibernateException
	{
		session.lock(entityName, object, lockMode);
	}

	@Override
	public Object merge(Object object) 
		throws HibernateException
	{
		return session.merge(object);
	}

	@Override
	public Object merge(String entityName, Object object)
		throws HibernateException
	{
		return session.merge(entityName, object);
	}

	@Override
	public void persist(Object object)
		throws HibernateException
	{
		session.persist(object);
	}

	@Override
	public void persist(String entityName, Object object)
		throws HibernateException
	{
		session.persist(entityName, object);
	}

	@Override
	public void reconnect(Connection connection) 
		throws HibernateException
	{
		session.reconnect(connection);
	}

	@Override
	public void refresh(Object object, LockMode lockMode)
		throws HibernateException
	{
		session.refresh(object, lockMode);
	}

	@Override
	public void refresh(Object object, LockOptions lockOptions)
		throws HibernateException
	{
		session.refresh(object, lockOptions);
	}

	@Override
	public void refresh(Object object) 
		throws HibernateException
	{
		session.refresh(object);
	}

	@Override
	public void replicate(Object object, ReplicationMode replicationMode)
		throws HibernateException
	{
		session.replicate(object, replicationMode);
	}

	@Override
	public void replicate(String entityName, Object object, ReplicationMode replicationMode)
		throws HibernateException
	{
		session.replicate(entityName, object, replicationMode);
	}

	@Override
	public Serializable save(Object object) 
		throws HibernateException
	{
		return session.save(object);
	}

	@Override
	public Serializable save(String entityName, Object object)
		throws HibernateException
	{
		return session.save(entityName, object);
	}

	@Override
	public void saveOrUpdate(Object object) 
		throws HibernateException
	{
		session.saveOrUpdate(object);
	}

	@Override
	public void saveOrUpdate(String entityName, Object object)
		throws HibernateException
	{
		session.saveOrUpdate(entityName, object);
	}

	@Override
	public void setCacheMode(CacheMode cacheMode)
	{
		session.setCacheMode(cacheMode);
	}

	@Override
	public void setDefaultReadOnly(boolean readOnly)
	{
		session.setDefaultReadOnly(readOnly);
	}

	@Override
	@Deprecated
	public void setFlushMode(FlushMode flushMode)
	{
		session.setFlushMode(flushMode);
	}

	@Override
	public void setReadOnly(Object entityOrProxy, boolean readOnly)
	{
		session.setReadOnly(entityOrProxy, readOnly);
	}

	@Override
	public void update(Object object) throws HibernateException
	{
		session.update(object);
	}

	@Override
	public void update(String entityName, Object object)
		throws HibernateException
	{
		session.update(entityName, object);
	}

	@Override
	public String getTenantIdentifier()
	{
		return session.getTenantIdentifier();
	}

	@Override
	public ProcedureCall getNamedProcedureCall(String name)
	{
		return session.getNamedProcedureCall(name);
	}

	@Override
	public ProcedureCall createStoredProcedureCall(String name) 
	{
		return session.createStoredProcedureCall(name);
	}

	@Override
	@SuppressWarnings("rawtypes")
	public ProcedureCall createStoredProcedureCall(String paramString, Class... paramArrayOfClass) 
	{
		return session.createStoredProcedureCall(paramString, paramArrayOfClass);
	}

	@Override
	public ProcedureCall createStoredProcedureCall(String paramString,
			String... paramArrayOfString) 
	{
		return session.createStoredProcedureCall(paramString, paramArrayOfString);
	}

	@Override
	public SharedSessionBuilder<?> sessionWithOptions() 
	{
		return session.sessionWithOptions();
	}

	@Override
	public void refresh(String paramString, Object paramObject)
	{
		session.refresh(paramString, paramObject);
	}

	@Override
	public void refresh(String paramString, Object paramObject,
			LockOptions paramLockOptions) 
	{
		session.refresh(paramString, paramObject, paramLockOptions);
	}

	@SuppressWarnings("rawtypes")
	@Override
	public IdentifierLoadAccess byId(String paramString) 
	{
		return session.byId(paramString);
	}

	@Override
	public <T> IdentifierLoadAccess<T> byId(Class<T> paramClass) 
	{
		return session.byId(paramClass);
	}

	@SuppressWarnings("rawtypes")
	@Override
	public NaturalIdLoadAccess byNaturalId(String paramString) 
	{
		return session.byNaturalId(paramString);
	}

	@Override
	public <T> NaturalIdLoadAccess<T> byNaturalId(Class<T> paramClass) 
	{
		return session.byNaturalId(paramClass);
	}

	@SuppressWarnings("rawtypes")
	@Override
	public SimpleNaturalIdLoadAccess bySimpleNaturalId(String paramString) 
	{
		return session.bySimpleNaturalId(paramString);
	}

	@Override
	public <T> SimpleNaturalIdLoadAccess<T> bySimpleNaturalId(Class<T> paramClass)
	{
		return session.bySimpleNaturalId(paramClass);
	}

	@Override
	public <T> T doReturningWork(ReturningWork<T> paramReturningWork)
			throws HibernateException 
	{
		return session.doReturningWork(paramReturningWork);
	}

	@Override
	public void addEventListeners(
			SessionEventListener... paramArrayOfSessionEventListener) 
	{
		session.addEventListeners(paramArrayOfSessionEventListener);
	}

	@Override
	public Integer getJdbcBatchSize() 
	{
		return session.getJdbcBatchSize();
	}

	@Override
	public void setJdbcBatchSize(Integer size) 
	{
		session.setJdbcBatchSize(size);
	}

	@Override
	public <T> MultiIdentifierLoadAccess<T> byMultipleIds(Class<T> entityClass) 
	{
		return session.byMultipleIds(entityClass);
	}

	@Override
	public MultiIdentifierLoadAccess<?> byMultipleIds(String entityName) 
	{
		return session.byMultipleIds(entityName);
	}

	@Override
	public boolean contains(String entityName, Object entity)
	{
		return session.contains(entityName, entity);
	}

	@Override
	public <T> Query<T> createNamedQuery(String name, Class<T> resultType) 
	{
		Query<T> q = getCachedQuery(name, resultType);
		if (q == null)
		{
			q = session.createNamedQuery(name, resultType);
			cacheQuery(name, q);
		}
		return q;
	}

	@SuppressWarnings("rawtypes")
	@Override
	public Query<?> createQuery(CriteriaDelete criteria) 
	{
		return session.createQuery(criteria);
	}

	@Override
	public <T> Query<T> createQuery(CriteriaQuery<T> criteria) 
	{
		return session.createQuery(criteria);
	}

	@SuppressWarnings("rawtypes")
	@Override
	public Query<?> createQuery(CriteriaUpdate criteria) 
	{
		return session.createQuery(criteria);
	}

	@Override
	public <T> Query<T> createQuery(String queryString, Class<T> resultType) 
	{
		Query<T> q = getCachedQuery(queryString, resultType);
		if (q == null)
		{
			q = session.createQuery(queryString, resultType);
			q.setFetchSize(HibernateUtil.getJdbcFetchSize());
			cacheQuery(queryString, q);
		}
		return q;
	}

	@Override
	public FlushMode getHibernateFlushMode() 
	{
		return session.getHibernateFlushMode();
	}

	@Override
	public void setHibernateFlushMode(FlushMode flushMode) 
	{
		session.setHibernateFlushMode(flushMode);
	}

	@Override
	public Query<?> createNamedQuery(String queryName) 
	{
		Query<?> q = getCachedQuery(queryName);
		if (q == null)
		{
			q = session.getNamedQuery(queryName);
			cacheQuery(queryName, q);
		}
		return q;
	}

	@SuppressWarnings({ "rawtypes" })
	@Override
	public NativeQuery createNativeQuery(String sql, Class resultClass) 
	{
		NativeQuery q = (NativeQuery)getCachedQuery(sql);
		if (q == null)
		{
			q = session.createNativeQuery(sql, resultClass);
			cacheQuery(sql, q);
		}
		return q;
	}

	@Override
	public NativeQuery<?> createNativeQuery(String sql, String resultSetMapping) 
	{
		NativeQuery<?> q = (NativeQuery<?>)getCachedQuery(sql);
		if (q == null)
		{
			q = session.createNativeQuery(sql, resultSetMapping);
			cacheQuery(sql, q);
		}
		return q;
	}

	@Override
	public NativeQuery<?> createNativeQuery(String sql) 
	{
		NativeQuery<?> q = (NativeQuery<?>)getCachedQuery(sql);
		if (q == null)
		{
			q = session.createNativeQuery(sql);
			cacheQuery(sql, q);
		}
		return q;
	}

	@Override
	public NativeQuery<?> getNamedNativeQuery(String queryName) 
	{
		NativeQuery<?> q = (NativeQuery<?>)getCachedQuery(queryName);
		if (q == null)
		{
			q = session.getNamedNativeQuery(queryName);
			cacheQuery(queryName, q);
		}
		return q;
	}

	@Override
	public <T> RootGraph<T> createEntityGraph(Class<T> rootType) 
	{
		return session.createEntityGraph(rootType);
	}

	@Override
	public RootGraph<?> createEntityGraph(String graphName) 
	{
		return session.createEntityGraph(graphName);
	}

	@Override
	public StoredProcedureQuery createNamedStoredProcedureQuery(String name) 
	{
		return session.createNamedStoredProcedureQuery(name);
	}

	@SuppressWarnings("rawtypes")
	@Override
	public StoredProcedureQuery createStoredProcedureQuery(String procedureName, Class... resultClasses)
	{
		return session.createStoredProcedureQuery(procedureName, resultClasses);
	}

	@Override
	public StoredProcedureQuery createStoredProcedureQuery(String procedureName, String... resultSetMappings) 
	{
		return session.createStoredProcedureQuery(procedureName, resultSetMappings);
	}

	@Override
	public StoredProcedureQuery createStoredProcedureQuery(String procedureName) 
	{
		return session.createStoredProcedureQuery(procedureName);
	}

	@Override
	public void detach(Object object) 
	{
		session.detach(object);
	}

	@Override
	public <T> T find(Class<T> entityClass, Object primaryKey, LockModeType lockMode, Map<String, Object> properties) 
	{
		return session.find(entityClass, primaryKey, lockMode, properties);
	}

	@Override
	public <T> T find(Class<T> entityClass, Object primaryKey, LockModeType lockMode) 
	{
		return session.find(entityClass, primaryKey, lockMode);
	}

	@Override
	public <T> T find(Class<T> entityClass, Object primaryKey, Map<String, Object> properties) 
	{
		return session.find(entityClass, primaryKey, properties);
	}

	@Override
	public <T> T find(Class<T> entityClass, Object primaryKey) 
	{
		return session.find(entityClass, primaryKey);
	}

	@Override
	public CriteriaBuilder getCriteriaBuilder() 
	{
		return session.getCriteriaBuilder();
	}

	@Override
	public Object getDelegate() 
	{
		return session;
	}

	@Override
	public RootGraph<?> getEntityGraph(String graphName) 
	{
		return session.getEntityGraph(graphName);
	}

	@Override
	public <T> List<EntityGraph<? super T>> getEntityGraphs(Class<T> entityClass) 
	{
		return session.getEntityGraphs(entityClass);
	}

	@Override
	public EntityManagerFactory getEntityManagerFactory() 
	{
		return session.getEntityManagerFactory();
	}

	@Override
	public LockModeType getLockMode(Object entity) 
	{
		return session.getLockMode(entity);
	}

	@Override
	public Metamodel getMetamodel() 
	{
		return session.getMetamodel();
	}

	@Override
	public Map<String, Object> getProperties() 
	{
		return session.getProperties();
	}

	@Override
	public <T> T getReference(Class<T> entityClass, Object primaryKey) 
	{
		return session.getReference(entityClass, primaryKey);
	}

	@Override
	public boolean isJoinedToTransaction() 
	{
		return session.isJoinedToTransaction();
	}

	@Override
	public void joinTransaction() 
	{
		session.joinTransaction();
	}

	@Override
	public void lock(Object entity, LockModeType lockMode, Map<String, Object> properties)
	{
		session.lock(entity, lockMode, properties);
	}

	@Override
	public void lock(Object entity, LockModeType lockMode) 
	{
		session.lock(entity, lockMode);
	}

	@Override
	public void refresh(Object entity, LockModeType lockMode, Map<String, Object> properties) 
	{
		session.refresh(entity, lockMode, properties);
	}

	@Override
	public void refresh(Object entity, LockModeType lockMode) 
	{
		session.refresh(entity, lockMode);
	}

	@Override
	public void refresh(Object entity, Map<String, Object> properties) 
	{
		session.refresh(entity, properties);
	}

	@Override
	public void remove(Object entity) 
	{
		session.remove(entity);
	}

	@Override
	public void setFlushMode(FlushModeType flushMode) 
	{
		session.setFlushMode(flushMode);
	}

	@Override
	public void setProperty(String propertyName, Object value) 
	{
		session.setProperty(propertyName, value);
	}

	@Override
	public <T> T unwrap(Class<T> cls) 
	{
		return cls == this.getClass() ? (T)this : session.unwrap(cls);
	}

	@Override
	@Deprecated
	public Session getSession() 
	{
		return this;
	}

}
