package net.sf.basedb.core.query;

import java.util.Collection;
import java.util.Collections;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Type;
import net.sf.basedb.util.EqualsHelper;


/**
	Expression that can be used to filter on the existence or non-existence
	of a named any-to-any link. 

	@author nicklas
	@since 3.6
*/
public class AnyToAnyRestriction 
	implements Restriction 
{

	/**
		Create a restriction that restrict a query to return
		items that has a linked item with the given link name.
		@param alias Alias to resolve the link against, use
			null to resolve the link agains the root entity
		@param linkName The name of the link (required)
		@param linkedTo The type of the linked item (optional)
		@param linkedQuery An optional query that the linked items must match
	*/
	public static AnyToAnyRestriction exists(String alias, String linkName, Item linkedTo, EntityQuery linkedQuery)
	{
		return new AnyToAnyRestriction(alias, linkName, linkedTo, true, linkedQuery);
	}
	
	/**
		Create a restriction that restrict a query to return
		items that is missing a linked item with the given link 
		name.
		@param alias Alias to resolve the link against, use
			null to resolve the link agains the root entity
		@param linkName The name of the link (required)
		@param linkedTo The type of the linked item (optional)
		@param linkedQuery An optional query that the linked items must not match
	*/
	public static AnyToAnyRestriction missing(String alias, String linkName, Item linkedTo, EntityQuery linkedQuery)
	{
		return new AnyToAnyRestriction(alias, linkName, linkedTo, false, linkedQuery);
	}
	
	private final String alias;
	private final String linkName;
	private final Item linkedTo;
	private final boolean hasLink;
	private final EntityQuery linkedQuery;
	private final Restriction linkedRestriction;
	private String parameterName;
	
	private AnyToAnyRestriction(String alias, String linkName, Item linkedTo, boolean hasLink, EntityQuery linkedQuery) 
	{
		this.alias = alias;
		this.linkName = linkName;
		this.linkedTo = linkedTo;
		this.hasLink = hasLink;
		this.linkedQuery = linkedQuery;
		this.linkedRestriction = linkedQuery == null ? null : Restrictions.in(Hql.property("at", "toId"), linkedQuery);
	}
	
	@Override
	public String toQl(Query query, DbControl dc)
	{
		if (parameterName == null)
		{
			parameterName = "LN"+dc.uniqueRandom();
		}
		query.setParameter(parameterName, linkName, Type.STRING);
		
		StringBuilder sb = new StringBuilder();
		sb.append(alias == null ? query.getRootAlias() : alias);
		sb.append(".id ");
		sb.append(hasLink ? "IN (" : "NOT IN (");
		sb.append("SELECT at.fromId FROM AnyToAnyData at WHERE at.name = :");
		sb.append(parameterName);
		if (linkedTo != null)
		{
			sb.append(" AND at.toType=").append(linkedTo.getValue());
			if (linkedRestriction != null)
			{
				sb.append(" AND ").append(linkedRestriction.toQl(query, dc));
			}
		}
		sb.append(")");
		
		return sb.toString();
	}

	@Override
	public Collection<? extends QueryElement> getChildren()
	{
		return linkedRestriction == null ? null : Collections.singleton(linkedRestriction);
	}

	
	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		if (alias != null) sb.append(alias).append(".");
		sb.append("id ");
		sb.append(hasLink ? "IN (" : "NOT IN(");
		sb.append("SELECT at.fromId FROM AnyToAnyData at WHERE at.name = '");
		sb.append(linkName).append("'");
		if (linkedTo != null)
		{
			sb.append(" AND at.toType=").append(linkedTo.getValue());
			if (linkedRestriction != null)
			{
				sb.append(" AND ").append(linkedRestriction.toString());
			}
		}
		sb.append(")");
		return sb.toString();
	}
	
	/**
		This expression is equal to another AnyToAnyRestriction if their 
		named link, 'hasLink' and subquery restriction are the same.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if ((other == null) || (super.getClass() != other.getClass())) return false;
		AnyToAnyRestriction o = (AnyToAnyRestriction)other;
		return EqualsHelper.equals(linkName, o.linkName) && EqualsHelper.equals(linkedTo, o.linkedTo) 
			&& hasLink == o.hasLink && EqualsHelper.equals(linkedRestriction, o.linkedRestriction);
	}
	
	@Override
	public int hashCode()
	{
		return EqualsHelper.hashCode(linkName) + EqualsHelper.hashCode(linkedTo) + EqualsHelper.hashCode(linkedRestriction);
	}

	
}
