/*
	$Id: Restrictions.java 7034 2015-12-04 07:47:16Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.util.filter.Filter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
	A factory class to create restrictions.

	@author Samuel, Nicklas
	@version 2.0
	@base.modified $Date: 2015-12-04 08:47:16 +0100 (Fri, 04 Dec 2015) $
*/
public class Restrictions
{
	/**
		Combine one or more restrictions with AND: 
		new restriction = r[0] AND r[1] AND ...

		@param r The restrictions to combine
		@return The new restriction
		@throws InvalidDataException If any of the elements is null or
			the array doesn't contain at least one element
	*/
	public static Restriction and(Restriction... r)
		throws InvalidDataException
	{
		int firstNull = getFirstNull(r);
		if (firstNull >= 0 || r == null)
		{
			throw new InvalidUseOfNullException("r["+firstNull+"]");
		}
		if (r.length < 1) throw new InvalidDataException("'r' must contain at least one element");
		return new AndRestriction(r);
	}

	/**
		Combine one or more restrictions with AND ignoring null elements.
		@param r The restrictions to combine
		@return The new restriction, or null if the array was null
			or didn't contain any not-null elements
		@since 2.15
	*/
	public static Restriction nullSafeAnd(Restriction... r)
		throws InvalidDataException
	{
		int notNulls = countNotNulls(r);
		Restriction result = null;
		if (notNulls > 0)
		{
			result = new AndRestriction(copyNotNulls(r, new Restriction[notNulls]));
		}
		return result;
	}
	
	/**
		Combine one or more restrictions with AND: 
		new restriction = r[0] AND r[1] AND ...
	
		@param restrictions The restrictions to combine
		@return The new restriction, or null if the collection was null
			or didn't contain any not-null elements
		@since 3.5
	*/
	public static Restriction nullSafeAnd(Collection<? extends Restriction> restrictions)
		throws InvalidDataException
	{
		int notNulls = countNotNulls(restrictions);
		Restriction result = null;
		if (notNulls > 0)
		{
			result = new AndRestriction(copyNotNulls(restrictions, new Restriction[notNulls]));
		}
		return result;
	}

	
	/**
		Combine one or more restrictions with OR: new restriction = r[0] OR r[1] ...
		@param r The restrictions to combine
		@return The new restriction
		@throws InvalidDataException If any of the elements is null or
			the array doesn't contain at least one element
	*/
	public static Restriction or(Restriction... r)
		throws InvalidDataException
	{
		int firstNull = getFirstNull(r);
		if (firstNull >= 0 || r == null)
		{
			throw new InvalidUseOfNullException("r["+firstNull+"]");
		}
		if (r.length < 1) throw new InvalidDataException("'r' must contain at least one element");
		return new OrRestriction(r);
	}

	/**
		Combine one or more restrictions with OR ignoring null elements.
		@param restrictions The restrictions to combine
		@return The new restriction, or null if the array was null
			or didn't contain any not-null elements
		@since 3.5
	*/
	public static Restriction nullSafeOr(Collection<? extends Restriction> restrictions)
		throws InvalidDataException
	{
		int notNulls = countNotNulls(restrictions);
		Restriction result = null;
		if (notNulls > 0)
		{
			result = new OrRestriction(copyNotNulls(restrictions, new Restriction[notNulls]));
		}
		return result;
	}

	/**
		Combine one or more restrictions with OR ignoring null elements.
		@param r The restrictions to combine
		@return The new restriction, or null if the array was null
			or didn't contain any not-null elements
		@since 2.15
	*/
	public static Restriction nullSafeOr(Restriction... r)
		throws InvalidDataException
	{
		int notNulls = countNotNulls(r);
		Restriction result = null;
		if (notNulls > 0)
		{
			result = new OrRestriction(copyNotNulls(r, new Restriction[notNulls]));
		}
		return result;
	}

	
	/**
		Negate a restriction: new restriction = NOT r
		@param r The restrictions to negate
		@return The new restriction
		@throws InvalidDataException If the restriction is null
	*/
	public static Restriction not(Restriction r)
		throws InvalidDataException
	{
		if (r == null) throw new InvalidUseOfNullException("r");
		return new NotRestriction(r);
	}

	/**
		Compare if two expressions are equal: new restriction = e1 == e2.
		If one of the expressions are null, the IS NULL operation is used:
		new restriction = e1 IS NULL (or e2 IS NULL).

		@param e1 The left expression
		@param e2 The right expression
		@return The new restriction
		@throws InvalidDataException If both expressions are null
	*/
	public static Restriction eq(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null && e2 == null) throw new InvalidUseOfNullException("e1 and e2");
		return new EqRestriction(e1, e2);
	}
	
	/**
		Compare if two expressions are inequal: new restriction = e1 &lt;&gt; e2.
		If one of the expressions are null, the NOT IS NULL operation is used:
		new restriction = NOT e1 IS NULL (or NOT e2 IS NULL).

		@param e1 The left expression
		@param e2 The right expression
		@return The new restriction
		@throws InvalidDataException If both expressions are null
	*/
	public static Restriction neq(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null && e2 == null) throw new InvalidUseOfNullException("e1 and e2");
		return new NeqRestriction(e1, e2);
	}

	/**
		Compare if one expression is greater than another: new restriction = e1 &gt; e2.

		@param e1 The left expression
		@param e2 The right expression
		@return The new restriction
		@throws InvalidDataException If any of the expressions are null
	*/
	public static Restriction gt(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new GtRestriction(e1, e2);
	}
	
	/**
		Compare if one expression is greater than or equal
		to another: new restriction = e1 &gt;= e2.

		@param e1 The left expression
		@param e2 The right expression
		@return The new restriction
		@throws InvalidDataException If any of the expressions are null
	*/
	public static Restriction gteq(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new GteqRestriction(e1, e2);
	}

	/**
		Compare if one expression is less than another: new restriction = e1 &lt; e2.

		@param e1 The left expression
		@param e2 The right expression
		@return The new restriction
		@throws InvalidDataException If any of the expressions are null
	*/
	public static Restriction lt(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new LtRestriction(e1, e2);
	}
	
	/**
		Compare if one expression is less than or equal
		to another: new restriction = e1 &lt;= e2.

		@param e1 The left expression
		@param e2 The right expression
		@return The new restriction
		@throws InvalidDataException If any of the expressions are null
	*/
	public static Restriction lteq(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new LteqRestriction(e1, e2);
	}
	
	/**
		Compare if one expression falls between two other expressions:
		new expression = low &lt;= e AND e &lt; high.
		Since the inclusion/exclusion of the low and high limits are dependent
		of the database BASE hase defined the between this way to make sure that 
		BASE will treat BETWEEN the same way regardless of database.

		@param e The expression to check against the limits
		@param low The expression for the low limit
		@param high The expression for the high limit
		@return The new restriction
		@throws InvalidDataException If any of the expressions are null
	*/
	public static Restriction between(Expression e, Expression low, Expression high)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		if (low == null) throw new InvalidUseOfNullException("low");
		if (high == null) throw new InvalidUseOfNullException("high");
		return new BetweenRestriction(e, low, high, false);
	}

	/**
		Compare if one expression falls between two other expressions (inclusive):
		new expression = low &lt;= e AND e &lt;= high.
		Since the inclusion/exclusion of the low and high limits are dependent
		of the database BASE hase defined the between this way to make sure that 
		BASE will treat BETWEEN the same way regardless of database.
	
		@param e The expression to check against the limits
		@param low The expression for the low limit
		@param high The expression for the high limit
		@return The new restriction
		@throws InvalidDataException If any of the expressions are null
		@since 3.0
	*/
	public static Restriction betweenInclusive(Expression e, Expression low, Expression high)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		if (low == null) throw new InvalidUseOfNullException("low");
		if (high == null) throw new InvalidUseOfNullException("high");
		return new BetweenRestriction(e, low, high, true);
	}

	
	/**
		Check if an expression is contained in a set of other expressions:
		e1 IN (e2[0], e2[1], ...)
		
		@param e1 The expression to check
		@param e2 The set of expression to check against
		@return The new restriction
		@throws InvalidDataException If the set is null or empty
	*/
	public static Restriction in(Expression e1, Expression... e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		if (e2.length < 1) throw new InvalidDataException("'e2' must contain at least one element");
		return new InRestriction(e1, e2);
	}

	/**
		Create a subquery restriction: <code>e1 IN (subquery)</code>.
		The expression is useful in the WHERE part of a query only.
		For example, to find all users that are members of a group
		having a name starting with A.
		<pre class="code">
	Query groupFilter = User.getQuery();
	groupFilter.join(Hql.innerJoin("groups", "grp"));
	groupFilter.restrict(Restrictions.like(Hql.property("grp", "name"), Expressions.string("A%")));
	Query userQuery = User.getQuery();
	userQuery.restrict(Restrictions.in(Hql.property("id"), groupFilter));
	</pre>
		
		@param e1 The left expression
		@param subquery The subquery
		@return The new restriction
		@throws InvalidDataException If the subquery is null
		@since 3.5
	*/
	public static Restriction in(Expression e1, Query subquery)
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (subquery == null) throw new InvalidUseOfNullException("subquery");
		return new InRestriction(e1, new SubqueryExpression("", subquery));
	}
	
	/**
		Check if one expression matches another: e1 LIKE e2
		@param e1 The left expression
		@param e2 The right expression
		@return The new restriction
		@throws InvalidDataException If any of the expressions are null
	*/
	public static Restriction like(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new LikeRestriction(e1, e2);
	}

	/**
	 	Check if one expression matches any expressions in an array:
	 	(e1 LIKE e2.1) OR (e1 LIKE e2.2) OR ....
	 	@param e1 The left expression
	 	@param e2 The array with expression to match against, null values are allowed
	 		and compared with IS NULL instead
	 	@return The new restriction.
	 	@throws InvalidDataException If any of the expressions are null.
	 	@since 2.6
	 */
	public static Restriction like_in(Expression e1, Expression...e2)
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		if (e2.length < 1) throw new InvalidDataException("'e2' must contain at least one element");
		List<Restriction> restrictions = new ArrayList<Restriction>(e2.length);		
		for (Expression e : e2)
		{
			if (e == null)
			{
				restrictions.add(Restrictions.eq(e1, null));
			}
			else
			{
				restrictions.add(Restrictions.like(e1, e));
			}
		}
		return Restrictions.or(restrictions.toArray(new Restriction[restrictions.size()]));
	}
	
	/**
	 	Check if one expression doesn't matches any of the expressions in an array.
	 	(e1 NOT LIKE e2.1) AND (e1 NOT LIKE e2.2) AND...
	  	@param e1 The left expression. Null is not allowed
	  	@param e2 The array with expressions that shouldn't match, null values
	  		are allows and compared with NOT IS NULL
	  	@return The new restriction
	  	@throws InvalidDataException If any of the required expressions are null.
	  	@since 2.6
	 */
	public static Restriction notLike_in(Expression e1, Expression...e2)
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		if (e2.length < 1) throw new InvalidDataException("'e2' must contain at least one element");
		List<Restriction> restrictions = new ArrayList<Restriction>(e2.length);		
		for (Expression e : e2)
		{
			if (e == null)
			{
				restrictions.add(Restrictions.neq(e1, null));
			}
			else
			{
				restrictions.add(Restrictions.not(Restrictions.like(e1, e)));
			}
		}		
		return Restrictions.and(restrictions.toArray(new Restriction[restrictions.size()]));
	}
	
	/**
		Check if an expression matches a regular expression. The support
		for this is database dependent, but both MySQL and Postgres supports it.
		If the database doesn't support regular expression, this method will 
		create a restriction for an exact match.
		
		@param e The expression to match
		@param regexp The regular expression to test against
		@return A restriction 
	  	@throws InvalidDataException If any of the values are null
		@since 2.8
	*/
	public static Restriction rlike(Expression e, Expression regexp)
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		if (regexp == null) throw new InvalidUseOfNullException("regexp");
		return new RlikeRestriction(e, regexp);
	}
	
	/**
		Create a restriction that falls back to one of two restrictions
		depending on if the query passes a filter or not. A typical use
		case is to apply a restriction depending on the return type of
		the query. This method will allow "null" restrictions but they
		are converted to "1=1" or "1=0" depending on the nullMatchNone
		parameter.
		
		@param filter A filter, if null the 'ifTrue' restriction is returned
		@param ifTrue The restriction to apply if the filter evaluates to TRUE, or null to not apply
			any restriction in this case
		@param ifFalse The restriction to apply if the filter evaluates to FALSE, or null to not apply
			any restriction in this case
		@param nullMatchNone FALSE to match all items if the filter is null, TRUE to
			match no items if the filter is null
		@return A conditional restriction, or the 'ifTrue' restriction if the filter is null, or
			null if both restrictions are null
		@since 3.7
	*/
	@SuppressWarnings("unchecked")
	public static Restriction conditionalRestriction(Filter<? extends Query> filter, 
		Restriction ifTrue, Restriction ifFalse, boolean nullMatchNone)
	{
		if (filter == null || (ifTrue == null && ifFalse == null)) return ifTrue;
		
		if (ifTrue == null) ifTrue = Restrictions.eq(Expressions.integer(1), Expressions.integer(nullMatchNone ? 0 : 1));
		if (ifFalse == null) ifFalse = Restrictions.eq(Expressions.integer(1), Expressions.integer(nullMatchNone ? 0 : 1));
		return new ConditionalQueryElement((Filter<Query>)filter, ifTrue, ifFalse);
	}
	
	/**
		Test if there is null in the array.
		@return The index of the first null element,
			or -1 if no null elements are found, or 0 if
			the array itself is null
	*/
	private static int getFirstNull(Object[] array)
	{
		if (array == null) return 0;
		for (int i = 0; i < array.length; ++i)
		{
			if (array[i] == null) return i;
		}
		return -1;
	}
	
	private static int countNotNulls(Object[] array)
	{
		if (array == null) return 0;
		int sum = 0;
		for (Object a : array)
		{
			if (a != null) ++sum;
		}
		return sum;
	}
	
	private static int countNotNulls(Collection<?> array)
	{
		if (array == null) return 0;
		int sum = 0;
		for (Object a : array)
		{
			if (a != null) ++sum;
		}
		return sum;
	}

	
	private static <T> T[] copyNotNulls(T[] source, T[] dest)
	{
		int index = 0;
		for (T s : source)
		{
			if (s != null)
			{
				dest[index] = s;
				++index;
			}
		}
		return dest;
	}
	
	private static <T> T[] copyNotNulls(Collection<? extends T> source, T[] dest)
	{
		int index = 0;
		for (T s : source)
		{
			if (s != null)
			{
				dest[index] = s;
				++index;
			}
		}
		return dest;
	}

}
