/**
	$Id: ProgressReporterSignalHandler.java 6898 2015-05-12 09:27:08Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.signal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import net.sf.basedb.core.ProgressReporter;

/**
	An implementation of a signal handler that uses a {@link ProgressReporter}
	to communicate signals back to the target. This handler supports any signals.
	This signal handler works by replacing the progress reporter a plug-in
	uses with it's own progress reporter. Each time the plug-in calls
	{@link ProgressReporter#display(int, String)} or {@link ProgressReporter#append(String)}
	and the percentage value is less than 100 the signal handler will check if a signal 
	has been received. If so, a {@link SignalReceivedException} will be thrown.
	<p>
	NOTE! Some parts of the BASE core only uses the <code>Thread.interrupt()</code>
	mechanism for aborting a process. For example, long-running queries against the
	dynamic database. In this case we recommend that a {@link ThreadSignalHandler},
	possible linked with {@link DelegatingSignalHandler}, is used instead.
	<p>
	Here is a general outline of a plug-in that uses the progress reporter signal handler:
	
	<pre class="code">
// To keep track of current progress
private int percent;
private String progressMessage;

// Set up signal handling. ABORT is the only supported signal
private ProgressReporterSignalHandler signalHandler;
public SignalHandler getSignalHandler()
{
   signalHandler = new ProgressReporterSignalHandler(Signal.ABORT);
   return signalHandler;
}
public void run(Request request, Response response, ProgressReporter progress)
{
   if (signalHandler != null)
   {
      signalHandler.forwardTo(progress);
      progress = signalHandler;
   }
   beginTransaction();
   boolean done = false;
   while (!done &amp;&amp; !interrupted)
   {
      try
      {
         done = doSomeWork(); // NOTE! This must not take forever!
         if (progress != null) progress.display(percent, progressMessage);
       }
       catch (SignalReceivedException ex)
       {
          // We only support the abort signal
          interrupted = true;
       }
   }
   if (interrupted)
   {
      rollbackTransaction();
      response.setError("Aborted by user", null);
   }
   else
   {
      commitTransaction();
      response.setDone("Done");
   }
}
</pre>
	
	@author nicklas
	@version 2.6
	@see ThreadSignalHandler
	@base.modified $Date: 2015-05-12 11:27:08 +0200 (Tue, 12 May 2015) $
*/
public class ProgressReporterSignalHandler
	extends AbstractSignalHandler
	implements ProgressReporter
{

	/**
		Log signals processing.
	*/
	private static final org.slf4j.Logger logger = 
		org.slf4j.LoggerFactory.getLogger(ProgressReporterSignalHandler.class);

	/**
		Original progress reporter that progress updates should be
		forwarded to.
	*/
	private ProgressReporter forwardTo;
	
	private List<Signal> received;
	
	/**
		The current percentage status.
	*/
	private int percent;

	/**
		When the progress has reached this limit signal
		checking becomes disabled. Default value is 100.
	*/
	private int limit;
	
	/**
		Create a new progress reporter signal handler that supports the specified 
		signals.
		@param supported A collection with the signals that are initially supported.
			More signals can be added with {@link #addSignal(Signal)}
	*/
	public ProgressReporterSignalHandler(Collection<Signal> supported)
	{
		super(supported);
		this.received = Collections.synchronizedList(new ArrayList<Signal>());
		this.limit = 100;
	}
	
	/*
		From the SignalHandler interface
		-------------------------------------------
	*/
	/**
		When signals are receieved, they are stored in a temporary
		list in the order they come in. When the signal target calls
		{@link #display(int, String)} or {@link #append(String)}
		a {@link SignalReceivedException} is throw if the list is not 
		empty. Once the exception has been thrown the internal list
		is cleared.
		@param signal The signal to handle
		@throws UnsupportedSignalException If the signal is not
			supported
	*/
	@Override
	public void handleSignal(Signal signal)
	{
		logger.debug("Got signal: " + signal);
		if (supports(signal))
		{
			received.add(signal);
		}
		else
		{
			logger.debug("Signal not supported " + signal);
			throw new UnsupportedSignalException(signal);
		}
	}
	// -------------------------------------------

	/*
		From the ProgressReporter interface
		-------------------------------------------
	*/
	@Override
	public void append(String message)
	{
		if (forwardTo != null) forwardTo.append(message);
		if (percent < limit) checkForSignals();
	}
	@Override
	public void display(int percent, String message)
	{
		if (forwardTo != null) forwardTo.display(percent, message);
		if (percent < limit) checkForSignals();
		this.percent = percent;
	}
	// -------------------------------------------
	
	/**
		Set the progress reporter that progress information should be
		forwarded to. 
		@param forwardTo The progress reporter, or null if progress should
			not be forwarded
	*/
	public void forwardTo(ProgressReporter forwardTo)
	{
		this.forwardTo = forwardTo;
	}
	
	/**
		Check if any signals has been received by this handler.
		@return TRUE if at least one signal has been received, FALSE otherwise
	*/
	public boolean hasReceivedSignals()
	{
		return received.size() > 0;
	}
	
	/**
		Get the list of received signals. Calling this method clears the internal
		list and if no signals are received in the meantime, the next call to
		this method will return null.
		@return A list with signals in the order they were
			received, or null if no signals has been received
	*/
	public List<Signal> getReceivedSignals()
	{
		List<Signal> toReturn = null;
		if (received.size() > 0)
		{
			toReturn = new ArrayList<Signal>(received);
			received.clear();
		}
		return toReturn;
	}
	
	/**
		Get the percentage limit where signal checking becomes disabled.
		@return The limit in percent
	*/
	public int getLimit()
	{
		return limit;
	}
	
	/**
		Set the percentage limit. When the progress has reached the limit
		signal checking will be disable. The default value is 100 since there is
		no need to abort a process when it has already been finished.
		@param limit The new limit
	*/
	public void setLimit(int limit)
	{
		this.limit = limit;
	}
	
	/**
		If at least one signal has been received a {@link SignalReceivedException}
		is thrown. This method clears the received signals.
	*/
	public void checkForSignals()
	{
		if (received.size() > 0)
		{
			throw new SignalReceivedException(getReceivedSignals());
		}
	}
	
}
