/**
	$Id $

	Copyright (C) 2022 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.snapshot;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Project;
import net.sf.basedb.util.filter.Filter;

/**
	Filter implementation that finds all annotation snapshots 
	for a given set of projects. Intended to be used with
	{@link SnapshotManager#findAnnotations(DbControl, AnnotationSetSnapshot, Filter, boolean)}
	when the manager is in the "all projects" mode. In other modes, the snapshot manager will
	select only the default values, or project-specific values for the currently active project.
	See {@link SnapshotManager#setAllProjects()}, {@link SnapshotManager#setProject(Project)}
	and {@link SnapshotManager#setNoProject()}.

	@author Nicklas
	@version 3.19.4
	@base.modified $Date$
*/
public class ProjectFilter
	implements Filter<AnnotationSnapshot>
{
	private final Set<Integer> projects;
	
	/**
		Create a filter that matches all projects that the current user has access
		to including the default values.
	*/
	public ProjectFilter(DbControl dc)
	{
		ItemQuery<Project> query = Project.getQuery();
		query.setIncludes(Include.ALL);
		this.projects = new HashSet<>(query.idList(dc));
		projects.add(0); // Matches default values
	}
	
	/**
		Create a filter that matches the given projects and optionally
		default values.
	*/
	public ProjectFilter(Collection<Project> projects, boolean includeDefaultValues)
	{
		this.projects = new HashSet<>();
		if (includeDefaultValues) this.projects.add(0);
		for (Project p : projects)
		{
			this.projects.add(p.getId());
		}
	}
	
	/**
		Create a filter that matches the given projects ID values. To match default values
		make sure that '0' is included in the given values.
	*/
	public ProjectFilter(Collection<Integer> projects)
	{
		this.projects = new HashSet<>(projects);
	}

	/*
		From the Filter interface
		-------------------------
	*/
	@Override
	public boolean evaluate(AnnotationSnapshot snapshot)
	{
		return projects.contains(snapshot.getProjectId());
	}
	// -------------------------


}
