package net.sf.basedb.util.excel;


import java.io.IOException;
import java.io.OutputStream;
import java.util.Collections;
import java.util.Map;

import org.apache.commons.io.output.NullWriter;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbookType;

import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.encode.EncoderDecoder;
import net.sf.basedb.util.export.TableWriter;
import net.sf.basedb.util.formatter.Formatter;

/**
	This is a simple wrapper for the TableWriter class to make it
	create an Excel workbook instead of a CSV file. There are some
	things to consider:
	
	* The  {@link #tablePrintHeaders(Object...)} and {@link #tablePrintData(Object...)} 
	  (including overloaded variants) methods can be used to write a single row at
	  a time. The various print() and write() methods write to a single cell and
	  will advance one column at a time. The println() variants ends the current row 
	  and starts over on the first column.
 	  
 	* Writing data to Excel is by default handled by the {@link AutoFormatter} implementation
 	  which will try to infer the correct Excel data type and format string from the class
 	  and value of the data.
 
    * It is possible to assign a different formatter implementation to a given column by
      calling {@link #setColumnFormatter(int, Formatter)}, but note that the formatter is
      only used by this class if it also implements the {@link ExcelFormatter} interface.
 
	* Call {@link #saveTo(OutputStream)} to write the Excel workbook to an output location.

	@since 3.15
*/
public class XlsxTableWriter 
	extends TableWriter
{

	private final String name;
	private final Workbook workbook;
	private final Sheet sheet;
	private final boolean externalWorkbook; // TRUE if the workbook was supplied externally
	private final CellStyleCreator styleCreator;
	private final AutoFormatter autoFormatter;
	
	private int rowNo;
	private Row currentRow;
	private int colNo;
	
	/**
		Create a new workbook with a single worksheet.
		@param name The name of the worksheet
	*/
	public XlsxTableWriter(String name)
	{
		this(null, name);
	}

	/**
		Add a sheet to the given Excel workbook.
		@param name The name of the worksheet
	*/
	public XlsxTableWriter(Workbook workbook, String name)
	{
		super(new NullWriter());
		this.name = name;
		this.externalWorkbook = workbook != null;
		if (workbook == null) workbook = new XSSFWorkbook(XSSFWorkbookType.XLSX);
		this.workbook = workbook; 
		this.sheet = workbook.createSheet(name);
		this.styleCreator = new CellStyleCreator(workbook);
		this.autoFormatter = new AutoFormatter();
	}
	
	/**
		Get the workbook.
	*/
	public Workbook getWorkbook()
	{
		return workbook;
	}
	
	/**
		Get the current sheet in the workbook.
	*/
	public Sheet getSheet()
	{
		return sheet;
	}

	/**
		Get row that was the last one to be printed.
	*/
	public Row getCurrentRow()
	{
		return currentRow;
	}
	
	/**
		Get the auto-formatter instance to use for columns that has no 
		specified formatter via {@link #setColumnFormatter(int, Formatter)}.
	*/
	public AutoFormatter getAutoFormatter()
	{
		return autoFormatter;
	}
	
	/**
		Get the style creator that can be used for custom styling of cells.
	*/
	public CellStyleCreator getCellStyleCreator()
	{
		return styleCreator;
	}
	
	@Override
	public void tablePrintHeaders(Object... headers)
	{
		super.tablePrintHeaders(headers);
		// Let the default implementation set the values and then we can try to adjust column widths
		// to match the text length
		for (int col = 0; col < headers.length; col++)
		{
			Cell cell = currentRow.getCell(col);
			if (cell == null) continue;
			String headerText = cell.getStringCellValue();
			if (headerText == null) continue;
			
			// Set column width based on title length + margin --  but not less than 10 or wider than 50
			int colWidth = 2 + Math.min(Math.max(8, headerText.length()), 48);
			sheet.setColumnWidth(col, colWidth*256);
		}
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	@Override
	public void tablePrintData(EncoderDecoder encoder, Map<Integer, Formatter<?>> formatters, Object... data) 
	{
		currentRow = sheet.createRow(rowNo++);
		colNo = 0;
		String nv = getNullValue();
		if (formatters == null) formatters = Collections.emptyMap();
		
		for (int col = 0; col < data.length; col++)
		{
			// Convert and format data value to ExcelValue
			Formatter fmt = formatters.get(col);
			ExcelFormatter excelFmt = (fmt instanceof ExcelFormatter) ? (ExcelFormatter)fmt : autoFormatter;
			ExcelValue ev = excelFmt.toExcelValue(data[col]);
			Cell cell = currentRow.createCell(col);
			ev.writeTo(cell, styleCreator);
		}
	}

	/**
		Closes the workbook unless it was supplied in the constructor from an
		external source.
	*/
	@Override
	public void close()
	{
		super.close();
		if (!externalWorkbook) FileUtil.close(workbook);
	}
	
	/**
		Save the workbook to the given stream.
	*/
	public void saveTo(OutputStream out)
		throws IOException
	{
		workbook.write(out);
	}
	
	@Override
	public void write(int c) 
	{
		writeData(c, false);
	}

	@Override
	public void write(char[] buf, int off, int len) 
	{
		writeData(String.copyValueOf(buf, off, len), false);
	}

	@Override
	public void write(char[] buf) 
	{
		write(buf, 0, buf.length);
	}

	@Override
	public void write(String s, int off, int len) 
	{
		writeData(s.substring(off, off+len), false);
	}

	@Override
	public void write(String s) 
	{
		writeData(s, false);
	}

	@Override
	public void print(boolean b) 
	{
		writeData(b, false);
	}

	@Override
	public void print(char c) 
	{
		writeData(c, false);
	}

	@Override
	public void print(int i) 
	{
		writeData(i, false);
	}

	@Override
	public void print(long l) 
	{
		writeData(l, false);
	}

	@Override
	public void print(float f) 
	{
		writeData(f, false);
	}

	@Override
	public void print(double d) 
	{
		writeData(d, false);
	}

	@Override
	public void print(char[] s) 
	{
		writeData(String.copyValueOf(s), false);
	}

	@Override
	public void print(String s) 
	{
		writeData(s, false);
	}

	@Override
	public void print(Object obj) 
	{
		writeData(obj, false);
	}

	@Override
	public void println() 
	{
		if (colNo == 0) rowNo++;
		colNo = 0;
	}

	@Override
	public void println(boolean b) 
	{
		writeData(b, true);
	}

	@Override
	public void println(char c) 
	{
		writeData(c, true);
	}

	@Override
	public void println(int i) 
	{
		writeData(i, true);
	}

	@Override
	public void println(long l) 
	{
		writeData(l, true);
	}

	@Override
	public void println(float f) 
	{
		writeData(f, true);
	}

	@Override
	public void println(double d) 
	{
		writeData(d, true);
	}

	@Override
	public void println(char[] s) 
	{
		writeData(String.copyValueOf(s), true);
	}

	@Override
	public void println(String s)
	{
		writeData(s, true);
	}

	@Override
	public void println(Object o) 
	{
		writeData(o, true);
	}

	/**
		Write the data to the next cell. If colNo=0 a new row is created, 
		otherwise the next cell in the current row is used. If
		'endOfRow' the current row is ended (by setting colNo to 0) so
		that a new row is created the next time this method is called.
	*/
	@SuppressWarnings({ "rawtypes", "unchecked" })
	private void writeData(Object data, boolean endOfRow)
	{
		if (colNo == 0) currentRow = sheet.createRow(rowNo++);
		Formatter fmt = getColumnFormatter(colNo);
		ExcelFormatter excelFmt = (fmt instanceof ExcelFormatter) ? (ExcelFormatter)fmt : autoFormatter;
		ExcelValue ev = excelFmt.toExcelValue(data);
		Cell cell = currentRow.createCell(colNo++);
		ev.writeTo(cell, styleCreator);
		if (endOfRow) colNo = 0;
	}
	
}
