/**
	$Id: ExtensionsManager.java 7233 2016-11-17 12:36:34Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.manager;

import java.io.File;
import java.io.FileFilter;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import net.sf.basedb.core.ItemAlreadyExistsException;
import net.sf.basedb.util.RegexpFileFilter;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.ExtensionsFile.WriteableExtensionsFile;
import net.sf.basedb.util.filter.Filter;

/**
	An extensions manager is responsible for handling a group of 
	extensions in the form of files stored in the file system. The 
	manager will keep track of added, updated and removed files.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2016-11-17 13:36:34 +0100 (Thu, 17 Nov 2016) $
*/
public class ExtensionsManager
{
	
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(ExtensionsManager.class);

	// Main registry for extensions
	private final Registry registry;
	// Settings file
	private final Settings settings;
	// Manually added ignore files
	private final Set<File> ignore;
	// Directories to scan for new/deleted/modified extensions
	private final Set<File> directories;
	// Extension files
	private final Map<URI, ExtensionsFile> xtFiles;
	// Map objects to the file they are defined in. The key is a unique object id.
	private final Map<ObjectKey<?>, ExtensionsFile> installedObjects;
	
	private ProcessResults lastResults;
	
	private int numNew;
	private int numModified;
	private int numUnmodified;
	private int numDeleted;
	private int numIgnored;
	
	/**
		Create a new extensions manager and use the given registry
		to store extensions.
		
		@param registry The registry to use
		@param settingsFile The file were settings are stored
	*/
	public ExtensionsManager(Registry registry, File settingsFile)
	{
		this.registry = registry;
		this.settings = new Settings(this, settingsFile);
		this.ignore = new TreeSet<File>(settings.getIgnoredFiles());
		this.directories = new HashSet<File>();
		this.xtFiles = new TreeMap<URI, ExtensionsFile>(new JarFirstURIComparator());
		this.installedObjects = new HashMap<ObjectKey<?>, ExtensionsFile>();
		addIgnore(settingsFile);
	}
	
	/**
		Get the registry in which all extensions that are managed by this
		manager are stored.
	*/
	public Registry getRegistry()
	{
		return registry;
	}
	
	/**
		Get the settings for the extension manager.
	*/
	public Settings getSettings()
	{
		return settings;
	}
	
	/**
		Save the last process results so that they can be used 
		later on.
		@since 3.3
	*/
	public void setLastProcessResults(ProcessResults results)
	{
		this.lastResults = results;
	}
	
	/**
		Get the results of the last processing.
		@since 3.3
	*/
	public ProcessResults getLastProcessResults()
	{
		return lastResults;
	}
	
	/**
		Add a file to the ignore list. Files in this list will never
		be checked for extensions again. This is only neccessary for XML/JAR
		files that doesn't contain extensions.
		@param file The file to add
		@see #removeIgnore(File)
	*/
	public synchronized void addIgnore(File file)
	{
		if (file == null) return;
		if (ignore.add(file))
		{
			settings.setIgnoredFile(file, true);
			removeFile(file);
		}
	}
	
	/**
		Remove a file from the ignore list.
		@param file The file to remove
		@see #addIgnore(File)
	*/
	public synchronized void removeIgnore(File file)
	{
		if (file == null) return;
		if (ignore.remove(file))
		{
			settings.setIgnoredFile(file, false);
		}
	}

	/**
		Get all ignored files. 
		
		@param onlyInManagedDirectory TRUE to only look for files that
			are located in a managed directory
		@since 3.10
	*/
	public List<File> getIgnoredFiles(boolean onlyInManagedDirectory)
	{
		// Create a copy, or the iterator will fail if new
		// files are added by another thread
		List<File> copy = new ArrayList<File>(ignore.size());
		if (onlyInManagedDirectory)
		{
			for (File f : ignore)
			{
				if (directories.contains(f.getParentFile())) copy.add(f);
			}
		}
		else
		{
			copy.addAll(ignore);
		}
		return copy;
	}
	
	/**
		Add a directory to this manager. The directory will be monitored
		for extension files in the future. Files in the directory that
		are installed as per the settings {@link Settings#isInstalledFile(File)}
		are loaded into manager. 
		<p>
		This method call is ignored if the path is not pointing to an
		existing directory.
		
		@param dir An existing directory
		@return The number of files in the directory that was
			installed
	*/
	public synchronized int addDirectory(File dir)
	{
		if (dir == null) throw new NullPointerException("dir");
		if (!dir.isDirectory()) return 0;
		directories.add(dir);
		return scanForInstalledFiles(dir, settings);
	}
	
	
	/**
	 	Add a file to the manager that is supposed to contain extensions.
	 	The file either be an XML file containing definitions, or a JAR
	 	file with the definitions in the META-INF/extensions.xml path.
	 	
		@param file An extensions file (null is not allowed)
		@return The extension file object representing the added file
	*/
	public synchronized ExtensionsFile addFile(File file)
	{
		if (file == null) throw new NullPointerException("file");
		ExtensionsFile xtFile = new ExtensionsFile(this, file);
		addExtensionsFile(xtFile);
		removeIgnore(file);
		return xtFile;
	}
	
	/**
		Remove a file from the extensions manager. This method will
		remove some information about the file from the manager itself,
		but it will not touch registered extensions, plug-ins, extracted
		resources, etc. The file itself is not deleted from the disk.
		<p>
		Note that if the file remains in a managed directory it will
		reappear in the manager if {@link #scanForChanges()} is
		called.
		
		@param file The file to remove (if null this method is ignored)
	*/
	ExtensionsFile removeFile(File file)
	{
		if (file == null) return null;
		settings.setUninstalledFile(file);
		ExtensionsFile xtFile = xtFiles.remove(file.toURI());
		if (xtFile != null)
		{
			unregisterAllObjects(xtFile);
		}
		return xtFile;
	}
	
	/**
	 	Add an URI that points to an XML file containing definitions.
	 	This method is mostly useful for system-defined extensions that
	 	are defined in a XML file inside a JAR file.
	 	<p>
	 	
	 	This method can't be used for JAR files and if used for
	 	XML files it doesn't detect changes that are made after 
	 	installation.
	 	
		@param uri An URI pointing to an extensions file (null is not allowed)
		@return The extension file object representing the added uri
	*/
	public synchronized ExtensionsFile addURI(URI uri)
	{
		if (uri == null) throw new NullPointerException("uri");
		ExtensionsFile xtFile = new ExtensionsFile(this, uri);
		addExtensionsFile(xtFile);
		return xtFile;
	}
	
	private void addExtensionsFile(ExtensionsFile xtFile)
	{
		xtFile.validate();
		xtFiles.put(xtFile.getURI(), xtFile);
		log.info("Added file: " + xtFile);
	}
	
	/**
		Scan the managed directories for new, updated and deleted files.
		This method is used for gathering statistics only. To do something
		with the files the {@link #processFiles(ExtensionsFileProcessor)} 
		method must be called.
		<p>
		Note! The new or modified status of files are NOT reset as a result
		of calling this method. The status can only be changed as a result
		of some processing.
		
		@return The number of new + modified + deleted files that was found
	*/
	public synchronized int scanForChanges()
	{
		numNew = 0;
		numModified = 0;
		numDeleted = 0;
		numUnmodified = 0;
		numIgnored = 0;
		
		// 1. Check and count all known files
		Iterator<ExtensionsFile> it = xtFiles.values().iterator();
		while (it.hasNext())
		{
			ExtensionsFile xtFile = it.next();
			if (!xtFile.exists()) 
			{
				log.debug("File '" + xtFile + "' has been deleted.");
				numDeleted++;
			}
			else if (xtFile.isNew())
			{
				log.debug("File '" + xtFile + "' is new. Validating...");
				xtFile.validate();
				log.debug("File '" + xtFile + "' is " + (xtFile.isValid() ? "valid" : "invalid"));
				numNew++;
			}
			else if (xtFile.checkModified())
			{
				log.debug("File '" + xtFile + "' is modified. Validating...");
				xtFile.validate();
				log.debug("File '" + xtFile + "' is " + (xtFile.isValid() ? "valid" : "invalid"));
				numModified++;
			}
			else
			{
				numUnmodified++;
			}
		}
		
		// Scan the managed directories for new files
		for (File dir : directories)
		{
			numNew += scanForNewFiles(dir);
		}
		
		return numModified + numNew + numDeleted;
	}
	
	/**
		Get the number of new files that was found in the last
		call to {@link #scanForChanges()}.
	*/
	public int getNumNew()
	{
		return numNew;
	}

	/**
		Get the number of modified files that was found in the last
		call to {@link #scanForChanges()}.
	*/
	public int getNumModified()
	{
		return numModified;
	}

	/**
		Get the number of unmodified files that was found in the last
		call to {@link #scanForChanges()}.
	*/
	public int getNumUnmodified()
	{
		return numUnmodified;
	}
	
	/**
		Get the number of deleted files that was found in the last
		call to {@link #scanForChanges()}.
	*/
	public int getNumDeleted()
	{
		return numDeleted;
	}
	
	/**
		Get the number of ignored files that was found in the last
		call to {@link #scanForChanges()}.
		@since 3.10
	*/
	public int getNumIgnored()
	{
		return numIgnored;
	}
	
	/**
		Scan the given directory and add files that are new.
		@param dir The directory to scan
		@return The number of new files in the given directory
	*/
	private int scanForNewFiles(File dir)
	{
		int numNew = 0;
		
		if (!dir.isDirectory()) 
		{
			log.warn("File '" + dir + "' is not a directory. Ignoring.");
			return numNew;
		}

		log.info("Scanning directory: " + dir);
		FileFilter filter = new RegexpFileFilter(".*\\.(xml|jar)", "");
		File[] files = dir.listFiles(filter);
		
		if (files != null && files.length > 0)
		{
			log.debug("Found " + files.length + " files");
			for (File file : files)
			{
				if (ignore.contains(file))
				{
					log.debug("File '" + file + "' is ignored.");
					numIgnored++;
					continue; // with the next file
				}
				
				URI uri = file.toURI();
				ExtensionsFile xtFile = xtFiles.get(uri);
				if (xtFile != null)
				{
					log.debug("File '" + xtFile + "' is a known file.");
				}
				else
				{
					log.debug("Adding file: " + file);
					xtFile = new ExtensionsFile(this, file);
					addExtensionsFile(xtFile);
					numNew++;
				}
			}
			log.info("Found " + numNew + " new *.xml and *.jar files");
		}
		else
		{
			log.info("Found no *.xml or *.jar files");
		}
		return numNew;
	}
	
	/**
		Scan the given directory and add files that are installed
		according to the settings.
		@param dir The directory to scan
		@return The number of installed files in the given directory
	*/
	private int scanForInstalledFiles(File dir, Settings settings)
	{
		int numInstalled = 0;
		
		if (!dir.isDirectory()) 
		{
			log.warn("File '" + dir + "' is not a directory. Ignoring.");
			return numInstalled;
		}
	
		log.info("Scanning directory: " + dir);
		FileFilter filter = new RegexpFileFilter(".*\\.(xml|jar)", "");
		File[] files = dir.listFiles(filter);
		
		if (files != null && files.length > 0)
		{
			log.debug("Found " + files.length + " files");
			for (File file : files)
			{
				if (ignore.contains(file))
				{
					log.debug("File '" + file + "' is ignored.");
					numIgnored++;
					continue; // with the next file
				}
				
				URI uri = file.toURI();
				ExtensionsFile xtFile = xtFiles.get(uri);
				if (xtFile != null)
				{
					log.debug("File '" + xtFile + "' is a known file.");
				}
				else if (settings.isInstalledFile(file))
				{
					log.debug("Adding installed file: " + file);
					xtFile = new ExtensionsFile(this, file);
					addExtensionsFile(xtFile);
					numInstalled++;
				}
				else
				{
					log.debug("File '" + file + "' is not installed.");
				}
			}
			log.info("Found " + numInstalled + " new *.xml and *.jar files");
		}
		else
		{
			log.info("Found no *.xml or *.jar files");
		}
		return numInstalled;
	}

	
	/**
		Get a list of all extension files managed by this manager.
		@return A list
	*/
	public List<ExtensionsFile> getFiles()
	{
		// Create a copy, or the iterator will fail if new
		// files are added by another thread
		List<ExtensionsFile> copy = new ArrayList<ExtensionsFile>(xtFiles.values());
		return copy;
	}

	/**
		Get the extension file that is located at the given URI. 
		This method can be used for all files even those that
		have been added with {@link #addFile(File)} or
		{@link #addDirectory(File)}.
		
		@param uri An URI that is pointing to an extensions file
 		@return An extensions file object, or null if no file is found
	*/
	public ExtensionsFile getFileByURI(String uri)
	{
		return xtFiles.get(URI.create(uri));
	}
	
	/**
		Find out which file the object with the given key was defined.
		See implementations of {@link ObjectKey} for various types
		of objects to look for.
		
		@param key The object key
		@return Information about the file the object is defined in,
			or null if no object is found for the given key
	*/
	public ExtensionsFile getFileByObjectKey(ObjectKey<?> key)
	{
		return installedObjects.get(key);
	}
	
	/**
		Register an object as defined by the given extensions file.
		
		@param key The object key
		@param xtFile The extensions file the object is defined in
	*/
	void registerObject(ObjectKey<?> key, ExtensionsFile xtFile)
	{
		ExtensionsFile xtOther = installedObjects.get(key);
		if (xtOther == null)
		{
			installedObjects.put(key, xtFile);
		}
		else if (!xtFile.equals(xtOther))
		{
			throw new ItemAlreadyExistsException(key.toDescription() + " is defined in '" + xtOther.getName() + "'");
		}
	}
	
	/**
		Unregister an object. The given extensions file must match
		be equal to the one that registered the object, otherwise
		the unregistration is ignored.
		
		@param key The object key
		@param xtFile The extensions file the object is defined in
	*/
	void unregisterObject(ObjectKey<?> key, ExtensionsFile xtFile)
	{
		if (xtFile.equals(installedObjects.get(key)))
		{
			installedObjects.remove(key);
		}
	}
	
	/**
		Unregister all objects know in the file.
	*/
	void unregisterAllObjects(ExtensionsFile xtFile)
	{
		Iterator<ExtensionsFile> it = installedObjects.values().iterator();
		while (it.hasNext())
		{
			if (it.next().equals(xtFile)) it.remove();
		}
	}
	
	/**
		Process all know file with the given processer.
		@param processor A processor implementation, null is not allowed
	*/
	public void processFiles(ExtensionsFileProcessor processor)
	{
		processFiles(processor, null);
	}
	
	/**
		Process all known extension files with the given processor.
		
		@param processor A processor implementation, null is not allowed
		@param filter An optional filter implementation that can be used to
			limit which files are sent to the processor
	*/
	public synchronized void processFiles(ExtensionsFileProcessor processor, Filter<ExtensionsFile> filter)
	{
		if (processor == null) throw new NullPointerException("processor");
		log.info("Processing files with processor: " + processor);

		// 1. Filter files with the given filter
		Collection<ExtensionsFile> filtered = null;
		if (filter != null)
		{
			filtered = new ArrayList<ExtensionsFile>(xtFiles.size());
			log.info("Filtering files with filter: " + filter);
			for (ExtensionsFile xtFile : xtFiles.values())
			{
				if (filter.evaluate(xtFile)) 
				{
					filtered.add(xtFile);
				}
				else
				{
					log.debug("File blocked by filter: " + xtFile);
				}
			}
		}
		else
		{
			filtered = xtFiles.values();
		}
		
		// 2. Process the remaining files
		int numFiles = filtered.size();
		List<WriteableExtensionsFile> writeableFiles = new ArrayList<WriteableExtensionsFile>(numFiles);
		try
		{
			log.debug("Processing " + numFiles + " file(s) with processor: " + processor);
			processor.begin(this, numFiles);
			for (ExtensionsFile xtFile : filtered)
			{
				log.debug("Processing file: " + xtFile);
				WriteableExtensionsFile wf = xtFile.getWriteableFile();
				writeableFiles.add(wf);
				processor.processFile(this, wf);
			}
			processor.done(this);
			log.debug("All files processed with processor: " + processor);
		}
		catch (Throwable t)
		{
			log.error("Error when processing extensions with " + processor, t);
			processor.done(null, t);
		}
		finally
		{
			// Make sure all WriteableExtensionFile:s are closed if the processor "forgets" it
			for (WriteableExtensionsFile wf : writeableFiles)
			{
				wf.close();
			}
		}
	}
	
	/**
		Compare URI:s so that 'jar:' scheme always are listed first. This means
		that the core defintion files are always displayed before any external
		extensions.
	*/
	public static class JarFirstURIComparator
		implements Comparator<URI>
	{
		public JarFirstURIComparator()
		{}

		@Override
		public int compare(URI u1, URI u2) 
		{
			String s1 = u1.getScheme();
			String s2 = u2.getScheme();
			if (!s1.equals(s2))
			{
				if (s1.equals("jar")) return -1;
				if (s2.equals("jar")) return 1;
			}
			return u1.compareTo(u2);
		}
	}
	
}
