/**
	$Id: IntegerFormatter.java 8138 2023-04-20 07:47:14Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.formatter;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.NumberFormat;

import net.sf.basedb.util.excel.ExcelFormatter;
import net.sf.basedb.util.excel.ExcelValue;

/**
 	Formats a number as an integer. 
 	
	@author nicklas
	@version 2.2
	@base.modified $Date: 2023-04-20 09:47:14 +0200 (Thu, 20 Apr 2023) $
*/
public class IntegerFormatter
	implements Formatter<Number>, ExcelFormatter<Number, Number>
{

	private final boolean longValues;
	private final NumberFormat format;
	
	/**
		Create a new integer formatter.
		@param longValues If the {@link #parseString(String)} method should return long
			or int values
	*/
	public IntegerFormatter(boolean longValues)
	{
		this.longValues = longValues;
		this.format = null;
	}
	
	/**
		Create a new integer formatter with a thousands separator.
		@param longValues If the {@link #parseString(String)} method should return long
			or int values
		@since 3.19.8
	*/
	public IntegerFormatter(boolean longValues, Character thousandsSeparator)
	{
		this.longValues = longValues;
		if (thousandsSeparator == null)
		{
			this.format = null;
		}
		else
		{
			DecimalFormatSymbols dfs = new DecimalFormatSymbols();
			dfs.setGroupingSeparator(thousandsSeparator);
			this.format = new DecimalFormat("#,##0", dfs);
		}
	}
	
	/*
		From the Formatter interface
		-------------------------------------------
	*/
	@Override
	public String format(Number value)
	{
		return value == null ? "" : 
			format != null ? format.format(value.longValue()) : String.valueOf(value.longValue());
	}
	@Override
	public Number parseString(String value)
	{
		if (value == null) return null;
		Double d = Double.valueOf(value);
		Number n = null;
		if (longValues)
		{
			n = Long.valueOf(d.longValue());
		}
		else
		{
			n = Integer.valueOf(d.intValue());
		}
		return n;
	}
	// -------------------------------------------

	/**
		@since 3.15
	*/
	@Override
	public ExcelValue<Number> toExcelValue(Number value) 
	{
		return ExcelValue.asInt(value);
	}

}
