/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	Source item transformer implementation that transform collected extract items 
	to derived bioassays. This implementations works on the derived bioassay level
	and merges the source derived bioassays with all derived bioassays having a 
	direct link to the set of collected extracts.
	
	@author Nicklas
	@since 3.5
*/
public class CollectedExtractsToDerivedBioAssayTransformer 
	extends AbstractSourceItemTransformer
{

	private final CollectExtracts collectedExtracts;
	private final boolean includeChildBioAssays;
	
	/**
		Create a new collected extracts to derived bioassay transformer.
	*/
	public CollectedExtractsToDerivedBioAssayTransformer(CollectExtracts collectedExtracts) 
	{
		this(collectedExtracts, true);
	}
	
	/**
		Create a new collected extracts to derived bioassay transformer.
		@param includeChildBioAssays TRUE to also include child bioassays to those that are directly
			linke (even if they are not linked or linked to other extracts)
		@since 3.19.10
	*/
	public CollectedExtractsToDerivedBioAssayTransformer(CollectExtracts collectedExtracts, boolean includeChildBioAssays) 
	{
		super(Item.DERIVEDBIOASSAY, Item.DERIVEDBIOASSAY);
		this.collectedExtracts = collectedExtracts;
		this.includeChildBioAssays = includeChildBioAssays;
	}
	
	@Override
	public Set<Integer> transform(TransformContext context, Set<Integer> source) 
	{
		DbControl dc = context.getDbControl();
		
		ItemQuery<DerivedBioAssay> query = DerivedBioAssay.getQuery();
		query.setIncludes(context.getInclude());
		query.join(Hql.leftJoin("extract", "xtr"));
		query.restrict(
			Restrictions.in(
				Hql.alias("xtr"), 
				Expressions.parameter("extracts")
			)
		);
		
		// We get the directly linked
		Set<Integer> all = safeIdList(dc, query, "extracts", collectedExtracts.getCollected());
		if (includeChildBioAssays)
		{
			//  and we need to get their child items as well
			all = new DerivedBioAssayToChildDerivedBioAssayTransformer(true).transform(context, all);
		}
		all.addAll(source);
		return all;
	}

}
