/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.SyncFilter.SourceItemTransform;

/**
	Transformer factory implementation for transforming to physical bioassays.
	The current implementation support transforming from:
	
	<ul>
	<li>{@link SourceItemTransform#PARENT_TO_CHILD}: BIOSOURCE, SAMPLE, EXTRACT
	<li>{@link SourceItemTransform#CHILD_TO_PARENT}: DERIVEDBIOASSAY, RAWBIOASSAY
	</ul>

	@author Nicklas
	@since 3.5
*/
public class ToPhysicalBioAssaySourceItemTransformerFactory 
	extends AbstractSourceItemTransformerFactory
{

	private static Item[] PARENT_TO_CHILD = 
		{ Item.BIOSOURCE, Item.SAMPLE, Item.EXTRACT };
	
	private static Item[] CHILD_TO_PARENT =
		{ Item.DERIVEDBIOASSAY, Item.RAWBIOASSAY };
	
	
	public ToPhysicalBioAssaySourceItemTransformerFactory() 
	{
		super(Item.PHYSICALBIOASSAY, PARENT_TO_CHILD, CHILD_TO_PARENT);
	}
	
	
	@Override
	public SourceItemTransformer create(final Item sourceItemType, final SourceItemTransform transform) 
	{
		List<SourceItemTransformer> chain = new ArrayList<SourceItemTransformer>();
		Item stepBySource = null;
		
		if (transform == SourceItemTransform.PARENT_TO_CHILD)
		{
			// Transforming a list of parent items to child physical bioassays
			// We do step-wise loading from BIOSOURCE -> SAMPLE -> EXTRACT -> PHYSICALBIOASSAY
			if (sourceItemType == Item.BIOSOURCE)
			{
				// Load child samples of the biosources
				chain.add(new BioSourceToSampleTransformer());
				stepBySource = Item.SAMPLE;
			}
			if (sourceItemType == Item.SAMPLE || stepBySource == Item.SAMPLE)
			{
				// Expand source selection to parents if the children push annotations upwards
				if (sourceItemType == Item.SAMPLE) chain.add(new PushToParentsTransformer(sourceItemType));
				// Load child samples including the original samples
				chain.add(new SampleToChildSampleTransformer(true));
				// Transform to child extracts
				chain.add(new SampleToExtractTransformer());
				stepBySource = Item.EXTRACT;
			}
			if (sourceItemType == Item.EXTRACT || stepBySource == Item.EXTRACT)
			{
				// Expand source selection to parents if the children push annotations upwards
				if (sourceItemType == Item.EXTRACT) chain.add(new PushToParentsTransformer(sourceItemType));
				// Load child extract and include the source extracts
				chain.add(new ExtractToChildExtractTransformer(true));
				// Load all related physical bioassays
				chain.add(new ExtractToPhysicalBioAssayTransformer());
			}
		}
		else if (transform == SourceItemTransform.CHILD_TO_PARENT)
		{
			// Transforming a list of child items to parent physical bioassays
			// We do step-wise loading from RAWBIOASSAY -> DERIVEDBIOASSAY -> PHYSICALBIOASSAY
			if (sourceItemType == Item.RAWBIOASSAY)
			{
				// Load parent derived bioassays
				chain.add(new RawBioAssayToDerivedBioAssayTransformer());
				stepBySource = Item.DERIVEDBIOASSAY;
			}
			if (sourceItemType == Item.DERIVEDBIOASSAY || stepBySource == Item.DERIVEDBIOASSAY)
			{
				// Load parent physical bioassays
				chain.add(new DerivedBioAssayToPhysicalBioAssayTransformer());
			}
		}

		
		return chain.size() == 0 ? null : new SourceItemTransformerChain(chain);
	}

}
