/**
	$Id$

	Copyright (C) 2020 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Subtypable;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for files. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_FILE}
	<li>Access denied: {@link Validator#DENIED_FILE}
	<li>Incorrect file type: {@link Validator#INCORRECT_FILETYPE} 
	</ul>

	@author Nicklas
	@since 3.17
*/
public class FileValidator
	extends NameableNodeValidator<File>
{
	
	public FileValidator()
	{
		super(Validator.MISSING_FILE, Validator.DENIED_FILE);
	}

	
	/* 
		From BasicValidator class
		-------------------------
	*/
	/**
		If the parent item has a subtype that is related to a FILE subtype, report
		the missing item as a {@link Validator#MISSING_FILE} failure. Otherwise,
		ignore the missing file.
		@return Always false
	*/
	@Override
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		BasicItem parentItem = parentNode.getItem(dc);
		ItemSubtype expectedSubtype = getExpectedFileSubtype(dc, parentItem);
		if (expectedSubtype != null)
		{
			context.createFailure(Validator.MISSING_FILE, parentNode, 
					"Missing file: " + expectedSubtype.getName(), getMissingItemFix(dc, parentNode));
		}
		return false;
	}


	/**
		Checks if the file is of the correct file type.
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		File file = (File)node.getItem(dc);
		
		// Validate the subtype of the software
		BasicItem parentItem = parentNode.getItem(dc);
		ItemSubtype expectedSubtype = getExpectedFileSubtype(dc, parentItem);
		try
		{
			ItemSubtype subtype = file.getItemSubtype();
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				Fix subtypeFix = null;
				if (subtype != null && parentItem instanceof Subtypable)
				{
					ItemSubtype mySubtype = ((Subtypable)parentItem).getItemSubtype();
					if (mySubtype != null)
					{
						subtypeFix = new Fix("Change related file subtype of '" + mySubtype.getName() + 
								"' to '" + subtype.getName() + "'", mySubtype);
					}
				}

				context.createFailure(Validator.INCORRECT_FILETYPE, parentNode, 
					"Expected a file with subtype: " + expectedSubtype.getName(),
					new Fix("Change file of '" + parentNode.getTitle() + "'", parentItem),
					new Fix("Change file type of '" + file.getName() + "' to '" + expectedSubtype.getName() + "'", file),
					subtypeFix
				);
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}

	/**
		@return Suggested fix is to add a file to the parent item
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		BasicItem parentItem = parentNode.getItem();		
		return new Fix("Add file to '" + parentNode.getTitle() + "'", parentItem);
	}
	// --------------------------------
	
	/**
		Get the subtype of the associated file that we expect for the given
		parent item.
	*/
	protected ItemSubtype getExpectedFileSubtype(DbControl dc, BasicItem parentItem)
	{
		ItemSubtype expectedSubtype = null;
		try
		{
			if (parentItem instanceof Subtypable)
			{
				expectedSubtype = ItemSubtype.getRelatedSubtype(dc, 
						(Subtypable)parentItem, Item.FILE, 0);
			}
		}
		catch (PermissionDeniedException ex)
		{}
		return expectedSubtype;
	}
	
}
