/**
	$Id: RawBioAssayValidator.java 6959 2015-10-01 09:37:58Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.List;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.RootRawBioAssay;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.extensions.ValidationRuleAction;

/**
	Validator implementation for raw bioassays and root raw bioassays. 
	Validation rules:
	<ul>
	<li>Non-default raw data type: {@link Validator#NONDEFAULT_RAWDATATYPE}
	<li>Parent extract mismatch:  {@link Validator#PARENT_EXTRACT_MISMATCH}
	<li>The number of raw data spots is different from the number of
		features on the array design: {@link Validator#NONMATCHING_SPOTCOUNT}.
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2015-10-01 11:37:58 +0200 (Thu, 01 Oct 2015) $
*/
public class RawBioAssayValidator
	extends NameableNodeValidator<Nameable>
{
	
	public RawBioAssayValidator()
	{
		super(null, null);
	}
	
	/* 
		From BasicValidator class
		-------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		
		// Check raw data type against default raw data type
		Project project = context.getProject();
		BasicItem item = node.getItem(dc);
		if (item instanceof RootRawBioAssay)
		{
			// The parent may be a root raw bioasay if the overview starts at the
			// experiment level
			item = ((RootRawBioAssay)item).getRawBioAssay();
		}
		RawBioAssay rba = (RawBioAssay)item;
		if (project != null)
		{
			project = Project.getById(dc, project.getId());
			RawDataType projectDefault = project.getDefaultRawDataType();
			if (projectDefault != null && rba.getRawDataType() != projectDefault)
			{
				context.createFailure(Validator.NONDEFAULT_RAWDATATYPE, node, null,
						new Fix("Change default raw data type of project", project)
					);
			}
		}
		
		// Check that parent items make sense
		Extract parentExtract = null;
		DerivedBioAssay parentBioAssay = null;
		Extract parentBioAssayExtract = null;
		List<PhysicalBioAssay> physicalBioAssays = null;
		try
		{
			parentBioAssay = rba.getParentBioAssay();
			parentExtract = rba.getParentExtract();
			if (parentBioAssay != null && parentExtract != null) 
			{
				physicalBioAssays = PhysicalBioAssayValidator.getPhysicalBioAssays(dc, context, parentBioAssay);
				parentBioAssayExtract = parentBioAssay.getExtract();
			}
		}
		catch (PermissionDeniedException ex)
		{}
		
		// Check that the extract is found on among the parents for at least on of the physical bioassays
		if (parentExtract != null && physicalBioAssays != null && physicalBioAssays.size() > 0)
		{
			boolean found = false;
			for (PhysicalBioAssay pba : physicalBioAssays)
			{
				if (PhysicalBioAssayValidator.getAllParentExtracts(dc, context, pba).contains(parentExtract))
				{
					found = true;
					break;
				}
			}
			
			if (!found)
			{
				if (physicalBioAssays.size() == 1)
				{
					PhysicalBioAssay pba = physicalBioAssays.get(0);
					context.createFailure(Validator.PARENT_EXTRACT_MISMATCH, node, 
						"The extract '" + parentExtract.getName() + "' is not a parent to the physical bioassay: " + 
							pba.getName(),
						new Fix("Change extract or parent bioassay on raw bioassay", rba),
						new Fix("Change extracts on physical bioassay", pba)
					);
				}
				else
				{
					context.createFailure(Validator.PARENT_EXTRACT_MISMATCH, node, 
							"The extract '" + parentExtract.getName() + "' is not a parent to any of the physical bioassays",
							new Fix("Change extract or parent bioassay on raw bioassay", rba)
						);
				}
			}
		}
		
		// Check that extract is related to the extract on the parent derived bioassay
		if (parentExtract != null && parentBioAssayExtract != null)
		{
			if (!ExtractValidator.isRelated(dc, context, parentExtract, parentBioAssayExtract) && !ExtractValidator.isRelated(dc, context, parentBioAssayExtract, parentExtract))
			{
				context.createFailure(Validator.PARENT_EXTRACT_MISMATCH, node, 
					"The parent bioassay '" + parentBioAssay.getName() + "' has unrelated extract: " + 
						parentBioAssayExtract.getName(), 
					new Fix("Change extract on raw bioassay", rba),
					new Fix("Change extract on parent bioassay", parentBioAssay)
				);
			}
		}

				
		if (rba.hasArrayDesign() && rba.hasData())
		{
			try
			{
				ArrayDesign design = rba.getArrayDesign();
				String threshold = context.getValidationOptions().getOption(
						(ValidationRuleAction)Validator.NONMATCHING_SPOTCOUNT, "threshold", "5%");
				
				int numDbSpots = rba.getNumDbSpots();
				int numDbFeatures = design.getNumDbFeatures();
				int numFileSpots = rba.getNumFileSpots();
				int numFileFeatures = design.getNumFileFeatures();
				int maxDbDelta = 0;
				int maxFileDelta = 0;
				if (threshold.endsWith("%"))
				{
					int percentage = Values.getInt(threshold.substring(0, threshold.length()-1));
					if (numDbFeatures > 0) maxDbDelta = (numDbFeatures * percentage) / 100;
					if (numFileFeatures > 0) maxFileDelta = (numFileFeatures * percentage) / 100;
				}
				else
				{
					maxDbDelta = Values.getInt(threshold);
					maxFileDelta = maxDbDelta;
				}
				
				
				if (numDbSpots > 0 && Math.abs(numDbFeatures - numDbSpots) > maxDbDelta)
				{
					context.createFailure(Validator.NONMATCHING_SPOTCOUNT, node, 
						"Raw spots (db) <> features by more than " + threshold + ": " + 
						numDbSpots  + "<>" + numDbFeatures);
				}
				if (numFileSpots > 0 && Math.abs(numFileFeatures - numFileSpots) > maxFileDelta)
				{
					context.createFailure(Validator.NONMATCHING_SPOTCOUNT, node, 
						"Raw spots (file) <> features by more than " + threshold + ": " + 
						numFileSpots  + "<>" + numFileFeatures);
				}
				
			}
			catch (RuntimeException ex)
			{}
		}
	}

	/**
		@return Always null
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		return null;
	}
	// ----------------------------
}
