/*
	$Id: Mapper.java 7703 2019-04-11 07:31:26Z nicklas $
	
	Copyright (C) 2006 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.
	
	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.parser;

import java.util.Date;

import net.sf.basedb.core.Type;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.parser.FlatFileParser.Data;

/**
	Represents a mapping expression that gets a single value from
 	a line of data. A typical mapping expression
	holds a reference to the column number or name, for example:
	<pre class="code">
\1\
\row\
Row: \row\, Col:\col\
=2 * col('Radius')
</pre>

	The actual syntax for the mapping string depends on the parser used.

	@author Nicklas
 	@version 2.0
 	@see net.sf.basedb.util.parser.FlatFileParser#getMapper(String)
	@base.modified $Date: 2019-04-11 09:31:26 +0200 (Thu, 11 Apr 2019) $
*/
public interface Mapper
{
	/**
		Get the value for the line of data.
		@param data The line of data
		@return A string value
		@deprecated In 3.15, use {@link #getString(FlatFileParser.Data)} instead or
			one of the other getXxxx() methods to get a given data type directly 
	*/
	@Deprecated
	public String getValue(Data data);
	
	/**
		Get the value for the line of data as a string.
		@param data The line of data
		@return A string value
		@since 3.15
	*/
	public String getString(Data data);
	
	/**
		Get the value for the line of data as an integer.

		@param data The line of data
		@return The value as an integer
		@throws NumberFormatException If the value can't be parsed as an integer
	*/
	public Integer getInt(Data data);

	/**
		Get the value for the line of data as a long.
	
		@param data The line of data
		@return The value as an long
		@throws NumberFormatException If the value can't be parsed as an long
		@since 3.15
	*/
	public Long getLong(Data data);
	
	/**
		Get the value for the line of data as a float.
	
		@param data The line of data
		@return The value as a float
		@throws NumberFormatException If the value can't be parsed as a float
	*/
	public Float getFloat(Data data);

	/**
		Get the value for the line of data as a double.
	
		@param data The line of data
		@return The value as a double
		@throws NumberFormatException If the value can't be parsed as a double
		@since 3.15
	*/
	public Double getDouble(Data data);

	/**
		Get the value for the line of data as a date.
	
		@param data The line of data
		@return The value as a date
		@since 3.15
	*/
	public Date getDate(Data data);

	@SuppressWarnings("unchecked")
	public default <T> T getByType(Data data, Type type)
	{
		T value = null;
		if (type.isNumerical())
		{
			if (type == Type.DOUBLE) value = (T)getDouble(data);
			else if (type == Type.FLOAT) value = (T)getFloat(data);
			else if (type == Type.INT) value = (T)getInt(data);
			else if (type == Type.LONG) value = (T)getLong(data);
		}
		else if (type.isTemporal())
		{
			value = (T)getDate(data);
		}
		else if (type == Type.BOOLEAN)
		{
			value = (T)(Boolean)Values.getBoolean(getString(data));
		}
		else
		{
			value = (T)getString(data);
		}
		return value;
	}
	
}
