/**
	$Id: PlateMappingImporter.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PlateGeometry;
import net.sf.basedb.core.PlateMapping;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.data.MappingCoordinate;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;
import net.sf.basedb.util.parser.FlatFileParser.Data;

/**
	Importer for plate mappings. It has a fixed file format:

	<pre class="code">
Source plates=1
Source rows=3
Source columns=3
Destination plates=1
Destination rows=3
Destination columns=3
Source plate	Source row	Source column	Destination plate	Destination row	Destination column
0	0	0	0	0	0
0	0	1	0	1	0
0	0	2	0	2	0
0	1	0	0	0	1
0	1	1	0	1	1
0	1	2	0	2	1
0	2	0	0	0	2
0	2	1	0	1	2
0	2	2	0	2	2
</pre>

	 @author nicklas
	 @version 2.0
	 @base.modified $Date: 2012-09-14 09:28:35 +0200 (Fri, 14 Sep 2012) $
*/
public class PlateMappingImporter
	extends AbstractFlatFileImporter
	implements InteractivePlugin
{
	
	private static final Set<GuiContext> guiContexts = 
		Collections.singleton(new GuiContext(Item.PLATEMAPPING, GuiContext.Type.LIST));
	
	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	private static final StringParameterType requiredStringParameter = 
		new StringParameterType(PlateMapping.MAX_NAME_LENGTH, null, true);

	private RequestInformation configureJob;

	private static final PluginParameter<String> nameParameter = new PluginParameter<String>(
		"name",
		"Name",
		"The name of the new plate mapping.",
		requiredStringParameter
	);
	
	private static final PluginParameter<String> descriptionParameter = new PluginParameter<String>(
		"description",
		"Description",
		"The description of the new plate mapping.",
		new StringParameterType(PlateMapping.MAX_DESCRIPTION_LENGTH, null, false)
	);


	/**
		Create a new importer.
	*/
	public PlateMappingImporter()
	{}

	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request create and write access to PlateMapping:s, create and use
	 	access to PlateGeometries and read access to File:s.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.PLATEMAPPING, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.PLATEGEOMETRY, EnumSet.of(Permission.USE), EnumSet.of(Permission.CREATE)));
		}
		return permissions;
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	/**
		Return a set containing the context:s [PLATEMAPPING, LIST].
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		Always null.
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		return null;
	}
	/**
		Store configuration settings for {@link Request#COMMAND_CONFIGURE_PLUGIN} and
		{@link Request#COMMAND_CONFIGURE_JOB}.
	*/
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				List<Throwable> errors = validateRequestParameters(getConfigureJobParameters(context).getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}

				storeValue(job, request, nameParameter);
				storeValue(job, request, descriptionParameter);
				storeValue(job, request, fileParameter);

				File file = (File)job.getValue("file");
				response.setSuggestedJobName("Import plate mapping '" + job.getValue("name") + 
						"' from file '" + file.getName() + "'");

				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}


	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command)
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
		}
		return requestInformation;
	}
	// -------------------------------------------
	/*
		From the AbstractFlatFileReporter class
		-------------------------------------------
	*/
	private DbControl dc;
	private PlateMapping pm;
	private Mapper sourcePlateMapper;
	private Mapper sourceRowMapper;
	private Mapper sourceColumnMapper;
	private Mapper destPlateMapper;
	private Mapper destRowMapper;
	private Mapper destColumnMapper;
	private FlatFileParser ffp;
	
	/**
		Save reference to parser.
	*/
	@Override
	protected void begin(FlatFileParser ffp)
		throws BaseException
	{
		super.begin(ffp);
		this.ffp = ffp;
	}
	
	/**
	 	Create a {@link DbControl} and a new {@link PlateMapping}.
		Initialise column <code>Mapper</code>:s.
	*/
	@Override
	protected void beginData()
	{
		dc = sc.newDbControl();
		String name = (String)job.getValue("name");
		String description = (String)job.getValue("description");
		
		int sourcePlates = Integer.parseInt(ffp.getHeader("Source plates"));
		int sourceRows = Integer.parseInt(ffp.getHeader("Source rows"));
		int sourceColumns = Integer.parseInt(ffp.getHeader("Source columns"));

		int destPlates = Integer.parseInt(ffp.getHeader("Destination plates"));
		int destRows = Integer.parseInt(ffp.getHeader("Destination rows"));
		int destColumns = Integer.parseInt(ffp.getHeader("Destination columns"));
		
		sourcePlateMapper = ffp.getMapper("\\Source plate\\");
		sourceRowMapper = ffp.getMapper("\\Source row\\");
		sourceColumnMapper = ffp.getMapper("\\Source column\\");
		destPlateMapper = ffp.getMapper("\\Destination plate\\");
		destRowMapper = ffp.getMapper("\\Destination row\\");
		destColumnMapper = ffp.getMapper("\\Destination column\\");
		
		PlateGeometry sourceGeometry = findGeometry(dc, sourceRows, sourceColumns);
		PlateGeometry destGeometry = sourceRows == destRows && sourceColumns == destColumns ?
			sourceGeometry : findGeometry(dc, destRows, destColumns);
		
		pm = PlateMapping.getNew(dc, sourceGeometry, sourcePlates, destGeometry, destPlates);
		pm.setName(name);
		pm.setDescription(description);
		dc.saveItem(pm);

	}

	/**
		Add a mapping detail.
	*/
	@Override
	protected void handleData(Data data) 
		throws BaseException
	{
		Integer sourcePlate = sourcePlateMapper.getInt(data);
		Integer sourceRow = sourceRowMapper.getInt(data);
		Integer sourceColumn = sourceColumnMapper.getInt(data);

		Integer destPlate = destPlateMapper.getInt(data);
		Integer destRow = destRowMapper.getInt(data);
		Integer destColumn = destColumnMapper.getInt(data);
		MappingCoordinate source = new MappingCoordinate(sourcePlate, sourceRow, sourceColumn);
		MappingCoordinate destination = new MappingCoordinate(destPlate, destRow, destColumn);
		pm.setSourceCoordinate(destination, source);
	}
	
	/**
		Close and commit/rollback the DbControl.
	*/
	@Override
	protected void end(boolean success)
		throws BaseException
	{
		try
		{
			if (success)
			{
				dc.commit();
			}
			else
			{
				dc.close();
			}
		}
		catch (BaseException ex)
		{
			dc.close();
			throw ex;
		}
		finally
		{
			super.end(success);
		}
	}
	/**
		Return <code>Platemapping 'name' created.</code>.
	*/
	@Override
	protected String getSuccessMessage(int skippedLines)
	{
		return "Platemapping " + pm.getName() + " created.";
	}
	
	@Override
	protected FlatFileParser getInitializedFlatFileParser()
		throws BaseException
	{
		FlatFileParser ffp = new FlatFileParser();
		ffp.setHeaderRegexp(Pattern.compile("(.+)=(.*)"));
		ffp.setDataHeaderRegexp(Pattern.compile("Source plate\\tSource row\\tSource column\\tDestination plate\\tDestination row\\tDestination column"));
		ffp.setDataSplitterRegexp(Pattern.compile("\\t"));
		ffp.setMinDataColumns(6);
		ffp.setMaxDataColumns(6);
		return ffp;
	}

	// -------------------------------------------

	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob == null)
		{
			// RequestInformation for CONFIGURE_JOB
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			
			parameters.add(nameParameter);
			parameters.add(descriptionParameter);
			parameters.add(fileParameter);
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Select a file to import plate mapping information from",
				"Here you specify the name and description of the new mapping and which file " +
				"to import the mappings details from.",
				parameters
			);
		}
		return configureJob;
	}

	
	private PlateGeometry findGeometry(DbControl dc, int rows, int columns)
	{
		ItemQuery<PlateGeometry> query = PlateGeometry.getQuery();
		query.restrict(
			Restrictions.and(
				Restrictions.eq(Hql.property("rows"), Expressions.integer(rows)),
				Restrictions.eq(Hql.property("columns"), Expressions.integer(columns))
			)
		);
		ItemResultList<PlateGeometry> result = query.list(dc);
		PlateGeometry pg = null;
		if (result.size() > 0)
		{
			pg = result.get(0);
		}
		else
		{
			pg = PlateGeometry.getNew(dc, rows, columns);
			int wells = rows * columns;
			pg.setName(wells + "-well (" + rows + " x " + columns + ")");
			pg.setDescription("Added by PlateMappingFlatFileImporter.");
			dc.saveItem(pg);
		}
		return pg;
	}
}
