/* $Id $
	------------------------------------------------------------------
	Copyright (C) 2020 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	@author Nicklas
*/
'use strict';

var Configure = function()
{
	var configure = {};
	
	var currentTargetSubtypes;
	var currentTargetAnnotationTypesAll;
	var currentTargetAnnotationTypesCategory;
	var currentTargetExtensionColumns;
	var currentTargetItemLists;
	 
	var currentChildItemTypes;
	var currentChildSubtypes;
	var currentChildItemLists;
	var currentChildAnnotationTypesAll;
	var currentChildAnnotationTypesCategory;
	var currentChildExtensionColumns;
	
	// See also net.sf.basedb.clients.web.extensions.list.RelatedItemColumn.pathTitles
	var nameCol = {'id': '.', 'text': 'Name'};
	var externalIdCol = {'id': 'externalId', 'text': 'External ID'};
	var bioPlateCol = {'id': 'bioWell.bioPlate', 'text': 'Bioplate'};
	var entryDateCol = {'id': 'entryDate', 'text': 'Registered'};
	var remainingQCol = {'id': 'remainingQuantity', 'text': 'Remaining quantity'};
	var originalQCol = {'id': 'originalQuantity', 'text': 'Original quantity'};
	var itemListCol = {'id': '§itemLists', 'text': 'Item list'};
	
	var creationRegistredCol = {'id': 'creationEvent.entryDate', 'text': 'Registration date'};
	var creationEventdateCol = {'id': 'creationEvent.eventDate', 'text': 'Creation date'};
	var creationProtocolCol = {'id': 'creationEvent.protocol', 'text': 'Protocol'};
	var creationHardwareCol = {'id': 'creationEvent.hardware', 'text': 'Hardware'};

	var protocolCol = {'id': 'protocol', 'text': 'Protocol'};
	var hardwareCol = {'id': 'hardware', 'text': 'Hardware'};
	var softwareCol = {'id': 'software', 'text': 'Software'};
	var tagCol = {'id': 'tag', 'text': 'Tag'};
	var jobTypeCol = {'id': 'job.itemSubtype', 'text': 'Job type'};

	var staticColumns = {
		'BIOSOURCE': [nameCol, externalIdCol, itemListCol, entryDateCol],
		'SAMPLE': [nameCol, externalIdCol, itemListCol, bioPlateCol, creationEventdateCol, creationRegistredCol, creationProtocolCol, remainingQCol, originalQCol],
		'EXTRACT': [nameCol, externalIdCol, itemListCol, bioPlateCol, creationEventdateCol, creationRegistredCol, creationProtocolCol, remainingQCol, originalQCol, tagCol],
		'PHYSICALBIOASSAY': [nameCol, itemListCol, creationEventdateCol, creationRegistredCol, creationProtocolCol, creationHardwareCol],
		'DERIVEDBIOASSAY': [nameCol, itemListCol, entryDateCol, protocolCol, softwareCol, hardwareCol, jobTypeCol],
		'RAWBIOASSAY': [nameCol, itemListCol, entryDateCol, protocolCol, softwareCol, jobTypeCol]
	}
	
	/**
		Initialize the page.
	*/
	configure.initPage = function()
	{
		Events.addEventHandler('targetItemType', 'change', configure.targetItemTypeOnChange);
		Events.addEventHandler('targetSubtype', 'change', configure.targetSubtypeOnChange);
		Events.addEventHandler('childItemType', 'change', configure.childItemTypeOnChange);
		Events.addEventHandler('childSubtype', 'change', configure.childSubtypeOnChange);
		Events.addEventHandler('extraFilter', 'change', configure.extraFilterOnChange);
		Events.addEventHandler('targetAnnotationValue', 'change', configure.targetAnnotationValueOnChange);
		Events.addEventHandler('showAllAnnotationTypes', 'change', configure.updateColumnsList);
		Events.addEventHandler(document.body, 'click', configure.hideMessage);
		
		// Buttons
		Buttons.addClickHandler('close', App.closeWindow);
		Buttons.addClickHandler('btnAdd', configure.addColumn);
	}
	
	// Clear existing columns
	configure.clearColumns = function()
	{
		configure.clearList('extraFilter');
		configure.clearList('targetAnnotationValue');
		Doc.element('columns').innerHTML = '';
		Doc.element('annotations').innerHTML = '';
		Doc.element('extensions').innerHTML = '';
		Doc.hide('targetPath');
		Doc.hide('childPath');
	}
	// Clear the list and disable it
	configure.clearList = function(listOrId)
	{
		var list = Doc.element(listOrId);
		list.disabled = true;
		list.length = 1;
	}
	
	configure.targetItemTypeOnChange = function(event)
	{
		// Reset everything that depends on the target item type
		configure.clearColumns();
		configure.clearList('targetSubtype');
		configure.clearList('childItemType');
		configure.clearList('childSubtype');
		Doc.element('allowDoublingBack').disabled = true;
		
		var frm = document.forms['relatedItems'];
		var sourceItemType = frm.item_type.value;
		var targetItemType = frm.targetItemType.value;
		var direction = Data.get(frm.targetItemType[frm.targetItemType.selectedIndex], 'direction');
		
		if (targetItemType)
		{
			var url = 'ajax.jsp?ID='+App.getSessionId();
			url += '&cmd=GetSubtypesAndAnnotationTypes&itemType='+encodeURIComponent(targetItemType);
			url += '&sourceItemType='+encodeURIComponent(sourceItemType);
			url += '&direction='+encodeURIComponent(direction);
			
			var request = Ajax.getXmlHttpRequest();
			request.open("GET", url, true);
			request.send(null);
			Ajax.setReadyStateHandler(request, configure.targetTypeInfoLoaded, configure.targetTypeInfoLoaded);
		}
	}
	
	configure.targetSubtypeOnChange = function(event)
	{
		// Reset everything that depends on the target subtype
		configure.clearColumns();
		
		var frm = document.forms['relatedItems'];
		var targetItemType = frm.targetItemType.value;
		var targetSubtype = frm.targetSubtype.value;
		var sourceItemType = frm.item_type.value;
		var category = frm.targetSubtype.selectedIndex > 0 ? frm.targetSubtype[frm.targetSubtype.selectedIndex].text : null;
		if (targetItemType && category)
		{
			var url = 'ajax.jsp?ID='+App.getSessionId();
			url += '&cmd=GetAnnotationTypesForCategory&itemType='+encodeURIComponent(targetItemType);
			url += '&subtype='+encodeURIComponent(targetSubtype);
			url += '&sourceItemType='+encodeURIComponent(sourceItemType);
			url += '&category='+encodeURIComponent(category);
			var request = Ajax.getXmlHttpRequest();
			request.open("GET", url, true);
			request.send(null);
			Ajax.setReadyStateHandler(request, configure.targetTypeInfoLoaded, configure.targetTypeInfoLoaded);
		}
	}
	
	configure.targetTypeInfoLoaded = function(request)
	{
		var response = JSON.parse(request.responseText);
		if (response.status != 'ok')
		{
			App.debug(request.responseText);
			return;
		}
		var frm = document.forms['relatedItems'];
		
		if (response.childTypes)
		{
			frm.childItemType.length = 1;
			currentChildItemTypes = response.childTypes;
			for (var i = 0; i < currentChildItemTypes.length; i++)
			{
				var st = currentChildItemTypes[i];
				frm.childItemType[frm.childItemType.length] = new Option('Child: '+st.name, st.id);
			}
			frm.childItemType.selectedIndex = 0;
			frm.childItemType.disabled = currentChildItemTypes.length == 0;
		}
		
		if (response.subtypes) 
		{
			currentTargetSubtypes = response.subtypes;
			frm.targetSubtype.length = 1;
			for (var i = 0; i < currentTargetSubtypes.length; i++)
			{
				var st = currentTargetSubtypes[i];
				frm.targetSubtype[frm.targetSubtype.length] = new Option(st.name, st.id);
			}
			frm.targetSubtype.selectedIndex = 0;
			frm.targetSubtype.disabled = currentTargetSubtypes.length == 0;
		}
		
		if (response.annotationTypes) currentTargetAnnotationTypesAll = response.annotationTypes;
		if (response.annotationTypesCategory) currentTargetAnnotationTypesCategory = response.annotationTypesCategory;
		if (response.extensionColumns) currentTargetExtensionColumns = response.extensionColumns;
		if (response.itemLists) currentTargetItemLists = response.itemLists;

		configure.updateColumnsList();
	}
	
	configure.childItemTypeOnChange = function(event)
	{
		// Reset everything that depends on the child item type
		configure.clearColumns();
		configure.clearList('childSubtype');
		Doc.element('allowDoublingBack').disabled = true;
		
		var frm = document.forms['relatedItems'];
		var targetItemType = frm.targetItemType.value;
		var childItemType = frm.childItemType.value;
		
		if (targetItemType && childItemType)
		{
			var url = 'ajax.jsp?ID='+App.getSessionId();
			url += '&cmd=GetSubtypesAndAnnotationTypes&itemType='+encodeURIComponent(childItemType);
			url += '&sourceItemType='+encodeURIComponent(targetItemType);
			url += '&direction=CHILD';
			
			var request = Ajax.getXmlHttpRequest();
			request.open("GET", url, true);
			request.send(null);
			Ajax.setReadyStateHandler(request, configure.childTypeInfoLoaded, configure.childTypeInfoLoaded);
		}
		else
		{
			configure.updateColumnsList();
		}
	}

	configure.childSubtypeOnChange = function(event)
	{
		// Reset everything that depends on the child subtype
		configure.clearColumns();

		var frm = document.forms['relatedItems'];
		var targetItemType = frm.targetItemType.value;
		var childItemType = frm.childItemType.value;
		var childSubtype = frm.childSubtype.value;
		var category = frm.childSubtype.selectedIndex > 0 ? frm.childSubtype[frm.childSubtype.selectedIndex].text : null;
		if (targetItemType && childItemType && category)
		{
			var url = 'ajax.jsp?ID='+App.getSessionId();
			url += '&cmd=GetAnnotationTypesForCategory&itemType='+encodeURIComponent(childItemType);
			url += '&subtype='+encodeURIComponent(childSubtype);
			url += '&sourceItemType='+encodeURIComponent(targetItemType);
			url += '&category='+encodeURIComponent(category);
			
			var request = Ajax.getXmlHttpRequest();
			request.open("GET", url, true);
			request.send(null);
			Ajax.setReadyStateHandler(request, configure.childTypeInfoLoaded, configure.childTypeInfoLoaded);
		}
	}

	configure.childTypeInfoLoaded = function(request)
	{
		var response = JSON.parse(request.responseText);
		if (response.status != 'ok')
		{
			App.debug(request.responseText);
			return;
		}
		var frm = document.forms['relatedItems'];

		if (response.subtypes) 
		{
			currentChildSubtypes = response.subtypes;
			frm.childSubtype.length = 1;
			for (var i = 0; i < currentChildSubtypes.length; i++)
			{
				var st = currentChildSubtypes[i];
				frm.childSubtype[frm.childSubtype.length] = new Option(st.name, st.id);
			}
			frm.childSubtype.selectedIndex = 0;
			frm.childSubtype.disabled = currentChildSubtypes.length == 0;
			frm.allowDoublingBack.disabled = currentChildSubtypes.length == 0;
		}	
		
		if (response.annotationTypes) currentChildAnnotationTypesAll = response.annotationTypes;
		if (response.annotationTypesCategory) currentChildAnnotationTypesCategory = response.annotationTypesCategory;
		if (response.extensionColumns) currentChildExtensionColumns = response.extensionColumns;
		if (response.itemLists) currentChildItemLists = response.itemLists;

		configure.updateColumnsList();
	}

	configure.updateColumnsList = function()
	{
		var frm = document.forms['relatedItems'];
		
		var targetItemType = frm.targetItemType.value;
		var targetSubtype = frm.targetSubtype.value;
		var childItemType = frm.childItemType.value;
		var childSubtype = frm.childSubtype.value;

		// At least a target item type and a subtype must be selected
		if (!targetItemType || !targetSubtype) return;
		// If a child item type is selected, so must a child subtype
		if (childItemType && !childSubtype) return;
		
		var staticCols = staticColumns[targetItemType] || [];
		var annotationCols = frm.showAllAnnotationTypes.checked ? currentTargetAnnotationTypesAll : currentTargetAnnotationTypesCategory;
		var xtCols = currentTargetExtensionColumns;
		var itemLists = currentTargetItemLists;
		
		if (childItemType && childSubtype)
		{
			staticCols = staticColumns[childItemType] || [];
			annotationCols = frm.showAllAnnotationTypes.checked ? currentChildAnnotationTypesAll : currentChildAnnotationTypesCategory;
			xtCols = currentChildExtensionColumns;
			itemLists = currentChildItemLists;
		}		
		
		var html = '';
		// Add static property columns depending on itemtype
		for (var i = 0; i < staticCols.length; i++)
		{
			var c = staticCols[i];
			var text = Strings.encodeTags(c.text);
			html += '<div><label><input type="checkbox" data-column-id="'+c.id+'" data-column-title="'+text+'">';
			html += text + '</label></div>';
		}
		Doc.element('columns').innerHTML = html;

		// Add annotation columns
		html = '';
		var extraFilter = Doc.element('extraFilter');
		var selectedOption = extraFilter[extraFilter.selectedIndex];
		extraFilter.length = 1;
		for (var i = 0; i < annotationCols.length; i++)
		{
			var at = annotationCols[i];
			var text = Strings.encodeTags(at.name);
			html += '<div><label><input type="checkbox" data-column-id="#+'+at.id+'" data-column-title="'+text+'">';
			html += text + '</label></div>';
			if (at.enumeration)
			{
				var selected = selectedOption.annotationType && selectedOption.value == at.id;
				var opt = new Option(at.name, at.id, selected, selected);
				opt.annotationType = at;
				extraFilter[extraFilter.length] = opt;
			}
		}
		if (itemLists.length > 0)
		{
			var opt = new Option('-- item lists --', '');
			opt.disabled = true;
			extraFilter[extraFilter.length] = opt;
			for (var i = 0; i < itemLists.length; i++)
			{
				var itemList = itemLists[i];
				var selected = selectedOption.preset && selectedOption.value == itemList.id;
				var opt = new Option(itemList.name + ' ('+itemList.size+')', itemList.id, selected, selected);
				opt.itemList = itemList;
				extraFilter[extraFilter.length] = opt;
			}
		}
		Doc.element('annotations').innerHTML = html;
		extraFilter.disabled = extraFilter.length < 2;
		if (!extraFilter[extraFilter.selectedIndex].annotationType) 
		{
			configure.clearList('targetAnnotationValue');
		}
		
		// Extension columns
		html = '';
		for (var i = 0; i < xtCols.length; i++)
		{
			var xtCol = xtCols[i];
			var id = "!"+xtCol.id+"#"+xtCol.columnId;
			var text = Strings.encodeTags(xtCol.name);
			if (xtCol.id.indexOf('net.sf.basedb.clients.web.related-item-columns') == 0)
			{
				// Special case when the extension is the related-item extionsion
				// It has built-in support for two hops and doen't have to use the 
				// extension mechanism recursively
				id = xtCol.columnId;
			}
			html += '<div title="'+Strings.encodeTags(xtCol.tooltip)+'">';
			html += '<label><input type="checkbox" data-column-id="'+id+'" data-column-title="'+text+'">';
			html += (xtCol.nameHTML || text) + '</label></div>';
		}
		Doc.element('extensions').innerHTML = html;
		
		configure.updateTargetPath();
	}
	
	configure.extraFilterOnChange = function(event)
	{
		var frm = document.forms['relatedItems'];
		configure.clearList('targetAnnotationValue');

		var annotationType = frm.extraFilter[frm.extraFilter.selectedIndex].annotationType;
		if (annotationType)
		{
			for (var i = 0; i < annotationType.enumeration.length; i++)
			{
				frm.targetAnnotationValue[frm.targetAnnotationValue.length] = new Option(annotationType.enumeration[i]);
			}
		}
		frm.targetAnnotationValue.disabled = frm.targetAnnotationValue.length < 2;
		configure.updateTargetPath();
	}
	
	configure.targetAnnotationValueOnChange = function(event)
	{
		configure.updateTargetPath();
	}
	
	configure.updateTargetPath = function()
	{
		var frm = document.forms['relatedItems'];
		
		var targetDirection = Data.get(frm.targetItemType[frm.targetItemType.selectedIndex], 'direction');
		var targetText = frm.targetSubtype[frm.targetSubtype.selectedIndex].text;
		var childText = frm.childSubtype.selectedIndex > 0 ? frm.childSubtype[frm.childSubtype.selectedIndex].text : null;
		var extraOption = frm.extraFilter[frm.extraFilter.selectedIndex];
		
		var filterText = '';
		if (extraOption.annotationType && frm.targetAnnotationValue.selectedIndex > 0)
		{
			var annotationType = extraOption.text;
			var annotationValue = frm.targetAnnotationValue.value;
			filterText = Strings.encodeTags(' & '+annotationType + '=' + annotationValue);
		}
		else if (extraOption.itemList)
		{
			filterText = Strings.encodeTags(' & '+extraOption.itemList.name);
		}

		if (childText)
		{
			Doc.element('childPath').innerHTML = 'Child('+childText+filterText+')';
			Doc.show('childPath', 'inline');
			filterText = '';
		}

		if (targetDirection == 'PARENT')
		{
			Doc.element('targetPath').innerHTML = 'Parent('+targetText+filterText+')';
			Doc.element('targetPath').className = 'parentitem';
		}
		else
		{
			Doc.element('targetPath').innerHTML = 'Child('+targetText+filterText+')';
			Doc.element('targetPath').className = 'childitem';
		}
		Doc.show('targetPath', 'inline');
	}
	
	configure.addColumn = function(event)
	{
		var frm = document.forms['relatedItems'];
		
		var data = {};
		data.targetItemType = frm.targetItemType.value;
		data.targetSubtype = frm.targetSubtype.value;
		data.childItemType = frm.childItemType.value;
		data.childSubtype = frm.childSubtype.value;

		var extraOption = frm.extraFilter[frm.extraFilter.selectedIndex];
		if (extraOption.annotationType)
		{
			data.targetAnnotationType = extraOption.value;
			data.targetAnnotationValue = frm.targetAnnotationValue.value;
		}
		else if (extraOption.itemList)
		{
			data.targetItemList = extraOption.value;
		}
		data.allowDoublingBack = frm.allowDoublingBack.checked?1:0;

		if (!data.targetItemType)
		{
			Forms.showNotification('targetItemType', 'Please select a parent or child item type!');
			return;
		}
		if (!data.targetSubtype)
		{
			Forms.showNotification('targetSubtype', 'Please select a subtype!');
			return;
		}
		if (data.childItemType && !data.childSubtype)
		{
			Forms.showNotification('childSubtype', 'Please select a subtype!');
			return;
		}
		if (data.targetAnnotationType && !data.targetAnnotationValue)
		{
			Forms.showNotification('targetAnnotationValue', 'Please select a value!');
			return;
		}
		
		data.targetSubtypeName = frm.targetSubtype[frm.targetSubtype.selectedIndex].text;
		data.direction = Data.get(frm.targetItemType[frm.targetItemType.selectedIndex], 'direction');
		
		var targetFilter = '';
		var childFilter = '';
		if (data.targetAnnotationType && data.targetAnnotationValue)
		{
			targetFilter = '&'+data.targetAnnotationType+'='+encodeURIComponent(data.targetAnnotationValue);
		}
		else if (data.targetItemList)
		{
			targetFilter = '&list='+data.targetItemList;
		}
		if (data.childItemType)
		{
			childFilter = targetFilter;
			targetFilter = '';
		}

		var propertyPrefix = '/'+data.direction+'/'+data.targetItemType+'/'+data.targetSubtype+targetFilter+'/';
		var titlePrefix = data.targetSubtypeName;
		if (data.childItemType)
		{
			data.childSubtypeName = frm.childSubtype[frm.childSubtype.selectedIndex].text;
			propertyPrefix += (data.allowDoublingBack?'/CHILD+/':'/CHILD/')+data.childItemType+'/'+data.childSubtype+childFilter+'/';
			titlePrefix += '.'+data.childSubtypeName;
		}
		if (data.targetAnnotationValue) titlePrefix += '['+data.targetAnnotationValue+']';

		var numSelected = 0;
		var allColumns = document.querySelectorAll("input[data-column-id]");
		for (var i = 0; i < allColumns.length; i++)
		{
			if (allColumns[i].checked)
			{
				var col = allColumns[i];
				data.column = Data.get(col, 'column-id');
				data.property = propertyPrefix+data.column;
				data.title = titlePrefix + (data.column == '.' ? '' : '.'+Data.get(col, 'column-title'));
				Events.sendCustomEvent(window.opener.Doc.element('selectRelatedItemColumn'), 'base-selected', data);
				numSelected++;
				allColumns[i].checked = false;
				// Remove the column from the list
				// var div = col.parentNode.parentNode;
				// div.parentNode.removeChild(div);
			}
		}
		
		if (numSelected == 0)
		{
			Forms.showNotification('columnstable', 'Please select at least one column!', null, 'pointer-below');
			return;
		}
		Doc.element('added-column-msg').innerHTML = numSelected == 1 ? 
			'1 column has been added to the table.' : numSelected + ' columns have been added to the table.';
		Doc.show('added-column-msg', 'table');
		event.stopPropagation(); // To prevent the hideMessage() method from executing
		setTimeout(configure.hideMessage, 2000);
	}
	
	configure.hideMessage = function(event)
	{
		Doc.hide('added-column-msg');
	}
	
	return configure;
}();

Doc.onLoad(Configure.initPage);